/* $Id: ddl_blkilu.c,v 1.1 1995/08/24 15:41:02 tim Exp $ */

/* non-permuted ilu factorization support routines */

/* Copyright 1995 IASC */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "ddl.h"

/* calculate communication pattern for formation of ilu factors */

int DDL_ILU_comm(DDL_Spa a)
{
  int ierr = DDL_SUCCESS;
  int nproc, proc, dummy, offset, ret;
  int rcv_tot_data, snd_tot_data, nblks;
  int *ia, *ja ;
  double *ad;
  int nr, nc, n, size[3], i, j, low, high, subtype;
  int format, block, blsq, dim_blk_mat, nblk_rows;
  MPI_Comm comm;
  DDL_Prop p;
  DDL_Glob g;
  DDL_Ind ind;
  /*  DDL_Dummy d;
  DDL_Vector v, vx;*/
  char *tmp;

#ifdef _DEBUG_MALLOC_INC
  union dbmalloptarg  debug;

  debug.i = 1;
  dbmallopt(MALLOC_CKCHAIN, &debug);
#endif

  /* check parameters */
  if (a == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* only packed rows supported at the moment */
  ierr = DDL_Format_sparse(a, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
    return ierr;}

  ierr = DDL_Get_ia(a, &ia);
  ierr = DDL_Get_ja(a, &ja);
  ierr = DDL_Gsize_sparse(a, size);
  nr = size[0]; /* total number of rows */
  nc = size[1]; /* total number of columns */
  ierr = DDL_Size_sparse(a, size);  
  n = size[0]; /* number of rows I have */
  ierr = DDL_Offset_sparse(a, &offset);
  ierr = DDL_Block_sparse(a, &block); 
  ierr = DDL_Comm(a, &comm);
  ierr = MPI_Comm_size(comm, &nproc);
  ierr = MPI_Comm_rank(comm, &proc);


/* form some useful parameters */
  nblk_rows = nr/block;
  blsq=block*block;
  dim_blk_mat = nc*n/blsq;

  /* create communication structure */
  /* call it py for now */
  ierr = DDL_Create_prop(a, DDL_COMM_PY, &p);
  g = (p->data);
  g->rcv.proc = malloc(nproc*sizeof(int));
  g->rcv.ind = malloc(dim_blk_mat*sizeof(int));
  g->rcv.cnt = malloc(nproc*sizeof(int));
  g->rcv.ptr = malloc((nproc+1)*sizeof(int));
  g->snd.proc = malloc(nproc*sizeof(int));
  g->snd.ind = malloc(dim_blk_mat*sizeof(int));
  g->snd.cnt = malloc(nproc*sizeof(int));
  g->snd.ptr = malloc((nproc+1)*sizeof(int));
  
  /* create communication structure */
  ierr = DDL_Create_prop(a, DDL_COMM_IND, &p);
  ind = (p->data);
  ind->rcv.proc = malloc(dim_blk_mat*sizeof(int)); 
  ind->snd.proc = malloc(dim_blk_mat*sizeof(int));
  ind->snd.ind = malloc(dim_blk_mat*sizeof(int));
  ind->snd.cnt = malloc(nc*sizeof(int));
  ind->snd.ptr = malloc((nc+1)*sizeof(int));

  /* How many blocks do I own ? */
  nblks=n*nc/(block*block);
  
  /* find which blocks exist using 1=exist 0=not exist representation */

  tmp = malloc(nblks*sizeof(char));
  ierr=DDL_Findnz_blk(ia,ja,n,nc,block,nproc,proc,tmp);

  /* pack information in row-gathered style */
  /* index=col_ind*block+row_ind */

  ierr=DDL_Pack_blk(tmp,g->rcv.ind,nblks,&rcv_tot_data);
  free(tmp); 

  /* send-receive data from all other processes */

  ierr = DDL_Calccomm_blk(ia, ja, n, nc, block,
                      proc, nproc, comm,
                      &(g->rcv.nproc), &(g->snd.nproc),
                      &rcv_tot_data, &snd_tot_data,
                      g->rcv.proc, g->snd.proc,
                      g->rcv.ptr, g->snd.ptr,
                      g->rcv.cnt, g->snd.cnt,
                      g->rcv.ind, g->snd.ind);

  /* truncate global rcv structures */

  ierr = DDL_Truncate(&(g->rcv.proc), rcv_tot_data);
  ierr = DDL_Truncate(&(g->rcv.ind), rcv_tot_data);
  ierr = DDL_Truncate(&(g->rcv.cnt), g->rcv.nproc);
  ierr = DDL_Truncate(&(g->rcv.ptr), g->rcv.nproc+1);

  /* order the send data by index=col_ind*block+row_ind */

  ierr= DDL_Blksend(nproc,nc,block,proc,
               &(g->rcv.nproc), &(g->snd.nproc),
               &rcv_tot_data, &snd_tot_data,
               g->snd.proc,
               g->snd.ptr,
               g->snd.cnt,
               g->snd.ind,
               ind->snd.ptr,
               ind->snd.cnt,
               ind->snd.ind);

  /* truncate global send structures */

  ierr = DDL_Truncate(&(g->snd.proc), (g->snd.nproc));
  ierr = DDL_Truncate(&(g->snd.ind), snd_tot_data);
  ierr = DDL_Truncate(&(g->snd.cnt), g->snd.nproc);
  ierr = DDL_Truncate(&(g->snd.ptr), g->snd.nproc+1);



  /* order the send data block row by block row */

  ierr = DDL_BlkReorder(proc,&(g->snd.nproc),
                  &snd_tot_data,
                  g->snd.proc,
                  g->snd.ptr,
                  g->snd.cnt,
                  g->snd.ind,
                  g->rcv.nproc,
                  &rcv_tot_data,
                  g->rcv.proc,
                  g->rcv.ptr,
                  g->rcv.cnt,
                  g->rcv.ind,
                  ind->rcv.proc,
                  ind->snd.ind, ind->snd.ptr,
                  ind->snd.cnt, ind->snd.proc);

   
  /* keep order, but re-number the data according to column index */

  ierr = DDL_NCBlks(ia, ja, a, ind->snd.ind,
                  ind->snd.ptr, ind->snd.cnt,
                  ind->snd.proc, 
                  block, n, nc,
                  nproc, proc);

  /* truncate index send structures */

  ierr = DDL_Truncate(&(ind->rcv.proc), rcv_tot_data);
  ierr = DDL_Truncate(&(ind->snd.proc), snd_tot_data);
  ierr = DDL_Truncate(&(ind->snd.ind), snd_tot_data);
  ierr = DDL_Truncate(&(ind->snd.cnt), nblk_rows);
  ierr = DDL_Truncate(&(ind->snd.ptr), nblk_rows+1);
}

/* find which blocks exist using 1=non-zero block */
/* 0=zero block representation */

int DDL_Findnz_blk(int *ia, int *ja, int nr,
                    int nc, int block, int nproc, int proc,
                             char *temp)
{
 int i,j,low,high,rowblocks,colblocks,rowlow,rowhigh;
 /* find allocated row blocks */
 rowblocks = nr/block;
 colblocks = nc/block;
 rowlow = DDL_Find_offset(nc, block, nproc, proc);
 rowhigh = DDL_Find_offset(nc, block, nproc, proc+1);

 for (i=0;i<nc*nr/(block*block);i++) temp[i] = 0;
 /* find non-zero blocks */

 for (i=0;i<nr;i++) {
    DDL_Coeff_ind(i, ia, &low, &high);
    for (j=low-1;j<high-1;j++) {
      /*store indices by using representation (j*colsize +i) */
      /* column indices in ja are global when stored by rows */
      /* only store lower triangular blocks */
      if(((ja[j]-1)/block) <= ((rowlow+i)/block))
        temp[((ja[j]-1)/block)*rowblocks+i/block]=1;
    }
  }
  return DDL_SUCCESS;
}



int DDL_Pack_blk(char *sparse, int *packed, int n, int *nz)
{
  int i, j;
  /* pack non-zero columns into ind vector as C indices */
  j = 0;
  for (i=0;i<n;i++) {
    if (sparse[i] != 0) {
      packed[j] = i;
      j=j+1;
    }
  }
  *nz = j;
  return DDL_SUCCESS;
}



int DDL_Calccomm_blk(int *ia, int *ja, int n, int nc, int block,
                 int proc, int nproc, MPI_Comm comm, 
                 int *rcv_tot_proc, int *snd_tot_proc,
                 int *rcv_tot_data, int *snd_tot_data,
                 int rcv[], int snd[],
                 int rcv_ptr[], int snd_ptr[],
                 int rcv_cnt[], int snd_cnt[],
                 int rcv_ind[], int snd_ind[])
{
  int cnt, i, next, next_proc, ierr, p;
  ierr = DDL_SUCCESS;

  /* set up receive structure */
  ierr = DDL_Packcomm_blk(nc,n,block,proc,nproc,rcv_tot_proc,
                rcv_tot_data, rcv, rcv_ptr, rcv_ind);

  /* exchange elements */
  ierr = DDL_Excomm_blk(nproc, comm, proc,
                    rcv_tot_proc, snd_tot_proc,
                    rcv_tot_data, snd_tot_data,
                    rcv, snd,
                    rcv_ptr, snd_ptr,
                    rcv_cnt, snd_cnt,
                    rcv_ind, snd_ind);

  return ierr;
}

int DDL_Packcomm_blk(int n, int nr, int block,
                 int dest, int nproc,
                 int *tot_proc, int *tot_data,
                 int *src, int *ptr, int *ind)
{
  int ierr = DDL_SUCCESS;
  int size, cnt, i, j, low, high, npg, num_proc, numblocks;
  int block_low, block_high;

  *tot_proc=0;

  num_proc = 0;
  numblocks=nr/block;
  /* get range of vector elements held by process proc */
  low = DDL_Find_offset(n, block, nproc, num_proc);
  high = DDL_Find_offset(n, block, nproc, num_proc+1);

  cnt = 0;
  npg = 0;
  ptr[0] = 0;
  for (i=0;i<*tot_data;i++) {
    if (ind[i]/numblocks >= high/block) {

      if (cnt != 0) {
        src[npg] = num_proc;
        ptr[npg+1] = ptr[npg]+cnt;
        cnt = 0;
        npg = npg+1;
      }
      while (ind[i]/numblocks >= high/block) {
        num_proc = num_proc+1;
        low=high;
        high=DDL_Find_offset(n, block, nproc, num_proc+1);
      }
    }
    cnt = cnt+1;
  }


  if (cnt != 0) {
    src[npg] = num_proc;
    ptr[npg+1] = ptr[npg]+cnt;
    npg = npg+1;
  }
  *tot_proc = npg;
    
}

/* exchange elements with other processes
   for sparse matrix vector multiply and triangular solve
   nproc:           is total number of processes distributed object uses
   comm:            communicator used by object
   rcv_tot_proc:    is number of procs src must get data from
   snd_tot_proc:    is number of procs src must send data to
   rcv_tot_data:    is total number of elements of x vector that src must get
   snd_tot_data:    is total number of elements of x vector that src must send
   rcv:             is vector of process numbers that src must get data from
   snd:             is vector of process numbers that src must send data to
   rcv_ind:         vector contains global C indices of elements of x that src
                    must get from each process in rcv
   snd_ind:         vector contains global C indices of elements of x that src
                    must send to each process in snd
   rcv_ptr[j]:      is C index of first element in rcv_ind that src must get from
                    rcv[j]
   snd_ptr[j]:      is C index of first element in snd_ind that src must send to
                    snd[j]
   rcv_cnt:         number of elements from each process
   snd_cnt:         number of elements to each process
   */
int DDL_Excomm_blk(int nproc, MPI_Comm comm, int proc,
               int *rcv_tot_proc, int *snd_tot_proc,
               int *rcv_tot_data, int *snd_tot_data,
               int rcv[], int snd[],
               int rcv_ptr[], int snd_ptr[],
               int rcv_cnt[], int snd_cnt[],
               int rcv_ind[], int snd_ind[])
{
  int cnt, i, next, next_proc, ierr;
  ierr = DDL_SUCCESS;

  for (i=0;i<nproc;i++)
    rcv_cnt[i] = 0;
  for (i=0;i<*rcv_tot_proc;i++)
    rcv_cnt[rcv[i]] = rcv_ptr[i+1]-rcv_ptr[i];

  /* everyone sends messages to everyone else saying how many
     elements they need from each process */
  MPI_Alltoall(rcv_cnt, 1, MPI_INT, snd_cnt, 1, MPI_INT, comm);

  /* set up arrays snd, snd_ptr */
  next_proc = 0;
  next = 0;
  for (i=0;i<nproc;i++) {
    cnt = snd_cnt[i];
    if (cnt != 0) {
      snd[next_proc] = i; /* remember process number */
      snd_ptr[next_proc] = next;
      next_proc = next_proc+1;
      next = next+cnt;
    }
  }
  snd_ptr[next_proc] = next;
  *snd_tot_proc = next_proc;
  *snd_tot_data = next;
 
  /* set up count vectors for DDL_Alltoallvs */

  for (i=0;i<*rcv_tot_proc;i++)
    rcv_cnt[i] = rcv_ptr[i+1]-rcv_ptr[i];
  for (i=0;i<*snd_tot_proc;i++)
    snd_cnt[i] = snd_ptr[i+1]-snd_ptr[i];

  /* exchange index vectors to complete snd_ind */
  ierr = DDL_Alltoallvs(rcv_ind, *rcv_tot_proc,
                        rcv, rcv_cnt, rcv_ptr, MPI_INT,
                        snd_ind, *snd_tot_proc,
                        snd, snd_cnt, snd_ptr, MPI_INT,
                        comm);
  return ierr;  
}


 int DDL_Blksend(int nproc, int nc, int block, int proc,
               int *rcv_tot_proc, int *snd_tot_proc,
               int *rcv_tot_data, int *snd_tot_data,
               int snd[],
               int snd_ptr[],
               int snd_cnt[],
               int snd_ind[],
               int blksnd_ptr[],
               int blksnd_cnt[],
               int blksnd_ind[])
{
  int cnt, i, j, next, next_proc, ierr, pind, nrblks, low, high;
  ierr = 0;

  for(i=0;i<*snd_tot_proc;i++) {
    low = DDL_Find_offset(nc, block, nproc, snd[i]); 
    high = DDL_Find_offset(nc, block, nproc, snd[i]+1);
    nrblks = (high-low)/block;
    for (j=snd_ptr[i];j<snd_ptr[i+1];j++) blksnd_ind[j]=snd_ind[j]/nrblks;
  }
  blksnd_ptr[0]=0;
  cnt=0;
  pind=0;
  for(i=1; i<=*snd_tot_data; i++) {
    if (blksnd_ind[i]!=blksnd_ind[i-1]) {
      cnt=cnt+1;
      blksnd_ptr[pind+1]=blksnd_ptr[pind]+cnt;
      blksnd_cnt[pind]=cnt;
      pind =pind+1;
      cnt=0;
      }
    else cnt=cnt+1;
  }
  if(cnt!=0) {
    cnt=cnt+1;
    blksnd_ptr[pind+1]=blksnd_ptr[pind]+cnt;
    blksnd_cnt[pind]=cnt;
  }
}


int DDL_BlkReorder(int proc, int *psnd_tot_proc,   
                  int *psnd_tot_data,
                  int psnd[],
                  int psnd_ptr[],
                  int psnd_cnt[],
                  int psnd_ind[],
                  int *prcv_tot_proc,
                  int *prcv_tot_data,
                  int prcv[],
                  int prcv_ptr[],
                  int prcv_cnt[],
                  int prcv_ind[],
                  int blkrcv_proc[],
                  int blksnd_ind[], int blksnd_ptr[],
                  int blksnd_cnt[], int blksnd_proc[])
  
{
  int ierr = 0;
  int *tmp, *tmp1, i, j, bi, p, cnt, ind;
  int blksnd_nind;

  /* expand ind->rcv.proc list */
  tmp = malloc(*prcv_tot_proc*sizeof(int));
  for (i=0;i<*prcv_tot_proc;i++) tmp[i] = prcv[i];
   /* expand process list */
  p = 0;
  for (i=0;i<*prcv_tot_data;i++) {
    if (i >= prcv_ptr[p+1]) p = p+1;
    blkrcv_proc[i] = tmp[p];
  }
  free(tmp);

   /* expand process list */
  p = 0;
  for (i=0;i<*psnd_tot_data;i++) {
    if (i >= psnd_ptr[p+1]) p = p+1;
    blksnd_proc[i] = psnd[p];
  }

  /* reorder send block data structure */
  tmp = malloc(*psnd_tot_data*sizeof(int));
  tmp1 = malloc(*psnd_tot_data*sizeof(int));
  for(i=0; i<*psnd_tot_data; i++) tmp1[i]=i;
  DDL_qs2i(blksnd_ind, tmp1, *psnd_tot_data, 1); /* sort indices in t2 */
  /* compact again grouping by index */

  i = 0;
  cnt = 1;
  ind = blksnd_ind[0];
  blksnd_ptr[i] = 0;
  for (j=1;j<*psnd_tot_data;j++) {
    if (blksnd_ind[j] != ind) {
      blksnd_cnt[i] = cnt;
      i = i+1;
      blksnd_ptr[i] = j;
      ind = blksnd_ind[j];
      cnt = 0;
    }
    cnt = cnt+1;
  }

  blksnd_cnt[i] = cnt;
  blksnd_ptr[i+1] = j;
  blksnd_nind = i+1;

  /* sort each column of tmp1 according to column blocks */

  for (i=0;i<blksnd_nind;i++) {
   DDL_qs1i(&tmp1[blksnd_ptr[i]], blksnd_cnt[i], 1);
  }

  for(i=0; i<*psnd_tot_data; i++) blksnd_ind[i]=psnd_ind[tmp1[i]];
  
  for(i=0; i<*psnd_tot_data; i++) tmp[i]=blksnd_proc[i];
  for(i=0; i<*psnd_tot_data; i++) blksnd_proc[i]=tmp[tmp1[i]];

/* place a minus one at the end of blksnd_proc */

 blksnd_proc[*psnd_tot_data]=-1;
 free(tmp);
 free(tmp1);
}

/* reorder the comms data so that each block column list
within each block row is ordered, with the diagonal first */

int DDL_NCBlks(int *ia, int *ja, double a[], int blksnd_ind[],
                  int blksnd_ptr[], int blksnd_cnt[],
                  int blksnd_proc[],
                   int block, int nr, int nc,
                  int nproc, int proc)

{
  int diag_cnt, blocknum, ierr;
  int low, high, offset, i, j, k;
  int blk_off,blkrows,blkrows_proc,low_proc,high_proc;
  int myrowdiag,gmyblockdiag,diagblk,myblockdiag,blk;
  int offset_proc, rows_proc;
  ierr=0;
  diag_cnt=0;
  low=DDL_Find_offset(nc, block, nproc, proc);
  high=DDL_Find_offset(nc, block, nproc, proc+1);
  blkrows=(high-low)/block;
  for (k=0; k<nr/block; k++) {
    diagblk=0;
    gmyblockdiag=k+low/block;
    myblockdiag=k;
    /* scan along row */
    for (blk=blksnd_ptr[myblockdiag];blk<blksnd_ptr[myblockdiag+1];blk++) {
      /* translate to block column indices for myblock row */
      if (diagblk!=0) {
        if (blksnd_proc[blk]==proc) { /* block resides on my proc */
          blk_off=blksnd_ind[blk]%blkrows;
          blksnd_ind[blk]=low/block+blk_off;
        }
        else {   /* block resides on another proc  */
          low_proc=DDL_Find_offset(nc, block, nproc, blksnd_proc[blk]);
          high_proc=DDL_Find_offset(nc, block, nproc, blksnd_proc[blk]+1);
          rows_proc=high-low;
          blkrows_proc=rows_proc/block;
          blksnd_ind[blk]=low_proc/block+blksnd_ind[blk]%blkrows_proc;
        }
      } 
      else {     /*  diagonal block  */
        blksnd_ind[blk]=gmyblockdiag;
        diagblk=1;
      }
    }
  }
}


/* perform global sparse ILU decomposition 
   a: IN           sparse triangular matrix
   d: OUT          block diagonal of ILU factors
*/

int DDL_ILU_form(DDL_Spa sp, DDL_Spa l,
           DDL_Spa u, DDL_Vec diag_blk)

{
  int ierr, i, j, k, low, high, next, ret;
  int x_size, x_itype, d_offset, offset;
  MPI_Comm x_comm;
  int sp_size[3], sp_itype;
  MPI_Comm sp_comm;
  MPI_Datatype sp_type, type, d_type;
  MPI_Status status;
  int *ia, *ja, *ipvt;
  void *a, *d, *snd_buf, *rcv_buf;
  int *snd_cnt, *rcv_cnt, *tmp;
  int format, sformat, size, proc;
  MPI_Aint typesize;
  int n, m, nr, ni, nelt, blsq, bli;
  DDL_Prop p, p1;
  DDL_Glob gx;
  DDL_Ind iy;
  double val, coeff;
  double *blk_mess, *dblk, *den_l, *dinv, *dtmp;
  int li, lj, nc, nextr, nextx, nexts, nblk_rows;
  int d_size, block, nproc, info, blk_col, rec_proc;
  int high_blk, loc_row, last_proc, messblks, numblks, recblks;
  int col, startr, d_ptr, dptr, blk_offset, d_loc, dummy, ri;
  int *l_ia, *l_ja, *u_ia, *u_ja;
  void *l_a, *u_a;
  int zero = 0;
  int one = 1;
  double dzero=0.0;
  double done=1.0;
  double dnegone=-1.0;
  char no[1];
  int row_cnt,row_no;
  ierr = DDL_SUCCESS;
 
  strcpy(no, "n");
  /* check input parameters */
  ierr = DDL_Comm_sparse(sp, &sp_comm);
  ierr = MPI_Comm_rank(sp_comm, &proc);
 
  ierr = DDL_Gsize_sparse(sp, sp_size);
  n = sp_size[0]; /* total number of rows */
  nc = sp_size[1]; /* total number of columns */
  ierr=DDL_Size_sparse(sp, sp_size);
  nr= sp_size[0];  /* number of rows I have */
  ierr = DDL_Type_sparse(sp, &sp_type);
  ierr = DDL_Comm_sparse(sp, &sp_comm);
  ierr = DDL_Type_vector(diag_blk, &d_type);
  
  ierr = DDL_Format_sparse(sp, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}

  /* get pointers to raw data values */
  ierr = DDL_Get_ia(sp, &ia);
  ierr = DDL_Get_ja(sp, &ja);
  ierr = DDL_Get_a(sp, &a);
  ierr = DDL_Get_vector(diag_blk, &d);
  ierr = DDL_Size_vector(diag_blk, &d_size);
  ierr = DDL_Size_vector(diag_blk, &d_offset);
  ierr = MPI_Type_size(d_type, &typesize);
  ierr = DDL_Block_sparse(sp, &block);
  ierr = MPI_Comm_size(sp_comm, &nproc);

  /* check global comms. data exists, else create */
  ierr = DDL_Find_prop(sp, DDL_COMM_PY, &p);
  ierr = DDL_Find_prop(sp, DDL_COMM_IND, &p1);

  /* get global comms structures */
  ierr = DDL_Find_prop(sp, DDL_COMM_PY, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }

  if (p!= NULL)
    gx= (DDL_Glob)(p->data);
  else
    gx=NULL;

  ierr=DDL_Find_prop(sp, DDL_COMM_IND, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE);
    if (ierr > 0) return ierr; /* cleanup and exit with error
 */
  } 
  iy = (DDL_Ind)(p->data);

  /* get pointers to raw data values */
  ierr = DDL_Get_ia(u, &u_ia);
  ierr = DDL_Get_ja(u, &u_ja);
  ierr = DDL_Get_a(u, &u_a);
  /* get pointers to raw data values */
  ierr = DDL_Get_ia(l, &l_ia);
  ierr = DDL_Get_ja(l, &l_ja);
  ierr = DDL_Get_a(l, &l_a);

/* form some useful parameters */
  offset = DDL_Find_offset(nc, block, nproc, proc);
  blk_offset=offset/block;
  high_blk=blk_offset+nr/block;
  nblk_rows = nr/block;
  row_cnt=0;

 /* form the L and U matrices */  
  ierr=DDL_LUblk_extract(ia,ja,a,
                       l_ia,l_ja,l_a,
                       u_ia,u_ja,u_a,
                       block,nr,nc,
                       nproc,proc);
  if(block!=1){
  /* form the skeleton blk data structures for u */
  ierr=DDL_U_setup(iy->snd.ptr,iy->snd.cnt,u_ia,block,nr,nc,
                   nproc,proc,nblk_rows);
  }

  blsq=block*block;
  dblk = malloc(blsq*typesize);
  dinv = malloc(blsq*typesize);
  den_l = malloc(blsq*typesize);
  dtmp = malloc(blsq*typesize);
  for (i=0;i<nr;i++) {
    ((double*)dblk)[i] = 0;
    ((double*)dinv)[i] = 0;
    ((double*)den_l)[i] = 0;
    ((double*)dtmp)[i] = 0;
  }


  /* fill DDL d object with block diagonals */

  for (k=0;k<d_size;k++) ((double*)d)[k]=0;

  for (k=0;k<nr/block;k++) {
    blk_col = k+blk_offset;
    ierr= DDL_Form_Mults_Blk(ia,ja,a,blk_col,dblk,
          k,block,nr,nc,nproc,proc);   
   for (i=0;i<blsq;i++)
    ((double*)d)[k*blsq+i]=dblk[i];
  }

  blk_mess=malloc(nr*block*sizeof(double));
  ipvt=malloc(block*sizeof(int));
  for (i=0;i<block;i++) ipvt[i]=0;

  /* main loop to form LU factors */
  /* use block indexing */

  /* initialize some counts */
  nextx=0;
  nextr=0;
  recblks=0;
  messblks=0;
  for (k=0;k<n/block;k++) {

    if (k==blk_offset+nextx && k < high_blk) {
          
      /* I own next block row */
      /* reset row_cnt */

      row_cnt =0;
      /* insert ones at the beginning of each line of U */
      for (row_no=k*block;row_no<k*block+block;row_no++){
        u_ja[u_ia[row_no-offset]-1]=row_no+1;
        ((double*)u_a)[u_ia[row_no-offset]-1]=1.0;
      }
      row_cnt=row_cnt+1;

      /* form the row's inverse diagonal block */
      for (i=0;i<blsq;i++) { 
        ((double*)dblk)[i]=((double*)d)[(k-blk_offset)*blsq+i];
      }

      F77(dgetrf)(&block,&block,dblk,&block,ipvt,&info);
     
      for (i=0;i<blsq;i++) den_l[i]=0.0; 

      for (i=0;i<blsq;i=i+block+1) ((double*)den_l)[i]=1.0;

      F77(dgetrs)("n",&block,&block,dblk,&block,ipvt,den_l,&block,&info);

      F77(dcopy)(&blsq,den_l,&one,dinv,&one);

      d_ptr=0;
      /* Find the other blocks (on the send list) on my procs */
          
      for (i=iy->snd.ptr[k-blk_offset];i<iy->snd.ptr[k+1-blk_offset];i++) {
        if (iy->snd.ind[i]!=k) {
          /* not a block diagonal */
          if (iy->snd.ind[i] < high_blk){
            /* internal calculation updates */
            /* find upper block */
            ierr= DDL_Form_Mults_Blk(ia, ja, a,
             iy->snd.ind[i],dblk,k-blk_offset,block,nr,nc,nproc,proc);

            /* multiply by dinv */

            F77(dgemm)(&no,&no,&block,&block,&block,&done,dinv,
                  &block,dblk,&block,&dzero,den_l,&block);

            ierr=DDL_Insert_Blk(u_ia, u_ja, u_a,
                      iy->snd.ind[i], den_l,
                      k-blk_offset, block, nr, nc,
                      nproc, proc, &row_cnt);

            /* find lower block and update diagonal */

            ierr=DDL_Form_Mults_Blk(ia, ja, a,
            k,dblk,iy->snd.ind[i]-blk_offset,block,nr,nc,nproc,proc);

            d_loc=(iy->snd.ind[i]-blk_offset)*blsq;

            F77(dfill)(&blsq,dtmp,&dzero);

            F77(dgemm)(&no,&no,&block,&block,&block,&done,dblk, 
               &block,den_l,&block,&dzero,dtmp,&block);
             
            for(bli=0;bli<blsq;bli++) 
               ((double*)d)[d_loc+bli]=((double*)d)[d_loc+bli]-dtmp[bli];

          }

          else if (iy->snd.ind[i] >= high_blk) {

            /* send to new processor or append to existing message */

            /* find upper block */

            ierr= DDL_Form_Mults_Blk(ia, ja, a, iy->snd.ind[i],
              dblk,k-blk_offset,block,nr,nc,nproc,proc);

            /* multiply by dinv */
 
            F77(dfill)(&blsq,den_l,&dzero);

            F77(dgemm)(&no,&no,&block,&block,&block,&done,dinv,    
                  &block,dblk,&block,&dzero,den_l,&block);

            ierr=DDL_Insert_Blk(u_ia, u_ja, u_a,              
                      iy->snd.ind[i], den_l,        
                      k-blk_offset, block, nr, nc,           
                      nproc, proc, &row_cnt);           

            /* append to message */

            for(bli=0;bli<blsq;bli++) blk_mess[bli+d_ptr]=den_l[bli];
            d_ptr = d_ptr+blsq;
            messblks=messblks+1;
            last_proc = iy->snd.proc[i];
                  
            /* send last message for block row */

            if (iy->snd.proc[i+1]!=iy->snd.proc[i]){ 
                                                   
              /* send to proc containing lower block and update diagonal */
            
              if (messblks){
                  DDL_Msend(blk_mess, messblks*blsq, MPI_DOUBLE,
                    1,&last_proc,0,sp_comm);
                  d_ptr=0;
                messblks=0;
              }
            } 
          }        
        }
      }
      nextx=nextx+1;
    }  

    else if (k==gx->rcv.ind[nextr]/nblk_rows) {
      /* find which processor to receive from */
      rec_proc = iy->rcv.proc[nextr];
      startr=nextr;
      /* find how many blocks to receive*/
      recblks=1;
      while(k==(gx->rcv.ind[nextr+1])/nblk_rows) {
        recblks=recblks+1;
        nextr=nextr+1; 
      }
      
      /*  receive the data */
      ierr=MPI_Recv(blk_mess,recblks*blsq,MPI_DOUBLE,rec_proc,0,sp_comm,&status);  
         
     /* update the block diagonals */
      for (ri=0;ri<recblks;ri++) {

        /* calculate ptr to (local) block diagonal */ 
        loc_row=(gx->rcv.ind[startr+ri]%(nblk_rows));
        d_loc = loc_row*blsq;

        /* col should equal k !! */
        col=(gx->rcv.ind[startr+ri]/(nblk_rows));
        d_ptr=loc_row*blsq;

        /* find lower block */

        ierr= DDL_Form_Mults_Blk(ia, ja, a,
                   col,den_l,loc_row,block,nr,nc,nproc,proc);

        F77(dfill)(&blsq,dtmp,&dzero);
      
        F77(dgemm)(&no,&no,&block,&block,&block,&done,den_l,
          &block,&blk_mess[ri*blsq],&block,&dzero,dtmp,&block);
       
        for(bli=0;bli<blsq;bli++) 
          ((double*)d)[d_loc+bli]=((double*)d)[d_loc+bli]-dtmp[bli];   
      }
    nextr=nextr+1;
    }
  }   

  if(block==1){
    /* form lower triangular matrix */
    /* assume diagonal entry is last index in every row ? */
    for(i=1;i<=n;i++)
      ((double*)l_a)[l_ia[i]-2]=((double*)d)[i-1];
  }   

  return ierr;
}
   
/* routine to extract actual coefficients a from data
and place appropriately in equivalent block */
/* optimized using bsearch ? */
  
int DDL_Form_Mults_Blk(int *ia, int *ja, double *a,
                      int blk_col, double blkcoeffs[],
                      int blk_row, int block, int nr, int nc,
                      int nproc, int proc)

{
  int first_row,last_row,i,j,jind,row_ind,offset,idx,blk_ind;
  int col_ind,range;
  int ierr;
  ierr=DDL_SUCCESS;
    
  /* equivalent of dfill */
  for (i=0; i<block*block; i++) blkcoeffs[i]=0.0;
  /* Find actual values in block */
  /* local indexing for rows */
  first_row = blk_row*block;
  last_row = first_row+block-1;

  row_ind=0;
  for (i=first_row; i<=last_row; i++) {
    for (j=0;j<block;j++) {
      col_ind = blk_col*block+j+1;
      range = ia[i+1]-ia[i];
      ierr=DDL_bsearch(&ja[ia[i]-1],range,col_ind,&idx);
      blk_ind = (i-first_row)+j*block;
      if (ierr==DDL_ERROR) {
        blkcoeffs[blk_ind]=0.0;
      }
      else blkcoeffs[blk_ind]=a[ia[i]-1+idx];
    }
  }
}  

/* subroutine to setup data structures for the (D-1)U matrix  */
/*  calling sequence :-                                       */
/*  ierr=DDL_U_setup(iy->snd.ptr,iy->snd.cnt,l_ia,block,nr,nc,*/
/*                   nproc,proc);                         */

int DDL_U_setup(int snd_ptr, int snd_cnt[], int *u_ia,
                int block, int nr, int nc,
                int nproc, int proc, int nblk_rows)
{  
  int i,k,bli,num_ents,row_ind;
  int ierr=DDL_SUCCESS;
  u_ia[0]=1;
  for(k=0; k<nblk_rows; k++){
    /* number of entries per row=diag_entry+blocksize*no.offdiag.blocks */
    num_ents=1+(snd_cnt[k]-1)*block;
    for(bli=0; bli<block; bli++){
      row_ind = k*block+bli;     
      u_ia[row_ind+1]=u_ia[row_ind]+num_ents;
    }
  }
}

     
/* routine to insert factored blocks into U matrix data structs */
  
int DDL_Insert_Blk(int *u_ia, int *u_ja, double *u_a,
                   int blk_col, double blkcoeffs[],
                   int blk_row, int block, int nr, int nc,
                   int nproc, int proc, int *row_cnt)
    
{
    
  int first_row,last_row,i,j,jind,row_ind,offset;
  int uia_ptr,row_insert,col_start;
  int ierr;
  ierr=DDL_SUCCESS;
       
  /* Find actual values in block */
  /* local indexing for rows */
  first_row = blk_row*block;   
  last_row = first_row+block-1;

  for(j=0;j<block;j++){
    for(i=0;i<block;i++){
      /* i = local row index within block, first_row = row index base */
      /* row_cnt = placing within row (-1) for fortran indexing */
      /* j = local col index, blk_col*block = global column index */
      /*   for start of block, +1 is for diagonal element  */
      uia_ptr = i+first_row;
      row_insert = *row_cnt-1;
      col_start=blk_col*block;
      u_ja[u_ia[uia_ptr]+row_insert]=col_start+j+1;
      ((double*)u_a)[u_ia[uia_ptr]+row_insert]=blkcoeffs[j*block+i];
    }
    *row_cnt=*row_cnt+1;
  }  
}



/* routine to insert factored blocks into U matrix
very similar to DDL_Form_Mults_Blk (will combine later) */
/* ok if block size = 1 */  

int DDL_Insert(int *ia, int *ja, double *a,
               int blk_col, double blkcoeffs[],
               int blk_row, int block, int nr, int nc,
               int nproc, int proc)
  
{
  
  int first_row,last_row,i,j,jind,row_ind,offset;
  int ierr;

  ierr=DDL_SUCCESS;

  /* Find actual values in block */
  /* local indexing for rows */
  first_row = blk_row*block;
  last_row = first_row+block-1;
  row_ind=0;
  for (i=first_row; i<=last_row; i++) {
    for (jind=ia[i]-1; jind<ia[i+1]-1; jind++) {
      if((ja[jind]-1)/block==blk_col) {
      a[jind]=blkcoeffs[row_ind+((ja[jind]-1)%block)*block];
      }
    }
  row_ind=row_ind+1;
  }
}


/* Extract L, U matrices from A matrix */ 

int DDL_LUblk_extract(int *ia, int *ja, double *a,
                       int *l_ia, int *l_ja, double *l_a,
                       int *u_ia, int *u_ja, double *u_a,
                       int block, int nr, int nc,
                       int nproc, int proc)
    
{
    
  int offset,l_count,u_count,l_ptr,u_ptr,i,j,jind,blk_add;
  int ierr;
  ierr=DDL_SUCCESS;
          
  offset=DDL_Find_offset(nc, block, nproc, proc);

  l_count=0;
  u_count=0;
  l_ia[0]=1;
  u_ia[0]=1;
                       
  for (i=0;i<nr;i++){
    /* insert 1.0 at start of upper row */
    u_ja[u_count]=i+1+offset;
    u_a[u_count]=1.0;
    u_count=u_count+1;
    for (j=ia[i]-1;j<ia[i+1]-1;j++){
       blk_add=(block-1)-i%block;
       if ((ja[j]-1)<=i+offset+blk_add){
          /* lower triangular element */
          l_ja[l_count]=ja[j];
          l_a[l_count]=a[j];
          l_count=l_count+1;
       }
       else if ((ja[j]-1)>i+offset){
          /* upper triangular element */
          u_ja[u_count]=ja[j];
          u_a[u_count]=a[j];
          u_count=u_count+1;
       }
    }
          
    l_ia[i+1]=l_count+1;
    u_ia[i+1]=u_count+1;

  }
        
}
