/* $Id: ddl_image.c,v 1.2 1995/03/15 13:25:35 tim Exp $ */

#include <stdio.h>
/*#include "ddl_debug.h"
#include "ddl_intmath.h"*/
#include <mpi.h>
#include <stdlib.h>
#include <malloc.h>
/*#include "ddl_image.h"*/

#include "ddl.h"

int
DDL_Ishost()
{
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  return rank==0;
}

int
DDL_Image_create(size, type, max_overlap, block, comm, image)
  int size[2];
  int max_overlap;
  int block;
  MPI_Comm comm;
  MPI_Datatype type;
  DDL_Image *image;
{
  struct DDLR_Image *im;
  int status, error;
  int nd, i;
  int dims[2], periods[2], coords[2];
  int is_vector_row, is_vector_col;
  MPI_Aint datasize;

  im = malloc(sizeof(struct DDLR_Image));
  MPI_Topo_test(comm, &status);
  if (status==MPI_CART) {
    MPI_Cartdim_get(comm, &nd);
    if (nd != 2) {
       error = DDL_Error(DDL_ERR_COMMTOP,DDL_ERR_HARD,DDL_IMAGE_CREATE);
       MPI_Abort(comm, 1);
    }
    MPI_Cart_get(comm, 2, dims, periods, coords);
    is_vector_row = (size[0] == 1 && dims[0] == 1);
    is_vector_col = (size[1] == 1 && dims[1] == 1);
    if ((!is_vector_row && size[0] % block != 0) || 
	(!is_vector_col && size[1] % block !=0))
    {
       error = DDL_Error(DDL_ERR_BLOCK,DDL_ERR_HARD,DDL_IMAGE_CREATE);
       MPI_Abort(comm,1);
    }
    for (i=0; i<2; i++) {
      int general_size;
      int nproc;
      if ((i == 0 && !is_vector_row) || (i == 1 && !is_vector_col))
      {
         general_size = block*DDL_CD(size[i]/block, dims[i]);
         nproc = DDL_CD(size[i],general_size);
      }
      else
      {
	 general_size = 1;
	 nproc = 1;
      }
      if (nproc != dims[i])
      {
          error = DDL_Error(DDL_ERR_BLOCK,DDL_ERR_HARD,DDL_IMAGE_CREATE);
          MPI_Abort(comm,1);
      }
      if (max_overlap > size[i] - general_size*(nproc-1)
	  && dims[i] > 1)
      {
          error = DDL_Error(DDL_ERR_OVER,DDL_ERR_HARD,DDL_IMAGE_CREATE);
          MPI_Abort(comm,1);
      }
      im->offset[i] = DDL_Min(coords[i]*general_size, size[i]);
      im->local_size[i] = DDL_Min((coords[i]+1)*general_size, size[i])
        - im->offset[i];
      im->global_size[i] = size[i];
      im->boundary[i][0] = max_overlap*(coords[i]>0);
      im->boundary[i][1] = max_overlap*(coords[i]<dims[i]-1);
    }
    im->overlap = max_overlap;
    im->comm = comm;
    im->block = block;          
    im->type = type;
    MPI_Type_extent(type, &datasize);
    im->data = malloc((unsigned)datasize *
                      (im->boundary[0][0]+im->boundary[0][1]+im->local_size[0]) *
                      (im->boundary[1][0]+im->boundary[1][1]+im->local_size[1]));
    im->rows = malloc(sizeof(void *)*(2*max_overlap+im->local_size[0]));
    im->rows = &im->rows[max_overlap];
    for (i = -max_overlap; i < im->local_size[0]+max_overlap; i++)
      im->rows[i] = (char *)im->data;
    for (i=0; i<im->boundary[0][0]+im->boundary[0][1]+im->local_size[0]; i++)
      im->rows[i-im->boundary[0][0]] = (char *)im->data
        + datasize*(i*
                    (im->boundary[1][0]+im->boundary[1][1]+im->local_size[1])
                    + im->boundary[1][0]);
    im->packed = 1;
    *image = im;
    return DDL_SUCCESS;
  } else {
       error = DDL_Error(DDL_ERR_COMMTOP,DDL_ERR_HARD,DDL_IMAGE_CREATE);
       MPI_Abort(comm, 1);
  }
  return DDL_ERR_HARD;
}


int
DDL_Image_free(im)
  DDL_Image im;
{
  free(&im->rows[-im->overlap]);
  free(im->data);
  free(im);
  return DDL_SUCCESS;
}


int
DDL_Image_comm(im, comm)
  DDL_Image im;
  MPI_Comm *comm;
{
  *comm = im->comm;
  return DDL_SUCCESS;
}


int
DDL_Image_type(im, type)
  DDL_Image im;
  MPI_Datatype *type;
{
  *type = im->type;
  return DDL_SUCCESS;
}


int
DDL_Image_global_size(im, size)
  DDL_Image im;
  int *size;
{
  size[0] = im->global_size[0];
  size[1] = im->global_size[1];
  return DDL_SUCCESS;
}


int
DDL_Image_local_size(im, size)
  DDL_Image im;
  int *size;
{
  size[0] = im->local_size[0];
  size[1] = im->local_size[1];
  return DDL_SUCCESS;
}


int
DDL_Image_boundary(im, boundary)
  DDL_Image im;
  int boundary[2][2];
{
  int i;

  for (i=0; i<2; i++) {
    boundary[i][0] = im->boundary[i][0];
    boundary[i][1] = im->boundary[i][1];
  }

  return DDL_SUCCESS;
}


int
DDL_Image_offset(im, offset)
  DDL_Image im;
  int *offset;
{
  offset[0] = im->offset[0];
  offset[1] = im->offset[1];
  return DDL_SUCCESS;
}


int
DDL_Image_max_overlap(im, overlap)
  DDL_Image im;
  int *overlap;
{
  *overlap = im->overlap;
  return DDL_SUCCESS;
}

int
DDL_Image_block(im, block)
  DDL_Image im;
  int *block;
{
  *block = im->block;
  return DDL_SUCCESS;
}


int
DDL_Image_rawdata(im, data)
  DDL_Image im;
  void **data;                  /* raw pointer to the data */
{
  *data = im->data;
  return DDL_SUCCESS;
}

int
DDL_Image_data(im, data)
  DDL_Image im;
  void **data;
{
  *data = im->rows;
  return DDL_SUCCESS;
}



