/* $Id: ddl_isr.c,v 1.2 1995/08/21 10:10:27 tim Exp $ */
#include <stdio.h>
#include "ddl_intmath.h"
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "ddl_isr.h"

static
struct {int num; ISR_TYPE type;}
isr_types[] = {{4, type_byte}, {5, type_int_2},
	       {6, type_int_4}, {7, type_real_4},
	       {8, type_real_8}, {9, type_cmpl_8},
	       {10, type_cmpl_16}, {-1, type_unknown}};

int
DDL_ISR_read_hdr(fp, n1, n2, isr_type, vmin, vmax)
  FILE *fp;
  int *n1;
  int *n2;
  ISR_TYPE *isr_type;
  double *vmin;
  double *vmax;
{
  char buffer[512];
  int i;
  int t;

  if (fread(buffer, 1, 512, fp) != 512)
    return MPI_ERR_UNKNOWN;
  
  sscanf(&buffer[88], "%d", n1);
  sscanf(&buffer[97], "%d", n2);
  sscanf(&buffer[124], "%lf", vmin);
  sscanf(&buffer[190], "%lf", vmax);
  sscanf(&buffer[115], "%d", &t);
  for (i=0; isr_types[i].num > 0; i++) {
    if (isr_types[i].num == t)
      break;
  }
  *isr_type = isr_types[i].type;
  if (*isr_type == type_unknown)
    return MPI_ERR_UNKNOWN;
  
  return MPI_SUCCESS;
}
  

int
DDL_ISR_read_data(fp, size, isr_type, data)
  FILE *fp;
  int size;
  ISR_TYPE isr_type;
  float *data;
{
  int i;
  fseek(fp, 512l, 0);
  if (isr_type==type_real_4) {
    while (size>0) {
      i = fread(data, 4, size, fp);
      size -= i;
      data += i;
    }
  } else {
    return MPI_ERR_UNKNOWN;
  }
  return MPI_SUCCESS;
}


int
DDL_ISR_write(fp, n1, n2, data)
  FILE *fp;
  int n1;
  int n2;
  float *data;
{
  char buffer[512];
  float vmin;
  float vmax;
  int i;

  vmin = vmax = data[0];
  for (i=1; i<n1*n2; i++) {
    if (data[i] < vmin)
      vmin = data[i];
    else if (data[i] > vmax)
      vmax = data[i];
  }

  sprintf(&buffer[0], "%6s\n", "INTRNL");
  sprintf(&buffer[7], "%80s\n", "Untitled");
  sprintf(&buffer[88], "%8d\n", n1);
  sprintf(&buffer[97], "%8d\n", n2);
  sprintf(&buffer[106], "%8d\n", 7); /* REAL*4 */
  sprintf(&buffer[115], "%8d\n", 7); /* REAL*4 */
  sprintf(&buffer[124], "%65f\n", vmin);
  sprintf(&buffer[190], "%65f\n", vmax);
  sprintf(&buffer[256], "%255s\n", " ");

  fwrite(buffer, 1, 512, fp);

  /* Write data */
  fwrite(data, 4, n1*n2, fp);
  return MPI_SUCCESS;
}
