/* $Id: ddlf.c,v 1.5 1995/08/24 15:41:32 tim Exp $ */

/* Copyright 1994 IASC */

/* Fortran wraps for C DDL routines */

#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include <malloc.h>
#include "ddl.h"

/* low-level MPICH-specific stuff !!!! */

/* need fix for 64 SGI --- take from mpich?? 
   This does not work yet... */

#ifdef POINTER_64_BITS
extern void *MPIR_ToPointer();
extern int MPIR_FromPointer();
extern void MPIR_RmPointer();
#else
#define MPIR_ToPointer(a) (a)
#define MPIR_FromPointer(a) (int)(a)
#define MPIR_RmPointer(a)
#endif

/* need trick to handle FORTRAN and C types not being the same
   ie. MPI_DOUBLE is not the same MPI_Datatype as MPI_DOUBLE_PRECISION 

   should we change primitive F types to equivalent C types ? 
   or should we pass the data types through and deal with ALL types
   in the lower-level C routines ? 

   quick fix: change standard Fortran float types to C types */

MPI_Datatype DDL_Ctype(MPI_Datatype ftype)
{
  if (ftype == MPI_DOUBLE_PRECISION) return MPI_DOUBLE;
  else if (ftype == MPI_REAL) return MPI_FLOAT;
  else return ftype;
}

MPI_Datatype DDL_Ftype(MPI_Datatype ctype)
{
  if (ctype == MPI_DOUBLE) return MPI_DOUBLE_PRECISION;
  else if (ctype == MPI_FLOAT) return MPI_REAL;
  else return ctype;
}

/* generic routines */

void F77(ddl_free)(DDL_Object obj, 
		   int *__ierr)
{
  *__ierr = DDL_Free((DDL_Object)*((int*)obj));
}

void F77(ddl_objecttype)(DDL_Object obj, int *type, 
		      int *__ierr)
{
  *__ierr = DDL_Objecttype((DDL_Object)*((int*)obj), type);
}

void F77(ddl_tag)(DDL_Object obj, int *tag, 
		  int *__ierr)
{
  *__ierr = DDL_Tag((DDL_Object)*((int*)obj), tag);
}

void F77(ddl_type)(DDL_Object obj, MPI_Datatype *type, 
		   int *__ierr)
{
  *__ierr = DDL_Type((DDL_Object)*((int*)obj), type);
  *type = DDL_Ftype(*type);
}

void F77(ddl_size)(DDL_Object obj, int *size, 
		   int *__ierr)
{
  *__ierr = DDL_Size((DDL_Object)*((int*)obj), size);
}

void F77(ddl_gsize)(DDL_Object obj, int *size, 
		   int *__ierr)
{
  *__ierr = DDL_Gsize((DDL_Object)*((int*)obj), size);
}

void F77(ddl_block)(DDL_Object obj, int *block, 
		   int *__ierr)
{
  *__ierr = DDL_Block((DDL_Object)*((int*)obj), block);
}

void F77(ddl_offset)(DDL_Object obj, int *offset, 
		     int *__ierr)
{
  *__ierr = DDL_Offset((DDL_Object)*((int*)obj), offset);
  /* remember FORTRAN indices from 1 not 0 */
  *offset = *offset+1;
}

void F77(ddl_format)(DDL_Object obj, int *format, 
		     int *__ierr)
{
  *__ierr = DDL_Format((DDL_Object)*((int*)obj), format);
}

void F77(ddl_comm)(DDL_Object obj, MPI_Comm *comm, 
		   int *__ierr)
{
  *__ierr = DDL_Comm((DDL_Object)*((int*)obj), comm);
}

void F77(ddl_read)(DDL_Object obj, FILE **file,
		   int *__ierr)
{
  *__ierr = DDL_Read((DDL_Object)*((int*)obj), *file);
}

void F77(ddl_write)(DDL_Object obj, FILE **file,
		    int *__ierr)
{
  *__ierr = DDL_Write((DDL_Object)*((int*)obj), *file);
}

void F77(ddl_copy)(DDL_Object src, DDL_Object dest,
		   int *__ierr)
{
  *__ierr = DDL_Copy((DDL_Object)*((int*)src), (DDL_Object)*((int*)dest));
}

void F77(ddl_fill)(DDL_Object obj, void *val,
		   int *__ierr)
{
  MPI_Datatype type;
  DDL_Type((DDL_Object)*((int*)obj), &type);
  if (type == MPI_FLOAT)
    *__ierr = DDL_Fill((DDL_Object)*((int*)obj), (double)(*(float*)val));
  else if (type == MPI_DOUBLE)
    *__ierr = DDL_Fill((DDL_Object)*((int*)obj), *(double*)val);
  else if (type == MPI_INT)
    *__ierr = DDL_Fill((DDL_Object)*((int*)obj), (double)(*(int*)val));
}

void F77(ddl_init)(MPI_Comm oldcomm, MPI_Comm *newcomm,
		   int *__ierr)
{
  *__ierr = DDL_Init((MPI_Comm)MPIR_ToPointer(*(int *)(oldcomm)), newcomm);
}

void F77(ddl_finalize)(MPI_Comm comm,
		       int *__ierr)
{
  *__ierr = DDL_Finalize((MPI_Comm)MPIR_ToPointer(*(int *)(comm)));
}

void F77(ddl_skip_host)(FILE **file,
		   int *__ierr)
{
  *__ierr = DDL_Skip_host(*file);
}

void F77(ddl_skip)(FILE **file, MPI_Comm comm,
		   int *__ierr)
{
  *__ierr = DDL_Skip(*file, (MPI_Comm)MPIR_ToPointer(*(int *)(comm)));
}

/* dense routines */

void F77(ddl_create_vector)(int size[1], MPI_Datatype type,
			    int *block, MPI_Comm comm, DDL_Object *vec,
			    int *__ierr)
{
  *__ierr = DDL_Create_vector(size[0],
	      DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))),
	      *block, (MPI_Comm)MPIR_ToPointer(*(int*)(comm)),
	      vec);
}

void F77(ddl_free_vector)(DDL_Object vec,
			  int *__ierr)
{
  *__ierr = DDL_Free_vector((DDL_Object)*((int*)vec));
}

void F77(ddl_get_vector)(DDL_Object vec, void **ptr,
			 int *__ierr)
{
  *__ierr = DDL_Get_vector((DDL_Object)*((int*)vec), ptr);
}

void F77(ddl_comm_vector)(DDL_Object vec, MPI_Comm *comm,
			  int *__ierr)
{
  *__ierr = DDL_Comm_vector((DDL_Object)*((int*)vec), comm);
}

void F77(ddl_format_vector)(DDL_Object vec, int *format,
			    int *__ierr)
{
  *__ierr = DDL_Format_vector((DDL_Object)*((int*)vec), format);
}

void F77(ddl_type_vector)(DDL_Object vec, MPI_Datatype *type,
			  int *__ierr)
{
  *__ierr = DDL_Type_vector((DDL_Object)*((int*)vec), type);
  *type = DDL_Ftype(*type);
}

void F77(ddl_gsize_vector)(DDL_Object vec, int *size,
			   int *__ierr)
{
  *__ierr = DDL_Gsize_vector((DDL_Object)*((int*)vec), size);
}

void F77(ddl_size_vector)(DDL_Object vec, int *size,
			  int *__ierr)
{
  *__ierr = DDL_Size_vector((DDL_Object)*((int*)vec), size);
}

void F77(ddl_offset_vector)(DDL_Object vec, int *offset,
			    int *__ierr)
{
  *__ierr = DDL_Offset_vector((DDL_Object)*((int*)vec), offset);
  /* remember FORTRAN indices from 1 not 0 */
  *offset = *offset+1;
}

void F77(ddl_read_vector)(DDL_Object vec, FILE **file,
			  int *__ierr)
{
  *__ierr = DDL_Read_vector((DDL_Object)*((int*)vec), *file);
}

void F77(ddl_write_vector)(DDL_Object vec, FILE **file,
			   int *__ierr)
{
  *__ierr = DDL_Write_vector((DDL_Object)*((int*)vec), *file);
}

/* BLAS routines */

void F77(ddl_dot)(DDL_Object x, DDL_Object y, void *dot,
		  int *__ierr)
{
  *__ierr = DDL_Dot((DDL_Object)*((int*)x), (DDL_Object)*((int*)y),
		    dot);
}

void F77(ddl_scal)(void *alpha, DDL_Object x,
		   int *__ierr)
{
  MPI_Datatype x_type;
  DDL_Type_vector((DDL_Object)*((int*)x), &x_type);
  if (x_type == MPI_FLOAT)
    *__ierr = DDL_Scal((double)(*(float*)alpha), (DDL_Object)*((int*)x));
  else if (x_type == MPI_DOUBLE)
    *__ierr = DDL_Scal(*(double*)alpha, (DDL_Object)*((int*)x));
}

void F77(ddl_nrm2)(DDL_Object x, void *nrm2,
		   int *__ierr)
{
  *__ierr = DDL_Nrm2((DDL_Object)*((int*)x), nrm2);
}

void F77(ddl_axpy)(void *alpha, DDL_Object x, DDL_Object y,
		   int *__ierr)
{
  int i, size;
  double *data;
  MPI_Datatype x_type;
  DDL_Type_vector((DDL_Object)*((int*)x), &x_type);
  if (x_type == MPI_FLOAT)
    *__ierr = DDL_Axpy((double)(*(float*)alpha), (DDL_Object)*((int*)x),
		       (DDL_Object)*((int*)y));
  else if (x_type == MPI_DOUBLE) {
    *__ierr = DDL_Axpy(*(double*)alpha, (DDL_Object)*((int*)x),
		       (DDL_Object)*((int*)y));
  }
}

void F77(ddl_rand)(void *alpha, DDL_Object x,
		   int *__ierr)
{
  MPI_Datatype x_type;
  DDL_Type_vector((DDL_Object)*((int*)x), &x_type);
  if (x_type == MPI_FLOAT)
    *__ierr = DDL_Rand((double)(*(float*)alpha), (DDL_Object)*((int*)x));
  else if (x_type == MPI_DOUBLE)
    *__ierr = DDL_Rand(*(double*)alpha, (DDL_Object)*((int*)x));
}

/* sparse routines */

void F77(ddl_create_sparse)(int size[3], MPI_Datatype type,
			    int *format, int *block, 
			    MPI_Comm comm, DDL_Object *sp,
			    int *__ierr)
{
  *__ierr = DDL_Create_sparse(size,
                DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))),
		*format, *block, 
		(MPI_Comm)MPIR_ToPointer(*(int*)(comm)),
		sp);
}

void F77(ddl_free_sparse)(DDL_Object sp,
			  int *__ierr)
{
  *__ierr = DDL_Free_sparse((DDL_Object)*((int*)sp));
}

void F77(ddl_get_ia)(DDL_Object sp, int **ia,
		     int *__ierr)
{
  *__ierr = DDL_Get_ia((DDL_Object)*((int*)sp), ia);
}

void F77(ddl_get_ja)(DDL_Object sp, int **ja,
		     int *__ierr)
{
  *__ierr = DDL_Get_ja((DDL_Object)*((int*)sp), ja);
}

void F77(ddl_get_a)(DDL_Object sp, void **a,
		    int *__ierr)
{
  *__ierr = DDL_Get_a((DDL_Object)*((int*)sp), a);
}

void F77(ddl_type_sparse)(DDL_Object sp, MPI_Datatype *type,
			  int *__ierr)
{
  *__ierr = DDL_Type_sparse((DDL_Object)*((int*)sp), type);
  *type = DDL_Ftype(*type);
}

void F77(ddl_size_sparse)(DDL_Object sp, int *size,
			  int *__ierr)
{
  *__ierr = DDL_Size_sparse((DDL_Object)*((int*)sp), size);
}

void F77(ddl_offset_sparse)(DDL_Object sp, int *offset,
			    int *__ierr)
{
  *__ierr = DDL_Offset_sparse((DDL_Object)*((int*)sp), offset);
  /* remember FORTRAN indices from 1 not 0 */
  *offset = *offset+1;
}

void F77(ddl_format_sparse)(DDL_Object sp, int *format,
			    int *__ierr)
{
  *__ierr = DDL_Format_sparse((DDL_Object)*((int*)sp), format);
}

void F77(ddl_setformat_sparse)(DDL_Object sp, int *format,
			       int *__ierr)
{
  *__ierr = DDL_Setformat_sparse((DDL_Object)*((int*)sp), *format);
}

void F77(ddl_comm_sparse)(DDL_Object sp, MPI_Comm *comm,
			  int *__ierr)
{
  *__ierr = DDL_Comm_sparse((DDL_Object)*((int*)sp), comm);
}

void F77(ddl_used_sparse)(DDL_Object sp, int *nelt,
			  int *__ierr)
{
  *__ierr = DDL_Used_sparse((DDL_Object)*((int*)sp), nelt);
}

void F77(ddl_gused_sparse)(DDL_Object sp, int *nelt,
			   int *__ierr)
{
  *__ierr = DDL_Gused_sparse((DDL_Object)*((int*)sp), nelt);
}

void F77(ddl_setused_sparse)(DDL_Object sp, int *nelt,
			     int *__ierr)
{
  *__ierr = DDL_Setused_sparse((DDL_Object)*((int*)sp), *nelt);
}

void F77(ddl_gsize_sparse)(DDL_Object sp, int *size,
			   int *__ierr)
{
  *__ierr = DDL_Gsize_sparse((DDL_Object)*((int*)sp), size);
}

void F77(ddl_init_sparse)(MPI_Comm oldcomm, MPI_Comm *newcomm,
			  int *__ierr)
{
  *__ierr = DDL_Init_sparse((MPI_Comm)MPIR_ToPointer(*(int*)(oldcomm)),
			    newcomm);
}

void F77(ddl_finalize_sparse)(MPI_Comm comm,
			      int *__ierr)
{
  *__ierr = DDL_Finalize_sparse((MPI_Comm)MPIR_ToPointer(*(int*)(comm)));
}

void F77(ddl_info_sparse)(DDL_Object sp,
			  int *__ierr)
{
  *__ierr = DDL_Info_sparse((DDL_Object)*((int*)sp));
}

void F77(ddl_read_sparse)(DDL_Object sp, FILE **file,
			  int *__ierr)
{
  *__ierr = DDL_Read_sparse((DDL_Object)*((int*)sp), *file);
}

void F77(ddl_write_sparse)(DDL_Object sp, FILE **file,
			   int *__ierr)
{
  *__ierr = DDL_Write_sparse((DDL_Object)*((int*)sp), *file);
}

/* support routines */

void F77(ddl_inhb)(FILE **file, int *n, int *nelt, int ia[], int ja[], double a[],
		   int *isym, double soln[], double rhs[], double xinit[], int *job,
		   int *__ierr)
{
  *__ierr = DDL_Inhb(*file, n, nelt, ia, ja, a,
		     isym, soln, rhs, xinit, job);
}

void F77(ddl_out_sparse)(FILE **file, int *n, int *nelt, 
			 int ia[], int ja[], double a[], MPI_Datatype type,
			 int *isym, int *format,
			 int *__ierr)
{
  *__ierr = DDL_Out_sparse(*file, *n, *nelt, 
	      ia, ja, a, 
	      DDL_Ctype((MPI_Datatype)MPIR_ToPointer( *(int*)(type) )),
	      *isym, *format);
}

void F77(ddl_in_sparse)(FILE **file, int *n, int *nelt, 
			int ia[], int ja[], double a[], MPI_Datatype *type,
			int *isym, int *format,
			int *__ierr)

{
  *__ierr = DDL_In_sparse(*file, n, nelt, 
			  ia, ja, a, type, isym, format);
  *type = DDL_Ftype(*type);
}

void F77(ddl_out_vector)(FILE **file, void *vec, MPI_Datatype type, int *n,
			 int *__ierr)
{
  *__ierr = DDL_Out_vector(*file, vec, 
         	   DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))), n);
}

void F77(ddl_in_vector)(FILE **file, void *vec, MPI_Datatype *type, int *n,
			int *__ierr)
{
  *__ierr = DDL_In_vector(*file, vec, type, n);
  *type = DDL_Ftype(*type);
}

void F77(ddl_open_host)(FILE **file, char *fname, int *mode,
			int *__ierr)
{
  *__ierr = DDL_Open_host(file, fname, *mode);
}

void F77(ddl_open)(FILE **file, char *fname, int *mode, MPI_Comm comm,
		   int *__ierr)
{
  *__ierr = DDL_Open(file, fname, *mode,
		     (MPI_Comm)MPIR_ToPointer(*(int*)(comm)));
}

void F77(ddl_close_host)(FILE **file,
			 int *__ierr)
{
  *__ierr = DDL_Close_host(*file);
}

void F77(ddl_close)(FILE **file, MPI_Comm comm,
		    int *__ierr)
{
  *__ierr = DDL_Close(*file, (MPI_Comm)MPIR_ToPointer(*(int*)(comm)));
}

void F77(ddl_fileformat_host)(FILE **file, int *format, MPI_Datatype *type, 
			      int a[],
			      int *__ierr)
{
  *__ierr = DDL_Fileformat_host(*file, format, type, a);
  *type = DDL_Ftype(*type);
}

void F77(ddl_fileformat)(FILE **file, int *format, MPI_Datatype *type, 
			 int a[], MPI_Comm comm,
			 int *__ierr)
{
  *__ierr = DDL_Fileformat(*file, format, type, a, 
			   (MPI_Comm)MPIR_ToPointer(*(int*)(comm)));
  *type = DDL_Ftype(*type);
}

void F77(ddl_coltotriad)(int *n, int *nelt, 
			 int ia[], int ja[], void *a, MPI_Datatype type, int *isym,
			 int *__ierr)
{
  *__ierr = DDL_Coltotriad(*n, *nelt, 
	       ia, ja, a, 
	       DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))), *isym);
}

void F77(ddl_rowtotriad)(int *n, int *nelt, 
			 int ia[], int ja[], void *a, MPI_Datatype type, int *isym,
			 int *__ierr)
{
  *__ierr = DDL_Rowtotriad(*n, *nelt, 
	       ia, ja, a, 
               DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))), *isym);
}

void F77(ddl_triadtocol)(int *n, int *nelt, 
			 int ia[], int ja[], void *a, MPI_Datatype type, int *isym,
			 int *__ierr)
{
  *__ierr = DDL_Triadtocol(*n, *nelt, 
	       ia, ja, a, 
	       DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))), *isym);
}

void F77(ddl_triadtorow)(int *n, int *nelt, 
			 int ia[], int ja[], void *a, MPI_Datatype type, int *isym,
			 int *__ierr)
{
  *__ierr = DDL_Triadtorow(*n, *nelt, 
  	       ia, ja, a, 
	       DDL_Ctype((MPI_Datatype)MPIR_ToPointer(*(int*)(type))), *isym);
}

void F77(ddl_mpitype)(int *DDLtype, MPI_Datatype *MPItype,
		      int *__ierr)
{
  *__ierr = DDL_MPItype(*DDLtype, MPItype);
  *MPItype = DDL_Ftype(*MPItype);	 
}

void F77(ddl_ddltype)(MPI_Datatype *MPItype, int *DDLtype,
		      int *__ierr)
{
  *__ierr = DDL_DDLtype(DDL_Ctype((MPI_Datatype)*((int*)MPItype)), DDLtype);
}

/* new sparse routines from sphigh.c */

void F77(ddl_smv_comm)(DDL_Object sp,
		       int *__ierr)
{
  *__ierr = DDL_Smv_comm((DDL_Object)*((int*)sp));
}

void F77(ddl_smv)(DDL_Object sp, DDL_Object x, DDL_Object y,
		  double *alpha, double *beta,
		  int *__ierr)
{
  *__ierr = DDL_Smv((DDL_Object)*((int*)sp), (DDL_Object)*((int*)x),
		    (DDL_Object)*((int*)y), *alpha, *beta);
}

void F77(ddl_compare_vecvec)(DDL_Object x, DDL_Object y, int *ret,
			     int *__ierr)
{
  *__ierr = DDL_Compare_vecvec((DDL_Object)*((int*)x), (DDL_Object)*((int*)y),
			       ret);
}

void F77(ddl_compare_matvec)(DDL_Object x, DDL_Object y, int *ret,
			     int *__ierr)
{
  *__ierr = DDL_Compare_matvec((DDL_Object)*((int*)x), (DDL_Object)*((int*)y),
			       ret);
}

void F77(ddl_compare_vecperm)(DDL_Object x, DDL_Object y, int *ret,
			     int *__ierr)
{
  *__ierr = DDL_Compare_vecperm((DDL_Object)*((int*)x), (DDL_Object)*((int*)y),
			       ret);
}

void F77(ddl_compare_matpermx)(DDL_Object x, DDL_Object y, int *ret,
			       int *__ierr)
{
  *__ierr = DDL_Compare_matpermx((DDL_Object)*((int*)x), (DDL_Object)*((int*)y),
			       ret);
}

void F77(ddl_compare_matpermy)(DDL_Object x, DDL_Object y, int *ret,
			       int *__ierr)
{
  *__ierr = DDL_Compare_matpermy((DDL_Object)*((int*)x), (DDL_Object)*((int*)y),
			       ret);
}

void F77(ddl_setsubtype_sparse)(DDL_Object x, int *subtype,
			       int *__ierr)
{
  *__ierr = DDL_Setsubtype_sparse((DDL_Object)*((int*)x), *subtype);
}

void F77(ddl_subtype_sparse)(DDL_Object x, int *subtype,
			       int *__ierr)
{
  *__ierr = DDL_Subtype_sparse((DDL_Object)*((int*)x), subtype);
}

void F77(ddl_check_perm)(DDL_Object x, int *ret,
			     int *__ierr)
{
  *__ierr = DDL_Check_perm((DDL_Object)*((int*)x), ret);
}

void F77(ddl_setperm_sparse)(DDL_Object a, DDL_Object x, DDL_Object y, 
			     int *__ierr)
{
  *__ierr = DDL_Setperm_sparse((DDL_Object)*((int*)a), (DDL_Object)*((int*)x),
			       (DDL_Object)*((int*)y));
}

void F77(ddl_invperm)(DDL_Object x, DDL_Object y, 
		      int *__ierr)
{
  *__ierr = DDL_Invperm((DDL_Object)*((int*)x),
			(DDL_Object)*((int*)y));
}

void F77(ddl_trisolve_comm)(DDL_Object sp,
			    int *__ierr)
{
  *__ierr = DDL_Trisolve_comm((DDL_Object)*((int*)sp));
}

void F77(ddl_trisolve)(DDL_Object sp, DDL_Object x, DDL_Object b,
		       int *__ierr)
{
  *__ierr = DDL_Trisolve((DDL_Object)*((int*)sp), (DDL_Object)*((int*)x),
		    (DDL_Object)*((int*)b));
}

/* other utility routines */

#include <stdarg.h>

#define DDL_MAX_PARAM 40

/* DDL access procedure
   provides FORTRAN programs with access to raw data vectors
   If more parameters are needed, increase DDL_MAX_PARAM and
   add more lines to call to func.

   Parameters:
     param 1:       FORTRAN function to call
     param 2:       n, count of vectors which follow immediately
     param n+3:     m, count of scalars which follow immediately
     param n+m+4:   ierr last parameter
   Status: H2 */
void F77(ddl_access)(void (*func)(), ...)
{
  va_list ap;
  void *p[DDL_MAX_PARAM];
  void **pdata, *pscal;
  int i;
  int *__ierr;
  int *num, cnt;

  va_start(ap, func);

  /* scan in pointers to raw data */
  num = va_arg(ap, int *);
  for (i=0;i<*num;i++) {
    pdata = va_arg(ap, void **);
    p[i] = *pdata;
  }

  cnt = *num;
  /* scan in pointers to scalars */
  num = va_arg(ap, int *);
  for (i=0;i<*num;i++) {
    pscal = va_arg(ap, void *);
    p[cnt+i] = pscal;
  }
  cnt = cnt+*num;

  p[cnt] = va_arg(ap, int *);
  cnt = cnt+1;
  va_end(ap);
  /* call user FORTRAN procedure */
  (*func)(p[0],
	  p[1],
	  p[2],
	  p[3],
	  p[4],
	  p[5],
	  p[6],
	  p[7],
	  p[8],
	  p[9],
	  p[10],
	  p[11],
	  p[12],
	  p[13],
	  p[14],
	  p[15],
	  p[16],
	  p[17],
	  p[18],
	  p[19],
	  p[20],
	  p[21],
	  p[22],
	  p[23],
	  p[24],
	  p[25],
	  p[26],
	  p[27],
	  p[28],
	  p[29],
	  p[30],
	  p[31],
	  p[32],
	  p[33],
	  p[34],
	  p[35],
	  p[36],
	  p[37],
	  p[38],
	  p[39]
    );
  return;
}

/* routine to check for 8byte aligned pointer
   for use when getting double precision workspace pointers
   from a users workspace declared as type integer (4byte aligned)
   Parameters:
     ptr: IN        pointer to check
     align: OUT     whether pointer is aligned or not:
                      0: aligned
		      1-3: not aligned
   Status: H3 */
void F77(ddl_isaligndbl)(int *ptr, int *align)
{
  *align = ((int)ptr & 03);
}

/* set CPU run time limit
   Parameters:
     sec: IN        maximum processor time allowed in seconds
   Status: H3 */
void F77(ddl_set_cpu_limit)(int *sec)
{
  DDL_Set_cpu_limit(*sec);
}

/* find time 
   take the difference between two calls to ddl_gettime to
   get time taken in seconds */
double F77(ddl_gettime)()
{
  return DDL_Gettime();
}

void F77(ddl_ilu_comm)(DDL_Object sp,
                            int *__ierr)
{
  *__ierr = DDL_ILU_comm((DDL_Object)*((int*)sp));
}

void F77(ddl_ilu_form)(DDL_Object sp, DDL_Object l, DDL_Object u, 
                DDL_Object diag_blk, int *__ierr)
{
  *__ierr = DDL_ILU_form((DDL_Object)*((int*)sp),
          (DDL_Object)*((int*)l), (DDL_Object)*((int*)u),
          (DDL_Object)*((int*)diag_blk));
}


