c     $Id: tfqmrnc2.f,v 1.2 1994/12/14 13:28:31 tim Exp tim $

c     MPI sparse DDL demo in FORTRAN (Andy's tfqmrnc)

c     Copyright 1994 IASC

c     DESCRIPTION   DDL based iterative solver for Ax=b.
c                   Block preconditioning techniques are  
c                   incorporated into this version.

      program tfqmrnc

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character hbfilename*(100), ddlfilename*(100)
      character solnfilename*(100), summaryfilename*(100)
      data hbfilename /'test4001.hbf\0'/
      data ddlfilename /'tfqmrdata\0'/
      data solnfilename /'soln\0'/
      data summaryfilename /'summary\0'/

c     Workspace for ilu blocks & solves 9/1/95
      integer maxiwork, maxrwork
      parameter(maxiwork=1000000, maxrwork=1000000) ! local size
      integer iwork(maxiwork)
      integer rwork(maxrwork)
      
      integer i,n,iter,itmax,loop
      double precision error,tol(10),gamma,delta,epsilon
      integer nelt, tolno, nc
      logical rhsexist,ilu,ilunc,expinv
      logical nopc
      character*80 ans
c      character*3 reply

      integer infile, outfile
      integer ierr
      integer block, format
      integer comm
      integer a_ddl, x_ddl, b_ddl, xtrue_ddl, tmp_ddl
      integer type, fformat, sizes(3), len, precon
      integer tot_blks

c     initialise MPI, DDL
      call MPI_Init(ierr)
      call DDL_Init_sparse(MPI_COMM_WORLD, comm, ierr)

      if (DDL_Ishost()) then 
        print *,' '
        print *,'                TFQMR SPARSE DDL SOLVER'
        print *,'               *************************'

c********************************************************************
c*                                                                  *
c*                   READ ITERATION PARAMETERS 
c*                                                                  *
c********************************************************************

        write(*,*) 'Input demonstration options'
        write(*,*) '(for default [in brackets] just press RETURN)'
        write(*,*) ''

        write(*,*) 'Input name of data file [test4001.hbf]'
        read(*,'(a)') ans
        if (ans .eq. '') then
           hbfilename = 'test4001.hbf'
        else
           hbfilename = ans
        end if
c       form C file name string for DDL procedures
        len=index(hbfilename,' ')
        hbfilename(len:len) = '\0'

        print *,'Is the problem supplied with right hand'
        print *,'side and solution vectors y/n? [Y]'
        read(*,'(a)') ans
        if (ans .eq. '') then
           ans = 'y'
        end if
        if ((ans.eq.'Y').or.(ans.eq.'y')) then
           rhsexist = .TRUE.
        else
           rhsexist = .FALSE.
        end if
        print *,''
        
        print *,'Enter Max. No. Iterations [100]'
        read(*,'(a)') ans
        if (ans .eq. '') then
           itmax = 100
        else
           read (ans, *) itmax
        end if

        print *,'Enter no. of tolerance levels required (max 10) [1]'
        read(*,'(a)') ans
        if (ans .eq. '') then
           tolno = 1
        else
           read (ans, *) tolno
        end if
        print *,''
        
        print *,'Enter set of tolerances (smallest last)'
        
        do i = 1,tolno
           print *,'Enter tolerance ',i,' [1e-6]'
           read(*,'(a)') ans
           if (ans .eq. '') then
              tol(i) = 1e-6
           else
              read (ans, *) tol(i)
           end if
        end do
        print *,''
        
        print *,'Enter e, where x_initial = x_true + e(rand(0,1))'
        print *,'(typically e should lie between 0.001 and 1.0):'
        print *,'An entry of zero gives an initial guess of zero [0]'
        read(*,'(a)') ans
        if (ans .eq. '') then
           epsilon = 0.0D0
        else
           read (ans, *) epsilon
        end if

        print *,'Is block preconditioning to be applied [Y]'   
        read(*,'(a)') ans
        if (ans .eq. '') then
           ans = 'y'
        end if
        if ((ans.eq.'Y').or.(ans.eq.'y')) then
 10        print *,'Choose the preconditioner'
           print *,'type "ilu" for incomplete LU factorization'
           print *,'     "inc" for incomplete LUnc factorization [INC]'
           read(*,'(a)') ans
           if (ans .eq. '') then
              ans = 'inc'
           end if
           if ((ans.eq.'ILU').or.(ans.eq.'ilu')) then
              ilu = .TRUE.
              print *,''
              print *,'ILU block preconditioner to be applied'
              print *,''
              ilunc = .FALSE.
              nopc = .FALSE.
           else if ((ans.eq.'INC').or.(ans.eq.'inc')) then
              ilunc = .TRUE.
              print *,''
              print *,'ILU(nc) block preconditioner to be applied'
              print *,''
              print *,'Enter nc parameter [4]'
              read(*,'(a)') ans
              if (ans .eq. '') then
                 nc = 4
              else
                 read (ans, *) nc
              end if
              print *,'Do you want to form explicit cell diagonals? [N]'
              read(*,'(a)') ans
              if (ans .eq. '') then
                 ans = 'n'
              end if
              if ((ans.eq.'Y').or.(ans.eq.'y')) then
                 expinv = .TRUE.
              else
                 expinv = .FALSE.
              end if
              ilu = .FALSE.
              nopc = .FALSE.
              print *,'Enter no. blocks in the preconditioner? [10]'
              read(*,'(a)') ans
              if (ans .eq. '') then
                 tot_blks = 10
              else
                 read (ans, *) tot_blks  
              end if
              print *,''
           else 
              print *,''
              print *,'*Unrecognizeable input* - try again y/n'
              print *,''
              read (*,'(a)') ans
              if ((ans.eq.'Y').or.(ans.eq.'y')) then 
                 go to 10
              else
                 print *,''
                 print *,'No Preconditioning applied'
                 print *,''
                 nopc = .TRUE.
                 ilu = .FALSE.
                 ilunc = .FALSE.
              end if
           end if
        else
           print *,''
           print *,'No Preconditioning applied'
           print *,''
           nopc = .TRUE.
           ilu = .FALSE.
           ilunc = .FALSE.
        end if

        print *, 'OK. Starting run...'

        call createinput(hbfilename, ddlfilename, rhsexist)
        
      end if ! DDL_Ishost

c     broadcast user input to all processes
      call MPI_Bcast(itmax, 1, MPI_INTEGER, 0, comm, ierr)
      call MPI_Bcast(rhsexist, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(ilu, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(ilunc, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(nc, 1, MPI_INTEGER, 0, comm, ierr)
      call MPI_Bcast(expinv, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(nopc, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(epsilon, 1, MPI_DOUBLE_PRECISION, 0, comm, ierr)
      call MPI_Bcast(tolno, 1, MPI_INTEGER, 0, comm, ierr)
      do i = 1, tolno
         call MPI_Bcast(tol(i), 1, MPI_DOUBLE_PRECISION, 0, comm, ierr)
      end do
c      call MPI_Bcast(tot_blks, 1, MPI_INTEGER, 0, comm, ierr)

c     open input and output files
      call DDL_Open(infile, ddlfilename, DDL_READ_MODE, comm, ierr)
      call DDL_Open(outfile, solnfilename, DDL_WRITE_MODE, comm, ierr)

c     read in the information heading
      call DDL_Fileformat(infile, fformat, type, sizes, comm, ierr)
      n = sizes(1)
      nelt = sizes(2)
      format = DDL_PCK_ROW
      block = tot_blks

c     set up size vector
      sizes(1) = n
      sizes(2) = n
      sizes(3) = nelt*2

c     create sparse matrix
      a_ddl = DDL_NULL
      call DDL_Create_sparse(sizes, type, format, 
     &                       block, comm, a_ddl, ierr)

c     create vector for solution
      x_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, x_ddl, ierr)
c     create vector for result
      b_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, b_ddl, ierr)
c     create vector for true solution
      xtrue_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, xtrue_ddl, ierr)
c     create temp vector
      tmp_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, tmp_ddl, ierr)

c     read data from file to sparse object
      call DDL_Read(a_ddl, infile, ierr)

      if (.not. rhsexist) then
c        initialise xtrue to one
         call DDL_Fill(xtrue_ddl, 1.0D0, ierr)
         call DDL_Smv(a_ddl, xtrue_ddl, b_ddl, gamma, delta, ierr)
      else
c        read rhs from file
         call DDL_Read(b_ddl, infile, ierr)
c        read soln vector from file
         call DDL_Read(xtrue_ddl, infile, ierr)
      endif

c     put in loop to fool SGI
c      do loop=1,100

c     initialise x 
      call DDL_Rand(1.0d0, tmp_ddl, ierr)
      call DDL_Copy(xtrue_ddl, x_ddl, ierr)
      if (epsilon.eq.0.0d0) then
         call DDL_Scal(0.0d0, x_ddl, ierr)
      else 
         call DDL_Axpy(epsilon, tmp_ddl, x_ddl, ierr)
      end if

      if (nopc) precon = 0
      if (ilu) precon = 1
      if (ilunc) precon = 2

      call DDL_Linsolve(a_ddl, x_ddl, b_ddl, itmax, iter, 
     &     tol, tolno, precon, nc, expinv, error, 
     &     iwork, maxiwork, rwork, maxrwork, ierr)

c     Why have we stopped?
      if (DDL_Ishost()) then
         if(iter .lt. itmax) THEN
            write (*,*)
            write (*,'(A22,I4,A11)') 'TFQMR converged after',iter,
     +           'iterations'
            write (*,*)
         else 
            write (*,*) 
            write (*,'(A37)') ' Specified Max. Iterations reached:' 
            write (*,'(A15,I4,A25,I4)') ' Max.iters = ',itmax,
     +           'Iterations reached = ',iter
            write (*,*)
         end if
      end if

c     end loop
c      end do

      call DDL_Write(x_ddl, outfile, ierr)

      call DDL_Close(outfile, comm, ierr)
      call DDL_Close(infile, comm, ierr)

c     check solution is correct
      if (DDL_Ishost()) then
         call check(ddlfilename, solnfilename, rhsexist, error)
         print *, 'Max error is', error
      end if

c     delete distributed objects
      call DDL_Free(a_ddl, ierr)
      call DDL_Free(x_ddl, ierr)
      call DDL_Free(b_ddl, ierr)
      call DDL_Free(xtrue_ddl, ierr)
      call DDL_Free(tmp_ddl, ierr)

c     tidy up and exit
      call DDL_Finalize_sparse(comm, ierr)
      call MPI_Finalize(ierr)

      stop
      end


c     convert Harwell-Boeing file to row-packed triad file
c     HB file has matrix; and  rhs, soln if .rhsexist.
c     output matrix; rhs, soln if .rhsexist.
      subroutine createinput(hbfilename, ddlfilename, rhsexist)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character*(*) hbfilename, ddlfilename
      logical rhsexist

      integer maxn, mxnelt, np
      parameter (maxn = 5005, mxnelt = 40000, np = 10)

      integer infile, outfile
      integer ierr, format
      integer type
      integer ia(mxnelt), ja(mxnelt)
      double precision a(mxnelt)
      integer nelt, job
      integer n, i, isym, size
      double precision soln(maxn), rhs(maxn), xinit(maxn)
      
c     read in sparse matrix from file
      job = 3
      n = maxn
      nelt = mxnelt
c     read in as column gathered
      call DDL_Open_host(infile, hbfilename, DDL_READ_MODE, ierr)
      call DDL_Inhb(infile, n, nelt, ia, ja, a, 
     &     isym, soln, rhs, xinit, job, ierr)
      call DDL_Close_host(infile, ierr)
      type = MPI_DOUBLE_PRECISION
c     convert to triad 
      call DDL_Coltotriad(n, nelt, ia, ja, a, type, isym, ierr)
c     convert triad to row
      call DDL_Triadtorow(n, nelt, ia, ja, a, type, isym, ierr)
c     convert row to triad
      call DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym, ierr)
c     write out as row ordered triad format
c     this data file will be used for rest of demo
      format = DDL_TRIAD_ROW
      call DDL_Open_host(outfile, ddlfilename, DDL_WRITE_MODE, ierr)
      call DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
     &     isym, format, ierr)

      if (rhsexist) then
        if (job .eq. 23 .or. job .eq. 31) then
c         write out rhs & sol
          call DDL_Out_vector(outfile, rhs, MPI_DOUBLE_PRECISION, 
     &                        n, ierr)
          call DDL_Out_vector(outfile, soln, MPI_DOUBLE_PRECISION, 
     &                        n, ierr)
        else
          print *, 'Error reading RHS and SOLN vectors'
        end if
      end if
      
      call DDL_Close_host(outfile, ierr)
      return
      end

c     check result vector
      subroutine check(ddlfilename, solnfilename, rhsexist, maxerr)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character*(*) solnfilename, ddlfilename
      logical rhsexist
      double precision maxerr

      integer maxn, mxnelt, np
      parameter (maxn = 5005, mxnelt = 40000, np = 10)

      integer infile, outfile
      integer ierr, format
      integer type
      integer n, i
      double precision soln(maxn), correct(maxn), error
      
c     get calculated solution vector
      call DDL_Open_host(infile, solnfilename, DDL_READ_MODE, ierr) 
      n = maxn
      call DDL_In_vector(infile, soln, type, n, ierr)  
      call DDL_Close_host(infile, ierr)
c     get correct solution vector
      if (rhsexist) then
         call DDL_Open_host(infile, ddlfilename, DDL_READ_MODE, ierr) 
         call DDL_Skip_host(infile, ierr)
         call DDL_Skip_host(infile, ierr)
         n = maxn
         call DDL_In_vector(infile, correct, type, n, ierr)  
         call DDL_Close_host(infile, ierr)
      else
c        set to 1
         do i=1, n
            soln(i) = 1.0D0
         end do
      end if

c     check for differences
      maxerr = 0.0D0
      do i=1, n
         error = ABS(ABS(soln(i))-ABS(correct(i)))
         if (error .gt. maxerr) then
	   maxerr = error
         end if
      end do

      return
      end


c     DDL procedure to solve sparse linear system of equations 
      subroutine DDL_Linsolve(a_ddl, x_ddl, b_ddl, itmax, iter, 
     &     tol, tolno, precon, nc, expinv, error, 
     &     iwork, iwork_dim, rwork, rwork_dim, ierr)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

c     DDL object handles
      integer a_ddl, x_ddl, b_ddl
      
      integer itmax, iter, tolno, precon, nc, ierr
      double precision tol(tolno), error
      logical expinv
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)

c     local variables
c     tfqmrnc variables
      integer i,n,m
      double precision alpha,theta,betan,
     + tau,omega,rhon,rhonm1,const,xnrm,c,sigma,malpha,oldcgs,
     + rnrm,rnrm0,eta,xtnrm,bnrm,gamma,delta
      integer nelt, tolcount
      logical ilu,ilunc

c     DDL/MPI 
      integer infile, outfile
      integer block, format
      integer comm
      integer m_ddl, l_ddl, u_ddl, dinv_ddl
      integer pip_ddl, rcgs_ddl, r0_ddl
      integer p_ddl, v_ddl, d_ddl, tmp_ddl, q_ddl, w_ddl, z_ddl
      integer uv_ddl, rvec_ddl, ivec_ddl
      integer type, fformat, sizes(3), len
      integer blocksize, lnelt, memory_int, memory_real
      integer ncblks
c     check preconditioning type
      if (precon .eq. 0) then
         ilu = .FALSE.
         ilunc = .FALSE.
      else if (precon .eq. 1) then
         ilu = .TRUE.
         ilunc = .FALSE.
      else if (precon .eq. 2) then
         ilu = .FALSE.
         ilunc = .TRUE.
      end if

c     find problem size
      call DDL_Gsize(a_ddl, sizes, ierr) 
      n = sizes(1)
      nelt = sizes(3)
      call DDL_Type(a_ddl, type, ierr)
      call DDL_Format(a_ddl, format, ierr)
      call DDL_Comm(a_ddl, comm, ierr)
      block = 1

c     check enough memory is allocated
      if (ilunc) then
         call DDL_Size(a_ddl, sizes, ierr)
         blocksize = sizes(1)
         lnelt = sizes(3)
         ncblks = blocksize/nc
         memory_int = 10+nc+2*(ncblks+blocksize)+ lnelt+blocksize+1
         memory_real = 8*blocksize+nc*(blocksize+3*nc)+ lnelt
         if (memory_int.gt.iwork_dim.or.memory_real.gt.rwork_dim) then
            ierr=2
            goto 1000
         end if
      end if

      if (ilu) then
         sizes(1) = n
         sizes(2) = n
         sizes(3) = nelt*2
         l_ddl = DDL_NULL
         call DDL_Create_sparse(sizes, type, format, 
     &                          block, comm, l_ddl, ierr)
         u_ddl = DDL_NULL
         call DDL_Create_sparse(sizes, type, format, 
     &                          block, comm, u_ddl, ierr)
      end if

      if (ilu) then
         dinv_ddl = DDL_NULL
         call DDL_Create_vector(n, type, block, comm, dinv_ddl, ierr)
c        factorize A to block preconditioner M
         call DDL_Ilu_block(a_ddl, l_ddl, u_ddl, dinv_ddl,
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
      end if

      if (ilunc) then
c        create integer vector for ilunc calls
         ivec_ddl = DDL_NULL
         call DDL_Create_vector(10*n, MPI_INTEGER, block, 
     &                          comm, ivec_ddl, ierr)
         rvec_ddl = DDL_NULL
         call DDL_Create_vector(10*n, MPI_DOUBLE_PRECISION, block, 
     &                          comm, rvec_ddl, ierr)
      write(*,*)
      write(*,*)
      write(*,*) 'About to form preconditioner'
      write(*,*)
c        factorize A to block preconditioner ILUnc
         call DDL_Ilunc_block(a_ddl, rvec_ddl, ivec_ddl, nc, expinv, 
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
      end if
      write(*,*) 'Formed preconditioner'
      write(*,*)
c     create vector for residual
      rcgs_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, rcgs_ddl, ierr)
c     create vector for initial residual
      r0_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, r0_ddl, ierr)
c     create vectors for tfqmr soln method
      uv_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, uv_ddl, ierr)
      p_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, p_ddl, ierr)
      v_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, v_ddl, ierr)
      d_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, d_ddl, ierr)
      tmp_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, tmp_ddl, ierr)
      q_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, q_ddl, ierr)
      w_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, w_ddl, ierr)
      if(ilu .or. ilunc) then
         z_ddl = DDL_NULL
         call DDL_Create_vector(n, type, block, 
     &                          comm, z_ddl, ierr)
      end if

      gamma = 1.0d0
      delta = 0.0d0
c     calculate mv communications
      call DDL_Smv_comm(a_ddl, ierr)

      call DDL_Nrm2(b_ddl, bnrm, ierr)

      call DDL_Copy(b_ddl, r0_ddl, ierr)
      call DDL_Smv(a_ddl, x_ddl, r0_ddl, -1.0d0, gamma, ierr)     

      call DDL_Nrm2(x_ddl, xnrm, ierr)

      if (ilu) then
         call DDL_Copy(r0_ddl, tmp_ddl, ierr)
         call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl, 
     &                      r0_ddl, ierr)
         call DDL_Copy(b_ddl, tmp_ddl, ierr)
         call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl, 
     &                      b_ddl, ierr)
      else if (ilunc) then
         call DDL_Copy(r0_ddl, tmp_ddl, ierr)
         call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &        r0_ddl, nc, expinv,
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
         call DDL_Copy(b_ddl, tmp_ddl, ierr)
         write(*,*)
         write(*,*) 'About to call DDL_Ilunc_solve'
         call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &        b_ddl, nc, expinv, 
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
      end if
         write(*,*)
         write(*,*) 'Solve Completed OK'     

      call DDL_Copy(r0_ddl, p_ddl, ierr)
      call DDL_Copy(r0_ddl, uv_ddl, ierr)
      call DDL_Copy(r0_ddl, rcgs_ddl, ierr)
      call DDL_Scal(0.0d0, d_ddl, ierr)

c     initialise v to Ap
      call DDL_Smv(a_ddl, p_ddl, v_ddl, gamma, delta, ierr)

      call DDL_Nrm2(x_ddl, xnrm, ierr)
      call DDL_Nrm2(rcgs_ddl, rnrm, ierr)
      call DDL_Nrm2(b_ddl, bnrm, ierr)

      rnrm0 = rnrm
      omega = rnrm
      tau = rnrm
      theta = 0.0d0
      eta = 0.0d0
      iter = 0
      tolcount = 1
      error = rnrm/bnrm
      rhonm1 = 1.0d0

      call DDL_Dot(r0_ddl, rcgs_ddl, rhonm1, ierr)

C*****************************************************************
C*****************************************************************
C
C                  START ITERATIVE LOOP
C
C*****************************************************************
C*****************************************************************
  

      do while ((error.GE.(tol(tolno))).AND.(iter.LT.itmax)) 
        iter = iter + 1
        print *, ''
        print *,' iter = ',iter
        
C*****************************************************************
C
C                      SECTION (c)
C
C*****************************************************************
        
         call DDL_Dot(r0_ddl, rcgs_ddl, rhon, ierr)

         betan = rhon/rhonm1

         if (iter.eq.1) then
            call DDL_Copy(rcgs_ddl, uv_ddl, ierr)
            call DDL_Copy(rcgs_ddl, p_ddl, ierr)
         else
            call DDL_Copy(rcgs_ddl, uv_ddl, ierr)
            call DDL_Axpy(betan, q_ddl, uv_ddl, ierr)
            call DDL_Copy(q_ddl, tmp_ddl, ierr)
            call DDL_Axpy(betan, p_ddl, tmp_ddl, ierr)
            call DDL_Copy(uv_ddl, p_ddl, ierr)
            call DDL_Axpy(betan, tmp_ddl, p_ddl, ierr)
         end if

         call DDL_Smv(a_ddl, p_ddl, v_ddl, gamma, delta, ierr)

         if (ilu) then
            call DDL_Copy(v_ddl, tmp_ddl, ierr)
            call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl,
     &                         v_ddl, ierr)
         else if (ilunc) then
            call DDL_Copy(v_ddl, tmp_ddl, ierr)
         call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &           v_ddl, nc, expinv,
     &           iwork, iwork_dim, rwork, rwork_dim, ierr)
         end if
      
C*****************************************************************
C
C                      SECTION (a)
C
C*****************************************************************

         call DDL_Dot(r0_ddl, v_ddl, sigma, ierr)
         alpha = rhon/sigma
         malpha = -alpha
         oldcgs = rnrm
         call DDL_Copy(uv_ddl, q_ddl, ierr)
         call DDL_Axpy(malpha, v_ddl, q_ddl, ierr)
         call DDL_Copy(uv_ddl, w_ddl, ierr)
         call DDL_Axpy(1.0d0, q_ddl, w_ddl, ierr)
         if (ilu) then
            call DDL_Smv(a_ddl, w_ddl, z_ddl, gamma, delta, ierr)
            call DDL_Copy(z_ddl, tmp_ddl, ierr)
            call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl,
     &                         z_ddl, ierr)
            call DDL_Axpy(malpha, z_ddl, rcgs_ddl, ierr)
         else if (ilunc) then
            call DDL_Smv(a_ddl, w_ddl, z_ddl, gamma, delta, ierr)
            call DDL_Copy(z_ddl, tmp_ddl, ierr)
            call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &           z_ddl, nc, expinv,
     &           iwork, iwork_dim, rwork, rwork_dim, ierr)
            call DDL_Axpy(malpha, z_ddl, rcgs_ddl, ierr)
         else
            call DDL_Smv(a_ddl, w_ddl, tmp_ddl, gamma, delta, ierr)
            call DDL_Axpy(malpha, tmp_ddl, rcgs_ddl, ierr) 
         end if

C*****************************************************************
C
C                      SECTION (b)
C
C*****************************************************************

         call DDL_Nrm2(rcgs_ddl, rnrm, ierr)  
         omega = dsqrt(oldcgs*rnrm)
         m = 2*iter-1   
         const = theta*theta*eta/alpha
         theta = omega/tau
         c = 1.0d0/(dsqrt(1.0d0+theta*theta))
         tau = tau*theta*c
         eta = c*c*alpha
         call DDL_Scal(const, d_ddl, ierr)
         call DDL_Axpy(1.0d0, uv_ddl, d_ddl, ierr)
         call DDL_Axpy(eta, d_ddl, x_ddl, ierr) 
         error = dsqrt(dble(m+1))*dabs(tau)/bnrm
         omega = rnrm
         m = 2*iter
         const = theta*theta*eta/alpha
         theta = omega/tau
         c = 1.0d0/(dsqrt(1.0d0+theta*theta))
         tau = tau*theta*c
         eta = c*c*alpha
         call DDL_Scal(const, d_ddl, ierr) 
         call DDL_Axpy(1.0d0, q_ddl, d_ddl, ierr)  
         call DDL_Axpy(eta, d_ddl, x_ddl, ierr)
         error = dsqrt(dble(m+1))*dabs(tau)/bnrm
         call DDL_Nrm2(x_ddl, xnrm, ierr)  

         if (error.LT.(tol(tolcount))) then
            tolcount = tolcount + 1
         end if
        
         rhonm1 = rhon
      end do

c     remove distributed objects
      if (ilu) then
         call DDL_Free(l_ddl, ierr)
         call DDL_Free(u_ddl, ierr)
         call DDL_Free(dinv_ddl, ierr)
         call DDL_Free(z_ddl, ierr)
      end if
      if (ilunc) then
         call DDL_Free(ivec_ddl, ierr)
         call DDL_Free(rvec_ddl, ierr)
         call DDL_Free(z_ddl, ierr)
      end if
      call DDL_Free(rcgs_ddl, ierr)
      call DDL_Free(r0_ddl, ierr)
      call DDL_Free(uv_ddl, ierr)
      call DDL_Free(p_ddl, ierr)
      call DDL_Free(v_ddl, ierr)
      call DDL_Free(d_ddl, ierr)
      call DDL_Free(tmp_ddl, ierr)
      call DDL_Free(q_ddl, ierr)
      call DDL_Free(w_ddl, ierr)
 1000 end

