/* $Id: cilu.c,v 1.1 1995/08/24 15:39:48 tim Exp $ */

/* non-permuted ilu solver */

/* Copyright 1995 IASC */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

#define hbfilename      "../parsim/data/sher1"
#define infilename      "indemo"
#define outfilename     "outdemo"

#define MAX_ERR         1.0E-6

int main(int argc, char **argv)
{
  MPI_Comm comm;                 /* communicator for processes */
  int block;
  int ierr;
  int maxnelt;
  int nelt, nr;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];                  /* object type parameters from file */
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, alpha, beta, err;
  FILE *infile, *outfile;
  int fformat;
  int dummy, isym;
  int *px, *ipx, *py, *ipy;
  double *x, *y;
  int nx, ny, nz, tmp;
  DDL_Vec d_ddl;
  DDL_Spa a_ddl,l_ddl,u_ddl;
  int pt, rot, k;

  /* limit CPU run-time for each node 
     optional, but probably wise! 
     (time given in seconds) 
     */
  DDL_Set_cpu_limit(1*60);

  /* initialise MPI and sparse DDL*/
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);


  if (DDL_Ishost()) {
    CreateInput();
  }

  /* open input and output files */
  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm);
  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm); 

  /* read in the information heading */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  nelt = sizes[1];
  format = DDL_PCK_ROW;
  block = 1;

  /* set up size vector */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt;

  /* create DDL sparse matrix + upper and lower matrices*/
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &a_ddl);

  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &u_ddl);

  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &l_ddl);

  /* read data from file to sparse object */
  ierr = DDL_Read(a_ddl, infile);

/* create diagonal block matrix */
    ierr = DDL_Create_vector(n*block, type, block, comm, &d_ddl);

  /* initialize constants on all processes */
  alpha = 1.0;
  beta = 1.0;

  /* create ilu form data */ 
  ierr = DDL_ILU_comm(a_ddl);

  /* now form ilu factors */
  ierr = DDL_ILU_form(a_ddl,l_ddl,u_ddl,d_ddl); 

  /* write result to file */
  ierr = DDL_Write(u_ddl, outfile);
  ierr = DDL_Write(l_ddl, outfile);
  ierr = DDL_Write(d_ddl, outfile);

  /* remove DDL objects */
  ierr = DDL_Free(a_ddl);
  ierr = DDL_Free(u_ddl);
  ierr = DDL_Free(l_ddl);
  ierr = DDL_Free(d_ddl);

  /* close input and output files */
  ierr = DDL_Close(infile, comm);
  ierr = DDL_Close(outfile, comm);

  /* tidy up and exit */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}


/* convert Harwell-Boeing file to row-packed triad file */
int CreateInput(void)
{
  FILE *infile, *outfile;
  int ierr, format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, isym, size;
  double soln[maxn], rhs[maxn], xinit[maxn];

  /* read in sparse matrix from file */
  job = 3;
  n = maxn;
  nelt = mxnelt;
  /* read in as column gathered */
  ierr = DDL_Open_host(&infile, hbfilename, DDL_READ_MODE);  
  ierr = DDL_Inhb(infile, &n, &nelt, ia, ja, a,
                  &isym, soln, rhs, xinit, &job);
  ierr = DDL_Close_host(infile);
  type = MPI_DOUBLE;
  /* convert to triad */
  ierr = DDL_Coltotriad(n, nelt, ia, ja, a, type, isym);
  /* convert triad to row */
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  /* convert row to triad */
  ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);
  /* write out as row ordered triad format */
  /* this data file will be used for rest of demo */
  format = DDL_TRIAD_ROW;
  ierr = DDL_Open_host(&outfile, infilename, DDL_WRITE_MODE);
  ierr = DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
                        isym, format);
  ierr = DDL_Out_vector(outfile, rhs, type, &n);
  ierr = DDL_Close_host(outfile);
}
