/* $Id: cperm.c,v 1.2 1995/08/21 10:19:09 tim Exp $ */

/* permuted sparse mv and triangular solve */

/* Copyright 1995 IASC */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

#define hbfilename      "test4001.hbf"
#define infilename      "indemo"
#define outfilename     "outdemo"

int main(int argc, char **argv)
{
  MPI_Comm comm;                 /* communicator for processes */
  int block;
  int ierr;
  int maxnelt;
  int nelt, nr;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];                  /* object type parameters from file */
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, alpha, beta;
  FILE *infile, *outfile;
  int fformat;
  int dummy, isym;
  int *px, *ipx, *py, *ipy;
  double *x, *b;
  int nx, ny, nz, tmp;
  DDL_Vec x_ddl, b_ddl, px_ddl, py_ddl;
  DDL_Spa a_ddl;
  int pt, rot;

  /* limit CPU run-time for each node 
     optional, but probably wise! 
     (time given in seconds) 
     */
  DDL_Set_cpu_limit(1*60);

  /* initialise MPI and sparse DDL*/
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);

  /* open input and output files */
/*  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm); */
/*  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm);*/


  nx = 5;
  ny = 5;
  nz = 5;
  n = nx*ny*nz;
  
  type = MPI_DOUBLE;

  /* make vectors on host--easier sequentially! */
  /* set up vectors so that result is always the same except it will
     have been permuted */
  if (DDL_Ishost()) {
    px = malloc(sizeof(int)*n);
    py = malloc(sizeof(int)*n);
    ipx = malloc(sizeof(int)*n);
    ipy = malloc(sizeof(int)*n);
    x = malloc(sizeof(double)*n);
    b = malloc(sizeof(double)*n);
    /* initialise data vectors */
    for (i=0;i<n;i++) x[i] = 1.0/(i+1);
    for (i=0;i<n;i++) b[i] = (i+1);
    /* set up permutation vectors */
    printf("Enter permutation number for px (1-5): ");
    scanf("%d", &pt);
    set_perm(pt, px, ipx, n, nx, ny, nz);
    printf("Enter permutation number for py (1-5): ");
    scanf("%d", &pt);
    set_perm(pt, py, ipy, n, nx, ny, nz);
    /* apply permutation to data vectors */
    ierr = DDL_Lpermute_vector(x, type, py, n); 
    ierr = DDL_Lpermute_vector(b, type, px, n); 
  }

  /* set up sizes */
  /* full 125x125 matrix */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt;

  type = MPI_DOUBLE;
  format = DDL_PCK_ROW;
  block = 1;

  /* create data vectors */
  type = MPI_DOUBLE;
  x_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &x_ddl);
  b_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &b_ddl);

  /* create permutation vectors */
  type = MPI_INT;
  px_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &px_ddl);
  py_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &py_ddl);

  /* distribute permutation vectors */
  ierr = DDL_Scatter_vector(px_ddl, px, n, 0);
  ierr = DDL_Scatter_vector(py_ddl, py, n, 0);

  /* distribute vectors */
  ierr = DDL_Scatter_vector(x_ddl, x, n, 0);
  ierr = DDL_Scatter_vector(b_ddl, b, n, 0);

  /* create sparse matrix */
  a_ddl = DDL_SPARSE_NULL;
  type = MPI_DOUBLE;
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &a_ddl);

  /* copy permutation vectors to sparse object permutation vectors */
  ierr = DDL_Addperm_sparse(a_ddl, px_ddl, py_ddl);
/*  ierr = DDL_Addperm_sparse(a_ddl, NULL, p2_ddl);*/

  /* initialise distributed matrix */
  ierr = DDL_Get_ia(a_ddl, &ia);
  ierr = DDL_Get_ja(a_ddl, &ja);
  ierr = DDL_Get_a(a_ddl, &a);
  ierr = DDL_Type_sparse(a_ddl, &type);
  ierr = DDL_Gsize_sparse(a_ddl, sizes);
  n = sizes[0];
  ierr = DDL_Size_sparse(a_ddl, sizes);
  maxnelt = sizes[1];
  nr = sizes[0];
  ierr = DDL_Offset_sparse(a_ddl, &offset);
  /* full matrix with a(i,j)=j+10*i */
  /* triad format */
  for (i=0;i<nr;i++) {
    for (j=0;j<n;j++) {
      ia[i*n+j] = i+1; /* ia stored as local F indices */
      ja[i*n+j] = j+1;
      a[i*n+j] = j+1+10*(i+offset);
    }
  }
  /* convert to row packed format */
  isym = 0;
  ierr = DDL_Triadtorow(nr, nr*n, ia, ja, a, type, isym);

  /* tell DDL about new matrix values */
  ierr = DDL_Setformat_sparse(a_ddl, DDL_PCK_ROW);
  ierr = DDL_Setused_sparse(a_ddl, nr*n);

  /* write matrix to file */
/*  ierr = DDL_Write(a_ddl, outfile);*/

  /* initialize constants on all pocesses */
  alpha = 1.0;
  beta = 1.0;
  
  /* create matrix-vector multiply data */
  ierr = DDL_Smvp_comm(a_ddl);

  /* now perform sparse matrix vector multiply */
  ierr = DDL_Smvp(a_ddl, x_ddl, b_ddl, alpha, beta);

  /* get result vector back to host */
  ierr = DDL_Gather_vector(b_ddl, b, n, 0);

  /* unpermute result vector */ 
  if (DDL_Ishost()) {
    /* apply inverse permutation to result vector */
    type = MPI_DOUBLE;
    ierr = DDL_Lpermute_vector(b, type, ipx, n); 
    /* write result to file */
    ierr = DDL_Open_host(&outfile, outfilename, DDL_WRITE_MODE);
    ierr = DDL_Out_vector(outfile, b, type, &n);
    ierr = DDL_Close_host(outfile);
  }

  /* release memory on host */ 
  if (DDL_Ishost()) {
    free(px);
    free(ipx);
    free(py);
    free(ipy);
    free(x);
    free(b);
  }

  /* write result to file */
/*  ierr = DDL_Write(b_ddl, outfile);*/

  /* remove DDL objects */
  DDL_Free(a_ddl);
  DDL_Free(x_ddl);
  DDL_Free(b_ddl);
  DDL_Free(px_ddl);
  DDL_Free(py_ddl);

  /* close input and output files */
/*  ierr = DDL_Close(infile, comm);*/
/*  ierr = DDL_Close(outfile, comm);*/

  /* tidy up and exit */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}

int set_perm(int pt, int *pv, int *ipv, int n, int nx, int ny, int nz)
{
  int i, rot, ierr;
  switch (pt) {
  case 1:
    /* identity permutation */
    for (i=0;i<n;i++)
      pv[i] = i;
    break;
  case 2:
    /* red/black permutation */
    ierr = DDL_calc_perm(nx, ny, nz, pv);
      break;
  case 3:
    /* rotate by rot */
    rot = 1;
    for (i=0;i<n;i++)
      pv[i] = (i+rot)%n;
    break;
  case 4:
    /* invert vector */
    for (i=0;i<n;i++)
      pv[i] = n-i-1;
    break;
  case 5:
    /* swap single pair (extremes) */
    for (i=0;i<n;i++)
      pv[i] = i;
    pv[0] = n-1;
    pv[n-1] = 0;
    break;
  default:
    /* identity permutation */
    for (i=0;i<n;i++)
      pv[i] = i;
    break;
  }      
  DDL_Calc_invperm(pv, ipv, n);
}
