/* $Id: ctesterr.c,v 1.1 1995/08/21 10:19:19 tim Exp $ */

/* test user error handler */

/* Copyright 1995 IASC */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

#define hbfilename      "../parsim/data/test4001.hbf"
#define infilename      "indemo"
#define outfilename     "outdemo"

int main(int argc, char **argv)
{
  MPI_Comm comm;                 /* communicator for processes */
  DDL_Spa m;
  DDL_Vec x, y;
  int ierr;
  int maxnelt;
  int nelt;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];                  /* object type parameters from file */
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, alpha, beta;
  FILE *infile, *outfile;
  int fformat;
  int dummy;
  int block;

  block = 5;

  /* limit CPU run-time for each node 
     optional, but probably wise! 
     (time given in seconds) 
     */
  DDL_Set_cpu_limit(1*60);

  /* initialise MPI and sparse DDL*/
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);

  if (DDL_Ishost()) {
    CreateInput();
  }

  /* open input and output files */
  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm);
  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm);

  /* read in the information heading */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  nelt = sizes[1];
  format = DDL_PCK_ROW;
  
  /* set up size vector */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt;

  /* create DDL sparse matrix */
  m = DDL_SPARSE_NULL;
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &m);

  /* read data from file to sparse object */
  ierr = DDL_Read(m, infile);

  /* print object details for each process */
  DDL_Info_sparse(m);

  /* create vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  x = DDL_VECTOR_NULL;
  /* deliberately cause error */
  ierr = DDL_Create_vector(-5, type, block, comm, &x);

  /* create second vector */
  y = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &y);

  /* fill vector with data from file */
  ierr = DDL_Read(x, infile);

  /* initialize vector y */
  ierr = DDL_Get_vector(y, &data);
  ierr = DDL_Size_vector(y, &size);
  for (i=0;i<size;i++) data[i] = 0.0;

  /* initialize constants on all pocesses */
  alpha = 1.0;
  beta = 1.0;
  
  /* create matrix-vector multiply data */
  ierr = DDL_Smv_comm(m);

  /* now perform sparse matrix vector multiply */
  ierr = DDL_Smv(m, x, y, alpha, beta); 

  /* write data to file */
/*  ierr = DDL_Write(m, outfile);*/
  ierr = DDL_Write(y, outfile);

  /* remove DDL objects */
  DDL_Free(m);
  DDL_Free(x);
  DDL_Free(y);

  /* close input and output files */
  ierr = DDL_Close(infile, comm);
  ierr = DDL_Close(outfile, comm);

  /* tidy up and exit */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}

/* convert Harwell-Boeing file to row-packed triad file */
int CreateInput(void)
{
  FILE *infile, *outfile;
  int ierr, format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, isym, size;
  double soln[maxn], rhs[maxn], xinit[maxn];

  /* read in sparse matrix from file */
  job = 3;
  n = maxn;
  nelt = mxnelt;
  /* read in as column gathered */
  ierr = DDL_Open_host(&infile, hbfilename, DDL_READ_MODE);
  ierr = DDL_Inhb(infile, &n, &nelt, ia, ja, a, 
		  &isym, soln, rhs, xinit, &job);
  ierr = DDL_Close_host(infile);
  type = MPI_DOUBLE;
  /* convert to triad */
  ierr = DDL_Coltotriad(n, nelt, ia, ja, a, type, isym);
  /* convert triad to row */
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  /* convert row to triad */
  ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);
  /* write out as row ordered triad format */
  /* this data file will be used for rest of demo */
  format = DDL_TRIAD_ROW;
  ierr = DDL_Open_host(&outfile, infilename, DDL_WRITE_MODE);
  ierr = DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
			isym, format);
  ierr = DDL_Out_vector(outfile, rhs, MPI_DOUBLE, &n);
  ierr = DDL_Close_host(outfile);
}

