c     $Id: tfqmrnc.f,v 1.4 1995/08/23 12:05:27 tim Exp $

c     MPI sparse DDL demo in FORTRAN (Andy's tfqmrnc)

c     Copyright 1994 IASC

c     DESCRIPTION   DDL based iterative solver for Ax=b.
c                   Block preconditioning techniques are  
c                   incorporated into this version.

      program tfqmrnc

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character hbfilename*(100), ddlfilename*(100), datadir*(100)
      character solnfilename*(100), summaryfilename*(100)
      data hbfilename /'test4001.hbf\0'/
      data datadir /'../parsim/data/'/
      data ddlfilename /'tfqmrdata\0'/
      data solnfilename /'soln\0'/
      data summaryfilename /'summary\0'/

c     Workspace for ilu blocks & solves 9/1/95
      integer maxiwork, maxrwork
      parameter(maxiwork=1000000, maxrwork=1000000) ! local size
      integer iwork(maxiwork)
      double precision rwork(maxrwork)
      
      integer i,n,iter,itmax,loop
      double precision error,tol(10),gamma,delta,epsilon
      double precision epsilon2
      integer nelt, tolno, nc
      logical rhsexist,ilu,ilunc,expinv
      logical nopc
      character*80 ans

      integer infile, outfile
      integer ierr
      integer block, format
      integer comm
      integer a_ddl, x_ddl, b_ddl, xtrue_ddl, tmp_ddl
      integer type, fformat, sizes(3), len, precon

c     initialise MPI, DDL
      call MPI_Init(ierr)
      call DDL_Init_sparse(MPI_COMM_WORLD, comm, ierr)

      if (DDL_Ishost()) then 
        print *,' '
        print *,'                TFQMR SPARSE DDL SOLVER'
        print *,'               *************************'

c********************************************************************
c*                                                                  *
c*                   READ ITERATION PARAMETERS 
c*                                                                  *
c********************************************************************

        write(*,*) 'Input demonstration options'
        write(*,*) '(for default [in brackets] just press RETURN)'
        write(*,*) ' '

        write(*,*) 'Input name of data file [test4001.hbf]'
        read(*,'(a)') ans
        if (ans .eq. ' ') then
           hbfilename = 
     +       datadir(1:index(datadir,' ')-1) // 'test4001.hbf'
        else
           hbfilename = datadir(1:index(datadir,' ')-1) // ans
        end if
c       form C file name string for DDL procedures
        len=index(hbfilename,' ')
        hbfilename(len:len) = '\0'

        print *,'Is the problem supplied with right hand'
        print *,'side and solution vectors y/n? [Y]'
        read(*,'(a)') ans
        if (ans .eq. ' ') then
           ans = 'y'
        end if
        if ((ans.eq.'Y').or.(ans.eq.'y')) then
           rhsexist = .TRUE.
        else
           rhsexist = .FALSE.
        end if
        print *,' '
        
        print *,'Enter Max. No. Iterations [100]'
        read(*,'(a)') ans
        if (ans .eq. ' ') then
           itmax = 100
        else
           read (ans, *) itmax
        end if

        print *,'Enter no. of tolerance levels required (max 10) [1]'
        read(*,'(a)') ans
        if (ans .eq. ' ') then
           tolno = 1
        else
           read (ans, *) tolno
        end if
        print *,' '
        
        print *,'Enter set of tolerances (smallest last)'
        
        do i = 1,tolno
           print *,'Enter tolerance ',i,' [1e-6]'
           read(*,'(a)') ans
           if (ans .eq. ' ') then
              tol(i) = 1e-6
           else
              read (ans, *) tol(i)
           end if
        end do
        print *,' '
        
        print *,'Enter e, where x_initial = x_true + e(rand(0,1))'
        print *,'(typically e should lie between 0.001 and 1.0):'
        print *,'An entry of zero gives an initial guess of zero [0]'
        read(*,'(a)') ans
        if (ans .eq. ' ') then
           epsilon = 0.0D0
        else
           read (ans, *) epsilon
        end if

        print *,'Is block preconditioning to be applied [Y]'   
        read(*,'(a)') ans
        if (ans .eq. ' ') then
           ans = 'y'
        end if
        if ((ans.eq.'Y').or.(ans.eq.'y')) then
 10        print *,'Choose the preconditioner'
           print *,'type "ilu" for incomplete LU factorization'
           print *,'     "inc" for incomplete LUnc factorization [INC]'
           read(*,'(a)') ans
           if (ans .eq. ' ') then
              ans = 'inc'
           end if
           if ((ans.eq.'ILU').or.(ans.eq.'ilu')) then
              ilu = .TRUE.
              print *,' '
              print *,'ILU block preconditioner to be applied'
              print *,' '
              ilunc = .FALSE.
              nopc = .FALSE.
           else if ((ans.eq.'INC').or.(ans.eq.'inc')) then
              ilunc = .TRUE.
              print *,' '
              print *,'ILU(nc) block preconditioner to be applied'
              print *,' '
              print *,'Enter nc parameter [4]'
              read(*,'(a)') ans
              if (ans .eq. ' ') then
                 nc = 4
              else
                 read (ans, *) nc
              end if
              print *,'Do you want to form explicit cell diagonals? [N]'
              read(*,'(a)') ans
              if (ans .eq. ' ') then
                 ans = 'n'
              end if
              if ((ans.eq.'Y').or.(ans.eq.'y')) then
                 expinv = .TRUE.
              else
                 expinv = .FALSE.
              end if
              ilu = .FALSE.
              nopc = .FALSE.
           else 
              print *,' '
              print *,'*Unrecognizeable input* - try again y/n'
              print *,' '
              read (*,'(a)') ans
              if ((ans.eq.'Y').or.(ans.eq.'y')) then 
                 go to 10
              else
                 print *,' '
                 print *,'No Preconditioning applied'
                 print *,' '
                 nopc = .TRUE.
                 ilu = .FALSE.
                 ilunc = .FALSE.
              end if
           end if
        else
           print *,' '
           print *,'No Preconditioning applied'
           print *,' '
           nopc = .TRUE.
           ilu = .FALSE.
           ilunc = .FALSE.
        end if

        print *, 'OK. Starting run...'

        call createinput(hbfilename, ddlfilename, rhsexist)
        
      end if ! DDL_Ishost

c     broadcast user input to all processes
      call MPI_Bcast(itmax, 1, MPI_INTEGER, 0, comm, ierr)
      call MPI_Bcast(rhsexist, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(ilu, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(ilunc, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(nc, 1, MPI_INTEGER, 0, comm, ierr)
      call MPI_Bcast(expinv, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(nopc, 1, MPI_LOGICAL, 0, comm, ierr)
      call MPI_Bcast(epsilon, 1, MPI_DOUBLE_PRECISION, 0, comm, ierr)
      call MPI_Bcast(tolno, 1, MPI_INTEGER, 0, comm, ierr)
      do i = 1, tolno
         call MPI_Bcast(tol(i), 1, MPI_DOUBLE_PRECISION, 0, comm, ierr)
      end do

c     open input and output files
      call DDL_Open(infile, ddlfilename, DDL_READ_MODE, comm, ierr)
      call DDL_Open(outfile, solnfilename, DDL_WRITE_MODE, comm, ierr)

c     read in the information heading
      call DDL_Fileformat(infile, fformat, type, sizes, comm, ierr)
      n = sizes(1)
      nelt = sizes(2)
      format = DDL_PCK_ROW
      block = 4

c     set up size vector
      sizes(1) = n
      sizes(2) = n
      sizes(3) = nelt*2

c     create sparse matrix
      a_ddl = DDL_NULL
      call DDL_Create_sparse(sizes, type, format, 
     &                       block, comm, a_ddl, ierr)

c     create vector for solution
      x_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, x_ddl, ierr)
c     create vector for result
      b_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, b_ddl, ierr)
c     create vector for true solution
      xtrue_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, xtrue_ddl, ierr)
c     create temp vector
      tmp_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, tmp_ddl, ierr)

c     read data from file to sparse object
      call DDL_Read(a_ddl, infile, ierr)

      if (.not. rhsexist) then
c        initialise xtrue to one
         call DDL_Fill(xtrue_ddl, 1.0D0, ierr)
         call DDL_Smv(a_ddl, xtrue_ddl, b_ddl, gamma, delta, ierr)
      else
c        read rhs from file
         call DDL_Read(b_ddl, infile, ierr)
c        read soln vector from file
         call DDL_Read(xtrue_ddl, infile, ierr)
      endif

c     initialise x 
      call DDL_Rand(1.0d0, tmp_ddl, ierr)
      call DDL_Copy(xtrue_ddl, x_ddl, ierr)
      if (epsilon.eq.0.0d0) then
         call DDL_Scal(0.0d0, x_ddl, ierr)
      else 
         call DDL_Axpy(epsilon, tmp_ddl, x_ddl, ierr)
      end if

      if (nopc) precon = 0
      if (ilu) precon = 1
      if (ilunc) precon = 2

      call DDL_Linsolve(a_ddl, x_ddl, b_ddl, itmax, iter, 
     &     tol, tolno, precon, nc, expinv, error, 
     &     iwork, maxiwork, rwork, maxrwork, ierr)

c     Why have we stopped?
      if (DDL_Ishost()) then
         if(iter .lt. itmax) THEN
            write (*,*)
            write (*,'(A22,I4,A11)') 'TFQMR converged after',iter,
     +           'iterations'
            write (*,*)
         else 
            write (*,*) 
            write (*,'(A37)') ' Specified Max. Iterations reached:' 
            write (*,'(A15,I4,A25,I4)') ' Max.iters = ',itmax,
     +           'Iterations reached = ',iter
            write (*,*)
         end if
      end if

      call DDL_Write(x_ddl, outfile, ierr)

      call DDL_Close(outfile, comm, ierr)
      call DDL_Close(infile, comm, ierr)

c     check solution is correct
      if (DDL_Ishost()) then
         call check(ddlfilename, solnfilename, rhsexist, error)
         print *, 'Max error is', error
      end if

c     delete distributed objects
      call DDL_Free(a_ddl, ierr)
      call DDL_Free(x_ddl, ierr)
      call DDL_Free(b_ddl, ierr)
      call DDL_Free(xtrue_ddl, ierr)
      call DDL_Free(tmp_ddl, ierr)

c     tidy up and exit
      call DDL_Finalize_sparse(comm, ierr)
      call MPI_Finalize(ierr)

      stop
      end


c     convert Harwell-Boeing file to row-packed triad file
c     HB file has matrix; and  rhs, soln if .rhsexist.
c     output matrix; rhs, soln if .rhsexist.
      subroutine createinput(hbfilename, ddlfilename, rhsexist)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character*(*) hbfilename, ddlfilename
      logical rhsexist

      integer maxn, mxnelt, np
      parameter (maxn = 5005, mxnelt = 40000, np = 10)

      integer infile, outfile
      integer ierr, format
      integer type
      integer ia(mxnelt), ja(mxnelt)
      double precision a(mxnelt)
      integer nelt, job
      integer n, i, isym, size
      double precision soln(maxn), rhs(maxn), xinit(maxn)
      
c     read in sparse matrix from file
      job = 3
      n = maxn
      nelt = mxnelt
c     read in as column gathered
      call DDL_Open_host(infile, hbfilename, DDL_READ_MODE, ierr)
      call DDL_Inhb(infile, n, nelt, ia, ja, a, 
     &     isym, soln, rhs, xinit, job, ierr)
      call DDL_Close_host(infile, ierr)
      type = MPI_DOUBLE_PRECISION
c     convert to triad 
      call DDL_Coltotriad(n, nelt, ia, ja, a, type, isym, ierr)
c     convert triad to row
      call DDL_Triadtorow(n, nelt, ia, ja, a, type, isym, ierr)
c     convert row to triad
      call DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym, ierr)
c     write out as row ordered triad format
c     this data file will be used for rest of demo
      format = DDL_TRIAD_ROW
      call DDL_Open_host(outfile, ddlfilename, DDL_WRITE_MODE, ierr)
      call DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
     &     isym, format, ierr)

      if (rhsexist) then
        if (job .eq. 23 .or. job .eq. 31) then
c         write out rhs & sol
          call DDL_Out_vector(outfile, rhs, MPI_DOUBLE_PRECISION, 
     &                        n, ierr)
          call DDL_Out_vector(outfile, soln, MPI_DOUBLE_PRECISION, 
     &                        n, ierr)
        else
          print *, 'Error reading RHS and SOLN vectors'
        end if
      end if
      
      call DDL_Close_host(outfile, ierr)
      return
      end

c     check result vector
      subroutine check(ddlfilename, solnfilename, rhsexist, maxerr)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character*(*) solnfilename, ddlfilename
      logical rhsexist
      double precision maxerr

      integer maxn, mxnelt, np
      parameter (maxn = 5005, mxnelt = 40000, np = 10)

      integer infile, outfile
      integer ierr, format
      integer type
      integer n, i
      double precision soln(maxn), correct(maxn), error
      
c     get calculated solution vector
      call DDL_Open_host(infile, solnfilename, DDL_READ_MODE, ierr) 
      n = maxn
      call DDL_In_vector(infile, soln, type, n, ierr)  
      call DDL_Close_host(infile, ierr)
c     get correct solution vector
      if (rhsexist) then
         call DDL_Open_host(infile, ddlfilename, DDL_READ_MODE, ierr) 
         call DDL_Skip_host(infile, ierr)
         call DDL_Skip_host(infile, ierr)
         n = maxn
         call DDL_In_vector(infile, correct, type, n, ierr)  
         call DDL_Close_host(infile, ierr)
      else
c        set to 1
         do i=1, n
            soln(i) = 1.0D0
         end do
      end if

c     check for differences
      maxerr = 0.0D0
      do i=1, n
         error = ABS(ABS(soln(i))-ABS(correct(i)))
         if (error .gt. maxerr) then
	   maxerr = error
         end if
      end do

      return
      end
