c     $Id: fdemo.f,v 1.3 1995/08/23 11:11:34 tim Exp $

c     Copyright 1994 IASC

c     A simple example of Sparse-DDL user programs in FORTRAN
c     using Cray pointers 

      program fdemo

c     programs must always include these two header files
      include "mpif.h"
      include "ddlf.h"

c     constants for sparse data structures 
      integer maxn, mxnelt, np
      parameter (maxn = 5005, mxnelt = 40000, np = 10)

c     filename constants
      character infilename*(*), outfilename*(*)
      parameter (infilename = "indemo\0")
      parameter (outfilename = "outdemo\0")

c     variables
      integer ierr
      double precision res, val, alpha, beta

c     MPI/DDL variables and handles are all integers
      integer comm                            ! MPI communicator handle
      integer type                            ! MPI datatype handle

c     DDL object variables
      integer m                               ! DDL_Spa sparse object handle
      integer x, y                            ! DDL_Vec vector object handles
      integer sizes(3), fformat, format, n, nelt, block
      integer sz, offset, maxnelt, dummy

c     data pointers and arrays for IBM xlf v.2, SUN/SGI f77
      double precision a(10), vec(10)
      integer ia(10), ja(10)
      pointer (aptr, a)
      pointer (iaptr, ia)
      pointer (japtr, ja)
      pointer (vecptr, vec)

c     file handles for use by DDL I/O routines
      integer infile, outfile                 ! C FILE handles

c     initialise MPI, DDL
c     these two lines must be included at the start of all user programs
c     DDL_Init_sparse sets up an MPI communicator, comm, compatible with 
c     the DDL from MPI_COMM_WORLD. 
      call MPI_Init(ierr)
      call DDL_Init_sparse(MPI_COMM_WORLD, comm, ierr)

c     create DDL input file on host process 
      if (DDL_Ishost()) call CreateInput()

c     block size to use for distributions
      block = 5

c     open input and output files (on all processes)
      call DDL_Open(infile, infilename, DDL_READ_MODE, comm, ierr)
      call DDL_Open(outfile, outfilename, DDL_WRITE_MODE, comm, ierr)

c     read in the header information about first object in file
c     which is a sparse matrix 
      call DDL_Fileformat(infile, fformat, type, sizes, comm, ierr)
      n = sizes(1)                         ! number of rows in matrix
      format = DDL_PCK_ROW                 ! we want row packed object 
  
c     set up size vector for DDL_Create_sparse
      sizes(1) = n
      sizes(2) = n
      sizes(3) = mxnelt ! maximum number of nonzeros to allocate space for

c     create DDL sparse matrix object
      call DDL_Create_sparse(sizes, type, format, block, comm, m, ierr)

c     read data from file to sparse matrix object
      call DDL_Read(m, infile, ierr)

c     create vector object for next vector in data file 
      call DDL_Fileformat(infile, fformat, type, sizes, comm, ierr)
      n = sizes(1)
      call DDL_Create_vector(n, type, block, comm, x, ierr)

c     create second identical vector
      call DDL_Create_vector(n, type, block, comm, y, ierr)

c     fill first vector with data from file
      call DDL_Read(x, infile, ierr)

c     initialize vector y  
c     first get pointer, size and offset for actual vector data values
c     on this process
      call DDL_Get_vector(y, vecptr, ierr)
      call DDL_Size_vector(y, sz, ierr)
      call DDL_Offset_vector(y, offset, ierr)

      val = 1.0

c     initialize y vector
      do i=1, sz
         vec(i) = val
      end do

c     repeat for other vector
      call DDL_Get_vector(x, vecptr, ierr)
      call DDL_Size_vector(x, sz, ierr)

c     initialize vector
      do i=1, sz
         vec(i) = dble(i)+offset
      end do

c     now perform dot product on distributed objects
      call DDL_Dot(x, y, res, ierr) 

c     now perform matrix vector multiply
      alpha = 1.0
      beta = 1.0
      call DDL_Smv_comm(m, ierr)
      call DDL_Smv(m, x, y, alpha, beta, ierr) 


c     second example initialises sparse matrix from the user program
c     --------------------------------------------------------------

c     update sparse matrix vectors directly
c     get pointer to actual vector data values for ia, ja and a
      call DDL_Get_ia(m, iaptr, ierr)
      call DDL_Get_ja(m, japtr, ierr)
      call DDL_Get_a(m, aptr, ierr)
c     find out data type of coefficient values
      call DDL_Type_sparse(m, type, ierr)
c     find out global size of matrix
      call DDL_Gsize_sparse(m, sizes, ierr)
      n = sizes(1) ! number of columns (rows) in global sparse matrix
c     find out local size of matrix
      call DDL_Size_sparse(m, sizes, ierr)
      maxnelt = sizes(2) ! maximum number of elements locally
      num_rows = sizes(1) ! number of rows on this process
c     find out offset of local part of matrix
      call DDL_Offset_sparse(m, offset, ierr)

c     insert nonzeros to create a diagonal matrix, globally a(i,i) = i
c     row-packed format, DDL_PCK_ROW 
      do i=1, num_rows
         ia(i) = i
         ja(i) = i+offset-1
         a(i) = i+offset-1
      end do
c     add end pointer
      ia(num_rows+1) = num_rows+1

c     tell DDL about new sparse matrix values
      call DDL_Setformat_sparse(m, DDL_PCK_ROW, ierr)
      call DDL_Setused_sparse(m, num_rows, ierr)

c     now perform matrix vector multiply
      alpha = 1.0
      beta = 1.0
      call DDL_Smv_comm(m, ierr)
      call DDL_Smv(m, x, y, alpha, beta, ierr) 

c     write data to file
      call DDL_Write(m, outfile, ierr)
      call DDL_Write(y, outfile, ierr)

c     remove sparse object
      call DDL_Free(m, ierr)
      call DDL_Free(x, ierr)
      call DDL_Free(y, ierr)

c     close input and output files
      call DDL_Close(infile, comm, ierr)
      call DDL_Close(outfile, comm, ierr)

c     print OK message
      if (DDL_Ishost()) then
         print *, 'Demonstration program finished OK (phew!)'
      end if

c     tidy up and exit
c     these calls must be included at the end of every DDL program
      call DDL_Finalize_sparse(comm, ierr)
      call MPI_Finalize(ierr)
      stop
      end

c     convert Harwell-Boeing file to row-packed triad file 
c     only executed by host process with file I/O capability
      subroutine createinput

      include "mpif.h"
      include "ddlf.h"

c     constants for sparse data structures 
      integer maxn, mxnelt, np
      parameter (maxn = 5005, mxnelt = 40000, np = 10)

c     filename constants
c     notice trailing \0 character since these names are all
c     passed to C character handling routines
      character hbfilename*(*), infilename*(*), outfilename*(*)
      parameter (hbfilename = "../parsim/data/sher1\0")
      parameter (infilename = "indemo\0")
      parameter (outfilename = "outdemo\0")

      integer infile, outfile
      integer ierr, format
      integer type
      integer ia(mxnelt), ja(mxnelt)
      double precision a(mxnelt)
      integer nelt, job
      integer n, i, isym, size
      double precision soln(maxn), rhs(maxn), xinit(maxn)

c     open Harwell-Boeing data file 
      call DDL_Open_host(infile, hbfilename, DDL_READ_MODE, ierr)

c     read in Harwell-Boeing file format to DDL_PCK_COL, 
c     column packed format matrix
c     (also reads in rhs vector, soln and xinit) 
      job = 3
      n = maxn
      nelt = mxnelt
      call DDL_Inhb(infile, n, nelt, ia, ja, a, 
     &     isym, soln, rhs, xinit, job, ierr)

c     close input file 
      call DDL_Close_host(infile, ierr)

c     data type of coefficient elements
      type = MPI_DOUBLE_PRECISION

c     convert column packed format, DDL_PCK_COL, 
c     to row gathered triad, DDL_TRIAD_ROW, for output to DDL data file
c     first, convert to column gathered triad, DDL_TRIAD_COL 
      call DDL_Coltotriad(n, nelt, ia, ja, a, type, isym, ierr)
c     then, convert column gathered triad to row packed format, DDL_PCK_ROW
      call DDL_Triadtorow(n, nelt, ia, ja, a, type, isym, ierr)
c     finally, convert row packed format to row gathered triad, DDL_TRIAD_ROW 
      call DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym, ierr)

c     write out matrix as row ordered triad format
c     this data file will be used for rest of demo
      format = DDL_TRIAD_ROW
      call DDL_Open_host(outfile, infilename, DDL_WRITE_MODE, ierr)
      call DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
     &     isym, format, ierr)
c     write out rhs vector as well 
      call DDL_Out_vector(outfile, rhs, type, n, ierr)
      call DDL_Close_host(outfile, ierr)
      return
      end

