c     $Id: perfsetup.f,v 1.2 1995/08/23 09:39:31 tim Exp tim $

c     Copyright 1995 IASC

c     setup ddl input data file from sources...for perf program
c     eg. cowi_big problem

      program perfsetup

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character cowifilename*(100), ddlfilename*(100)
      data cowifilename 
     &     /'../parsim/data/cowi_big '/
      data ddlfilename 
     &     /'../parsim/data/cowi_big.ddl\0'/

c     Workspace for ilu blocks & solves 9/1/95
      integer maxiwork, maxrwork
      parameter(maxiwork=1000000, maxrwork=1000000) ! local size
      integer iwork(maxiwork)
      double precision rwork(maxrwork)
      
      integer i,n,iter,itmax,loop
      double precision error,tol(10),gamma,delta,epsilon
      double precision epsilon2
      integer nelt, tolno, nc
      logical rhsexist,ilu,ilunc,expinv
      logical nopc

      integer infile, outfile
      integer ierr
      integer block, format
      integer comm
      integer a_ddl, x_ddl, b_ddl, xtrue_ddl, tmp_ddl
      integer type, fformat, sizes(3), len, precon

c     initialise MPI, DDL
      call MPI_Init(ierr)
      call DDL_Init_sparse(MPI_COMM_WORLD, comm, ierr)

c     create DDL input file on host process 
      if (DDL_Ishost()) call createinput(cowifilename, ddlfilename)
        
c     tidy up and exit
      call DDL_Finalize_sparse(comm, ierr)
      call MPI_Finalize(ierr)

      stop
      end


c     convert big cowi data files to row-packed triad file
c     cowi file has matrix in .mat; and  rhs, soln if .rhsexist.
c     output matrix; rhs, soln if .rhsexist.
      subroutine createinput(cowifilename, ddlfilename)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

      character*(*) cowifilename, ddlfilename

      character*(100) filename
      integer maxn, mxnelt
      parameter (maxn = 20000, mxnelt = 200000)
      integer iu, ln
      parameter (iu = 1)

      integer infile, outfile
      integer ierr, format
      integer type
      integer ia(mxnelt), ja(mxnelt)
      double precision a(mxnelt)
      integer nelt, job
      integer n, i, isym, size
      double precision soln(maxn), rhs(maxn), xinit(maxn)
      
      ln = index(cowifilename, ' ')-1

c     read in sparse matrix from cowi file .mat
      filename = cowifilename(1:ln)//'.mat'
      open(unit=iu, file=filename)
      read (iu, *) n
      read (iu, *) nelt

c     read in elements as triples row, col, val
      do i=1, nelt
         read (iu, *) ia(i),ja(i),a(i) 
      end do
      close(iu)

c     read in rhs vector cowi file .rhs
      filename = cowifilename(1:ln)//'.rhs'
      open(unit=iu, file=filename)
      do i=1, n
         read (iu, *) rhs(i) 
      end do
      close(iu)

c     read in soln vector cowi file .sol
      filename = cowifilename(1:ln)//'.sol'
      open(unit=iu, file=filename)
         read (iu, *) (soln(i),i=1,n) 
      close(iu)

      type = MPI_DOUBLE_PRECISION
      isym = 0
c     convert triad to row
      call DDL_Triadtorow(n, nelt, ia, ja, a, type, isym, ierr)
c     convert row to triad
      call DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym, ierr)
c     write out as row ordered triad format
c     this data file will be used for rest of demo
      format = DDL_TRIAD_ROW
      call DDL_Open_host(outfile, ddlfilename, DDL_WRITE_MODE, ierr)
      call DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
     &     isym, format, ierr)

c     write out rhs & sol
      call DDL_Out_vector(outfile, rhs, MPI_DOUBLE_PRECISION, 
     &     n, ierr)
      call DDL_Out_vector(outfile, soln, MPI_DOUBLE_PRECISION, 
     &     n, ierr)
      
      call DDL_Close_host(outfile, ierr)
      return
      end



