/* $Id: ddl_dense.h,v 1.5 1995/08/29 11:27:30 tim Exp $ */

/* general header file for dense DDL routines */

/* Copyright 1994 IASC */

#ifndef _DDL_DENSE_H
#define _DDL_DENSE_H

#include <mpi.h>
#include "ddl_general.h"
#include "ddl_image.h"

/* dense DDL object types */

/* dense DDL object formats */
#define DDL_VECTOR_TYPE              200
#define DDL_VEC_BLK                  DDL_VECTOR+      1
#define DDL_VEC_REP                  DDL_VECTOR+      2
#define DDL_VEC_FMT                  DDL_VECTOR+      3

/* DDL dense library routine numbers */

extern int DDL_CREATE_VECTOR;
extern int DDL_FREE_VECTOR;
extern int DDL_WRITE_VECTOR;
extern int DDL_READ_VECTOR;
extern int DDL_SIZE_VECTOR;
extern int DDL_TYPE_VECTOR;
extern int DDL_FORMAT_VECTOR;
extern int DDL_OFFSET_VECTOR;
extern int DDL_IN_VECTOR;
extern int DDL_OUT_VECTOR;
extern int DDL_FILEFORMAT;
extern int DDL_DOT;
extern int DDL_GET_VECTOR;
extern int DDL_FILEFORMAT_HOST;
extern int DDL_COMM_VECTOR;
extern int DDL_GSIZE_VECTOR;
extern int DDL_COPY_VECTOR;
extern int DDL_AXPY;
extern int DDL_FILL_VECTOR;
extern int DDL_GATHER_VECTOR;
extern int DDL_SCATTER_VECTOR;
extern int DDL_BLOCK_VECTOR;

/* DDL library routine names */
#define DDL_T_CREATE_VECTOR          "DDL_Create_vector"  
#define DDL_T_FREE_VECTOR            "DDL_Free_vector"  
#define DDL_T_WRITE_VECTOR           "DDL_Write_vector"  
#define DDL_T_READ_VECTOR            "DDL_Read_vector"  
#define DDL_T_SIZE_VECTOR            "DDL_Size_vector"  
#define DDL_T_TYPE_VECTOR            "DDL_Type_vector"  
#define DDL_T_FORMAT_VECTOR          "DDL_Format_vector"  
#define DDL_T_OFFSET_VECTOR          "DDL_Offset_vector"  
#define DDL_T_IN_VECTOR              "DDL_In_vector"  
#define DDL_T_OUT_VECTOR             "DDL_Out_vector"  
#define DDL_T_FILEFORMAT             "DDL_Fileformat"  
#define DDL_T_DOT                    "DDL_Dot"
#define DDL_T_GET_VECTOR             "DDL_Get_vector"
#define DDL_T_FILEFORMAT_HOST        "DDL_Fileformat_host"  
#define DDL_T_COMM_VECTOR            "DDL_Comm_vector"  
#define DDL_T_GSIZE_VECTOR           "DDL_Gsize_vector"
#define DDL_T_COPY_VECTOR            "DDL_Copy_vector"
#define DDL_T_AXPY                   "DDL_Axpy"
#define DDL_T_FILL_VECTOR            "DDL_Fill_vector"
#define DDL_T_GATHER_VECTOR          "DDL_Gather_vector"
#define DDL_T_SCATTER_VECTOR         "DDL_Scatter_vector"
#define DDL_T_BLOCK_VECTOR           "DDL_Block_vector"
  
/* DDL constant definitions */

#define DDL_Vectortoimage(vec) vec->image
#define DDL_Image_find(ptr) ptr
#define DDL_Raw_vector(ptr) ptr->image->data

/* function prototpes */
int DDLL_Create_vector(int size, MPI_Datatype type,
		       int block, MPI_Comm comm, DDL_Vector *vec);
int DDL_Create_vector(int size, MPI_Datatype type,
		      int block, MPI_Comm comm, DDL_Object *vec);
int DDL_Free_vector(DDL_Object vec);
int DDL_Get_vector(DDL_Object vec, void *ptr);
int DDL_Comm_vector(DDL_Object vec, MPI_Comm *comm);
int DDL_Format_vector(DDL_Object vec, int *format);
int DDL_Type_vector(DDL_Object vec, MPI_Datatype *type);
int DDL_Gsize_vector(DDL_Object vec, int *size);
int DDL_Size_vector(DDL_Object vec, int *size);
int DDL_Offset_vector(DDL_Object vec, int *offset);
int DDL_Write_vector(DDL_Object vec, FILE *file);
int DDL_Write_vectorblk(FILE *file, void *data, MPI_Datatype type, int *count);
int DDL_Read_vector(DDL_Object vec, FILE *file);
int DDL_Read_vectorblk(FILE *file, void *data, MPI_Datatype type, int *count);
int DDL_Dot(DDL_Object x, DDL_Object y, void *dot);
int DDL_Scal(double alpha, DDL_Object x);
int DDL_Nrm2(DDL_Object x, void *nrm2);
int DDL_Axpy(double alpha, DDL_Object x, DDL_Object y);
int DDL_Copy_vector(DDL_Object x, DDL_Object y);
int DDL_Rand(double alpha, DDL_Object x);
int DDL_Fill_vector(DDL_Object x, double val);
int DDL_Gather_vector(DDL_Object vec, void *dest, int size, int root);
int DDL_Block_vector(DDL_Object vec, int *block);
int DDL_Check_err(DDL_Object x, DDL_Object true, double *err);

#endif
