/* $Id: ddl_generic.h,v 1.3 1995/08/29 11:28:13 tim Exp $ */

/* header file for generic DDL stuff */

/* Copyright 1994 IASC */

#ifndef _DDL_GENERIC_H
#define _DDL_GENERIC_H

/* macro to convert C procedure names to Fortran names
   so that C procedures are callable from Fortran and vice versa */
/* underscored macro for names with underscore character and using f2c? */
#ifdef HAVE_FORTRAN_UNDERSCORE
#define F77(name) name ## _
#define F77_(name) name ## _
#else
#define F77(name) name
#define F77_(name) name
#endif

#define DDL_GENERIC                  100

/* DDL error severity */
extern int DDL_ERR_HARD;               
extern int DDL_ERR_SOFT;               

/* DDL error numbers */

extern int DDL_SUCCESS;
extern int DDL_ERR;                      
extern int DDL_ERR_UNALLOCATED;          
extern int DDL_ERR_MEMORY;               
extern int DDL_ERR_SIZE;                 
extern int DDL_ERR_TYPE;                 
extern int DDL_ERR_FORMAT;               
extern int DDL_ERR_BLOCK;                
extern int DDL_ERR_COMMTOP;              
extern int DDL_ERR_COMMDIM;              
extern int DDL_ERR_USED;                 
extern int DDL_ERR_FILE;                 
extern int DDL_ERR_NOTIMPLEMENTED;       
extern int DDL_ERR_TOOSMALL;             
extern int DDL_ERR_FILEFORMAT;           
extern int DDL_ERR_ISIZE;                
extern int DDL_ERR_ITYPE;                
extern int DDL_ERR_IFORMAT;              
extern int DDL_ERR_ICOMM;                
extern int DDL_ERR_DDLTYPE;              
extern int DDL_ERR_TOOBIG;               
extern int DDL_ERR_IOBJ;                 
extern int DDL_ERR_OVER;                 
extern int DDL_ERR_DIST;                 
extern int DDL_ERR_IPROPTAG;             
extern int DDL_ERR_NOTEXIST;             
extern int DDL_ERR_NOCOMMTAG;            

/* DDL error texts */
#define DDL_ERRT_SUCCESS             "No error"
#define DDL_ERRT_ERR                 "Unspecified error"
#define DDL_ERRT_UNALLOCATED         "Unallocated object"
#define DDL_ERRT_MEMORY              "Out of memory"
#define DDL_ERRT_SIZE                "Invalid SIZE parameter"
#define DDL_ERRT_TYPE                "Invalid TYPE parameter"
#define DDL_ERRT_FORMAT              "Invalid FORMAT parameter"
#define DDL_ERRT_BLOCK               "Invalid BLOCK parameter"
#define DDL_ERRT_COMMTOP             "Invalid communicator topology"
#define DDL_ERRT_COMMDIM             "Invalid communicator dimension"
#define DDL_ERRT_USED                "Object in use"
#define DDL_ERRT_FILE                "Invalid file handle"
#define DDL_ERRT_NOTIMPLEMENTED      "Operation not implemented"
#define DDL_ERRT_TOOSMALL            "Object too small to hold data"
#define DDL_ERRT_FILEFORMAT          "Invalid file format"
#define DDL_ERRT_ISIZE               "Incompatible object size"
#define DDL_ERRT_ITYPE               "Incompatible object type"
#define DDL_ERRT_IFORMAT             "Incompatible object format"
#define DDL_ERRT_ICOMM               "Incompatible object communicator"
#define DDL_ERRT_DDLTYPE             "Invalid object type"
#define DDL_ERRT_TOOBIG              "Data space too small for object data"
#define DDL_ERRT_IOBJ                "Incompatible objects"
#define DDL_ERRT_OVER                "Invalid overlap"
#define DDL_ERRT_DIST                "Not all processes used by distribution"
#define DDL_ERRT_IPROPTAG            "Illegal property tag" 
#define DDL_ERRT_NOTEXIST            "Not exist"
#define DDL_ERRT_NOCOMMTAG           "No COMM tag"

/* generic DDL library routine numbers */

extern int DDL_FREE;
extern int DDL_READ;
extern int DDL_WRITE;
extern int DDL_SIZE;
extern int DDL_OFFSET;
extern int DDL_TYPE;
extern int DDL_FORMAT;
extern int DDL_COMM;
extern int DDL_ERROR;
extern int DDL_INIT;
extern int DDL_FINALIZE;
extern int DDL_COPY;
extern int DDL_FILL;
extern int DDL_GSIZE;
extern int DDL_SKIP_HOST;
extern int DDL_BLOCK;
extern int DDL_OBJECTTYPE;
extern int DDL_TAG;

/* generic DDL library routine names */
#define DDL_T_FREE                   "DDL_Free"  
#define DDL_T_OBJECTTYPE             "DDL_Objecttype"  
#define DDL_T_TAG                    "DDL_Tag"  
#define DDL_T_READ                   "DDL_Read"
#define DDL_T_WRITE                  "DDL_Write"
#define DDL_T_SIZE                   "DDL_Size"
#define DDL_T_OFFSET                 "DDL_Offset"
#define DDL_T_TYPE                   "DDL_Type"
#define DDL_T_FORMAT                 "DDL_Format"
#define DDL_T_COMM                   "DDL_Comm"
#define DDL_T_ERROR                  "DDL_Error"  
#define DDL_T_INIT                   "DDL_Init"  
#define DDL_T_FINALIZE               "DDL_Finalize"  
#define DDL_T_COPY                   "DDL_Copy"  
#define DDL_T_FILL                   "DDL_Fill"  
#define DDL_T_GSIZE                  "DDL_Gsize"
#define DDL_T_SKIP_HOST              "DDL_Skip_host"
#define DDL_T_BLOCK                  "DDL_Block"

/* size of blocks for communication using temporary buffers */
#define DDL_COMM_BLKSIZE             10000

/* basic data types like MPI but integers */
#define DDL_FLOAT                    1
#define DDL_DOUBLE                   2
#define DDL_INT                      3

/* null object types for initialization */
#define DDL_NULL                     NULL
#define DDL_SPARSE_NULL              NULL
#define DDL_VECTOR_NULL              NULL

/* file I/O mode constants */
#define DDL_READ_MODE                1
#define DDL_WRITE_MODE               2

/* error function */
#define DDL_Error(err_no, err_severity, ddl_routine) \
  F77(ddl_error)(&err_no, &err_severity, &ddl_routine)

/* function prototypes */

int DDL_Free(DDL_Object obj);
int DDL_Objecttype(DDL_Object obj, int *type);
int DDL_Tag(DDL_Object obj, int *tag);
int DDL_Type(DDL_Object obj, MPI_Datatype *type);
int DDL_Size(DDL_Object obj, int *size);
int DDL_Offset(DDL_Object obj, int *offset);
int DDL_Format(DDL_Object obj, int *format);
int DDL_Comm(DDL_Object obj, MPI_Comm *comm);
int DDL_Read(DDL_Object obj, FILE *file);
int DDL_Write(DDL_Object obj, FILE *file);
int DDL_Copy(DDL_Object src, DDL_Object dest);
int DDL_Fill(DDL_Object obj, double val);
/*int DDL_Print_error(int err_no, int err_severity, int ddl_routine);*/
/*int DDL_Error(int err_no, int err_severity, int ddl_routine);*/
/*int DDL_Errhard(int err_no, int err_severity, int ddl_routine);*/
/*int DDL_Errsoft(int err_no, int err_severity, int ddl_routine);*/
int F77(ddl_error)(int *err_no, int *err_severity, int *ddl_routine);
int F77(ddl_print_error)(int *err_no, int *err_severity, int *ddl_routine);
int DDL_Init(MPI_Comm oldcomm, MPI_Comm *newcomm);
int DDL_Finalize(MPI_Comm comm);
int DDL_Skip_host(FILE *file);
int DDL_Skip(FILE *file, MPI_Comm comm);
int DDL_Gsize(DDL_Object obj, int *size);
int DDL_Block(DDL_Object obj, int *block);

#endif

