/* $Id: ddl_image.h,v 1.4 1995/08/29 11:23:49 tim Exp $ */

#ifndef _DDL_IMAGE_H
#define _DDL_IMAGE_H

#include <mpi.h>
#include "ddl_general.h"

struct DDLR_Image {
  MPI_Comm comm;
  MPI_Datatype type;
  int global_size[2];
  int local_size[2]; /* data storage is actually */
		     /* (local_size[0]+2*overlap)*
			(local_size[1]+2*overlap) */ 
  int overlap, block; /* All but the last process has a multiple of */
		      /* block elements */
  int boundary[2][2];		/* Holds the overlap sizes for each direction */
  int offset[2]; /* This is the offset to the bit that you own, not to */
		 /* the overlap region */
  int packed;			/* Whether the structure is packed, */
				/* i.e., whether it contains space for */
				/* the overlaps that don't exist */
				/* because we're on the boundary */
  void *data;
  void **rows;			/* A dope vector to the rows of the data */
};

typedef struct DDLR_Image *DDL_Image;
  
extern int DDL_IMAGE_CREATE;

/* DDL library routine names */
#define DDL_T_IMAGE_CREATE           "DDL_Image_create"  


#ifndef _DEBUG_MALLOC_INC
#define malloc_enter(func)
#define malloc_leave(func)
#define malloc_chain_check()
#define malloc_dump(fd)
#define malloc_list(a,b,c)
#define malloc_inuse(hist)    (*(hist) = 0, 0)
#endif

int DDL_Image_create P((int size[2], MPI_Datatype type, int max_overlap,
		 int block, MPI_Comm comm, DDL_Image *image));
int DDL_Image_free P((DDL_Image image));
int DDL_Image_comm P((DDL_Image image, MPI_Comm *comm));
int DDL_Image_type P((DDL_Image image, MPI_Datatype *type));
int DDL_Image_global_size P((DDL_Image image, int *size));
int DDL_Image_local_size P((DDL_Image image, int *size));
int DDL_Image_offset P((DDL_Image image, int *offset));
int DDL_Image_max_overlap P((DDL_Image image, int *overlap));
int DDL_Image_rawdata P((DDL_Image image, void **data));
int DDL_Image_data P((DDL_Image image, void **data));
int DDL_Image_block P((DDL_Image im, int *block));
int DDL_Image_boundary P((DDL_Image im, int boundary[2][2]));
int DDL_Ishost();
#endif

