/* $Id: ddl_sparse.h,v 1.8 1995/08/29 11:25:32 tim Exp $ */

/* general header file for sparse DDL routines */

/* Copyright 1994 IASC */

#ifndef _DDL_SPARSE_H
#define _DDL_SPARSE_H

#include <mpi.h>
#include "ddl_general.h"
#include "ddl_image.h"

/* DDL constant definitions */

/* sparse DDL matrix types */
#define DDL_SPARSE_TYPE              300
#define DDL_ROW_FMT                  DDL_SPARSE_TYPE+      1
#define DDL_COL_FMT                  DDL_SPARSE_TYPE+      2
#define DDL_TRIAD_FMT                DDL_SPARSE_TYPE+      3
#define DDL_PCK_ROW                  DDL_SPARSE_TYPE+      4
#define DDL_PCK_COL                  DDL_SPARSE_TYPE+      5
#define DDL_PAD_ROW                  DDL_SPARSE_TYPE+      6
#define DDL_PAD_COL                  DDL_SPARSE_TYPE+      7
#define DDL_TRIAD_ROW                DDL_SPARSE_TYPE+      8
#define DDL_TRIAD_COL                DDL_SPARSE_TYPE+      9
#define DDL_TRIAD                    DDL_SPARSE_TYPE+      10

/* tag constants for global communications structure */
#define DDL_COMMTAG_MV               1 /* mv multiply */
#define DDL_COMMTAG_PX               2 /* triangular solve row perm */
#define DDL_COMMTAG_PY               3 /* triangular solve column perm */

/* matrix subtype */
#define DDL_GENERAL                  0 /* general matrix */
#define DDL_UPPER                    1 /* upper triangular */
#define DDL_LOWER                    2 /* lower triangular */

/* DDL library routine numbers */

extern int DDL_CREATE_SPARSE;
extern int DDL_FREE_GLOBAL;
extern int DDL_FREE_SPARSE;
extern int DDL_TYPE_SPARSE;
extern int DDL_SIZE_SPARSE;
extern int DDL_FORMAT_SPARSE;
extern int DDL_INIT_SPARSE;
extern int DDL_FINALIZE_SPARSE;
extern int DDL_INFO_SPARSE;
extern int DDL_INHB;
extern int DDL_IN_SPARSE;
extern int DDL_OUT_SPARSE;
extern int DDL_COLTOTRIAD;
extern int DDL_OFFSET_SPARSE;
extern int DDL_COMM_SPARSE;
extern int DDL_GSIZE_SPARSE;
extern int DDL_READ_SPARSE;
extern int DDL_USED_SPARSE;
extern int DDL_SETUSED_SPARSE;
extern int DDL_SETFORMAT_SPARSE;
extern int DDL_WRITE_SPARSE;
extern int DDL_GUSED_SPARSE;
extern int DDL_GET_IA;
extern int DDL_GET_JA;
extern int DDL_GET_A;
extern int DDL_SMV_COMM;
extern int DDL_SMV;
extern int DDL_COPY_SPARSE;

extern int DDL_TRISOLVE;
extern int DDL_SETSUBTYPE_SPARSE;
extern int DDL_SUBTYPE_SPARSE;
extern int DDL_CHECK_PERM;
extern int DDL_COMPARE_VECVEC;
extern int DDL_COMPARE_MATVEC;
extern int DDL_BLOCK_SPARSE;
extern int DDL_CREATE_PROP;
extern int DDL_FREE_PROP;
extern int DDL_FIND_PROP;
extern int DDL_FREE_ALLPROP;
extern int DDL_SETPERM_SPARSE;
extern int DDL_TRISOLVE_COMM;
extern int DDL_ADDPERM_VECTOR;
extern int DDL_PERMUTE_VECTOR;
extern int DDL_INVPERM;
extern int DDL_ILU_COMM;
extern int DDL_ILU_FORM;

/* DDL library routine names */
#define DDL_T_CREATE_SPARSE          "DDL_Create_sparse"  
#define DDL_T_FREE_GLOBAL            "DDL_Free_global"              
#define DDL_T_FREE_SPARSE            "DDL_Free_sparse"               
#define DDL_T_TYPE_SPARSE            "DDL_Type_sparse"  
#define DDL_T_SIZE_SPARSE            "DDL_Size_sparse"  
#define DDL_T_FORMAT_SPARSE          "DDL_Format_sparse"            
#define DDL_T_INIT_SPARSE            "DDL_Init_sparse"  
#define DDL_T_FINALIZE_SPARSE        "DDL_Finalize_sparse"  
#define DDL_T_INFO_SPARSE            "DDL_Info_sparse"  
#define DDL_T_INHB                   "DDL_Inhb"
#define DDL_T_IN_SPARSE              "DDL_In_sparse"
#define DDL_T_OUT_SPARSE             "DDL_Out_sparse"
#define DDL_T_COLTOTRIAD             "DDL_Coltotriad"
#define DDL_T_OFFSET_SPARSE          "DDL_Offset_sparse"
#define DDL_T_COMM_SPARSE            "DDL_Comm_sparse"
#define DDL_T_GSIZE_SPARSE           "DDL_Gsize_sparse"
#define DDL_T_READ_SPARSE            "DDL_Read_sparse"
#define DDL_T_USED_SPARSE            "DDL_Used_sparse"
#define DDL_T_SETUSED_SPARSE         "DDL_Setused_sparse"
#define DDL_T_SETFORMAT_SPARSE       "DDL_Setformat_sparse"            
#define DDL_T_WRITE_SPARSE           "DDL_Write_sparse"
#define DDL_T_GUSED_SPARSE           "DDL_Gused_sparse"
#define DDL_T_GET_IA                 "DDL_Get_ia"
#define DDL_T_GET_JA                 "DDL_Get_ja"
#define DDL_T_GET_A                  "DDL_Get_a"
#define DDL_T_SMV_COMM               "DDL_Smv_comm" 
#define DDL_T_SMV                    "DDL_Smv" 
#define DDL_T_COPY_SPARSE            "DDL_Copy_sparse" 

#define DDL_T_TRISOLVE               "DDL_Trisolve" 
#define DDL_T_SETSUBTYPE_SPARSE      "DDL_Setsubtype_sparse" 
#define DDL_T_SUBTYPE_SPARSE         "DDL_Subtype_sparse" 
#define DDL_T_CHECK_PERM             "DDL_Check_perm" 
#define DDL_T_COMPARE_VECVEC         "DDL_Compare_vecvec" 
#define DDL_T_COMPARE_MATVEC         "DDL_Compare_matvec" 
#define DDL_T_BLOCK_SPARSE           "DDL_Block_sparse" 
#define DDL_T_CREATE_PROP            "DDL_Create_prop"
#define DDL_T_FREE_PROP              "DDL_Free_prop" 
#define DDL_T_FIND_PROP              "DDL_Find_prop" 
#define DDL_T_FREE_ALLPROP           "DDL_Free_allprop" 
#define DDL_T_SETPERM_SPARSE         "DDL_Addperm_sparse" 
#define DDL_T_TRISOLVE_COMM          "DDL_Trisolve_comm" 
#define DDL_T_ADDPERM_VECTOR         "DDL_Addperm_vector" 
#define DDL_T_PERMUTE_VECTOR         "DDL_Permute_vector" 
#define DDL_T_INVPERM                "DDL_Invperm" 
#define DDL_T_ILU_COMM               "DDL_ILU_comm" 
#define DDL_T_ILU_FORM               "DDL_ILU_form" 


/* function prototypes */
int DDLL_Create_sparse(int size[3], MPI_Datatype type, int format,
		       int block, MPI_Comm comm, DDL_Sparse *sparse);
int DDL_Create_sparse(int size[3], MPI_Datatype type, int format,
		      int block, MPI_Comm comm, DDL_Object *obj);
int DDL_Free_sparse(DDL_Object obj);
int DDL_Get_ia(DDL_Object obj, int **ia);
int DDL_Get_ja(DDL_Object obj, int **ja);
int DDL_Get_a(DDL_Object obj, void **a);
int DDL_Type_sparse(DDL_Object obj, MPI_Datatype *type);
int DDL_Size_sparse(DDL_Object obj, int size[]);
int DDL_Offset_sparse(DDL_Object obj, int *offset);
int DDL_Format_sparse(DDL_Object obj, int *format);
int DDL_Setformat_sparse(DDL_Object obj, int format);
int DDL_Comm_sparse(DDL_Object obj, MPI_Comm *comm);
int DDL_Used_sparse(DDL_Object obj, int *nelt);
int DDL_Gused_sparse(DDL_Object obj, int *nelt);
int DDL_Setused_sparse(DDL_Object obj, int nelt);
int DDL_Gsize_sparse(DDL_Object obj, int size[]);
int DDL_Init_sparse(MPI_Comm oldcomm, MPI_Comm *newcomm);
int DDL_Finalize_sparse(MPI_Comm comm);
int DDL_Info_sparse(DDL_Object obj); 
int DDL_Check_comm(MPI_Comm comm);
int DDL_Read_sparse(DDL_Object obj, FILE *file);
int DDL_Read_sparseline(FILE *file, int nelt, int ia[], int ja[], double a[],
			MPI_Datatype type, int format, int *count, int *total, 
			int *ia_last, int *ja_last, double *a_last);
int DDL_Write_sparse(DDL_Object obj, FILE *file);
int DDL_Write_sparseline(FILE *file, int ia[], int ja[], double a[],
			 MPI_Datatype type, int *count);
int DDL_Copy_sparse(DDL_Object src, DDL_Object dest);
int DDL_Alltoallvs(void *snd_buf, int snd_tot_proc, int snd[], 
		   int snd_cnt[],
		   int snd_ptr[], MPI_Datatype snd_type,
		   void *rcv_buf, int rcv_tot_proc, int rcv[], 
		   int rcv_cnt[],
		   int rcv_ptr[], MPI_Datatype rcv_type,
		   MPI_Comm comm);
int DDL_Bcasts(void *buf, int cnt, MPI_Datatype type, 
	       int root, int nproc, int *proc, MPI_Comm comm);
int DDL_Smv_seq(int n, int m, int ia[],int ja[], 
		void *a, MPI_Datatype type,
		int nelt, int format, void *x, void *y,
		double alpha, double beta);
int DDL_Coeff_ind(int row, int ia[], int *low, int *high);

int DDL_Compare_vecvec(DDL_Object x, DDL_Object y, int *ret);
int DDL_Compare_matvec(DDL_Object x, DDL_Object y, int *ret);
int DDL_Compare_vecperm(DDL_Object x, DDL_Object p, int *ret);
int DDL_Compare_matpermx(DDL_Object x, DDL_Object p, int *ret);
int DDL_Compare_matpermy(DDL_Object x, DDL_Object p, int *ret);
int DDL_Setsubtype_sparse(DDL_Object obj, int subtype);
int DDL_Subtype_sparse(DDL_Object obj, int *subtype);
int DDL_Check_perm(DDL_Object p, int *ret);
int DDL_Create_prop(DDL_Object obj, int tag, DDL_Prop *prop);
int DDLL_Free_prop(DDL_Prop p);
int DDL_Free_prop(DDL_Object obj, int tag);
int DDL_Free_allprop(DDL_Object obj);
int DDL_Find_prop(DDL_Object obj, int tag, DDL_Prop *p);
int DDL_Setperm_sparse(DDL_Object obj, DDL_Object px_ddl,
		    DDL_Object py_ddl);
int DDL_Smv_comm(DDL_Object sp);
int DDL_Smv(DDL_Object sp, DDL_Object x, DDL_Object y,
	    double alpha, double beta);
int DDL_Trisolve_comm(DDL_Object a);
int DDL_Trisolve(DDL_Object sp, DDL_Object x, DDL_Object b);
int DDL_Sparse_nzcol(int *ia, int *ja, int n, int nc, 
		     int *ind, int *nz);
int DDL_Packcomm(int n, int block,
		 int dest, int nproc, int *tot_proc, int *tot_data, 
		 int *src, int *ptr, int *ind);
int DDL_Excomm(int nproc, MPI_Comm comm, 
	       int *rcv_tot_proc, int *snd_tot_proc, 
	       int *rcv_tot_data, int *snd_tot_data, 
	       int rcv[], int snd[], 
	       int rcv_ptr[], int snd_ptr[],
	       int rcv_cnt[], int snd_cnt[], 
	       int rcv_ind[], int snd_ind[]);
int DDL_Col(int *ia, int *ja, int n, int nc, int block, int *pv,
	    int proc, int nproc, MPI_Comm comm, 
	    int *rcv_tot_proc, int *snd_tot_proc, 
	    int *rcv_tot_data, int *snd_tot_data, 
	    int rcv[], int snd[], 
	    int rcv_ptr[], int snd_ptr[], 
	    int rcv_cnt[], int snd_cnt[],
	    int rcv_ind[], int snd_ind[],
	    int rcv_pind[]);
int DDL_Row(int *ia, int *ja, int n, int nr, int block, int *pv,
	    int proc, int nproc, MPI_Comm comm, 
	    int *rcv_tot_proc, int *snd_tot_proc, 
	    int *rcv_tot_data, int *snd_tot_data, 
	    int rcv[], int snd[], 
	    int rcv_ptr[], int snd_ptr[], 
	    int rcv_cnt[], int snd_cnt[],
	    int rcv_ind[], int snd_ind[],
	    int rcv_pind[]);
int DDL_Calc_invperm(int *pv, int *ipv, int n);
int DDL_calc_perm(int nx, int ny, int nz, int *pv);
int DDL_Lpermute_vector(void *v, MPI_Datatype type, int *pv, int n);
int DDL_cat_plane(int **pv, int base, int plane, int nx, int ny); 
int DDL_cat_part(int **pv, int base, int odd, int nx, int range); 
int DDL_cmp(const int *x, const int *y);
int DDL_qs1i(int *ia, int n, int kflag);
int DDL_qs2i1d(int *ia, int *ja, double *a, int n, int kflag);
int DDL_qs2i(int *ia, int *ja, int n, int kflag);
int DDL_Block_sparse(DDL_Object sp, int *block);
int DDL_Create_dummy(DDL_Object *d);
int DDL_Free_dummy(DDL_Object d);
int DDL_Truncate(void *ptr, int n);
int DDL_Lpermute_sparse(int *ia, int *ja, void *a, int nelt, 
			int n, int m, MPI_Datatype type, int format,
			int *px, int *py);
int DDL_Msend(void *buf, int cnt, MPI_Datatype type, 
	      int nproc, int *proc, int tag, MPI_Comm comm);
int DDL_Setperm_vector(DDL_Object v_ddl, DDL_Object pv_ddl);
int DDL_Invperm(DDL_Object pv_ddl, DDL_Object ipv_ddl);

int DDL_ILU_comm(DDL_Object a);
int DDL_ILU_form(DDL_Object sp, DDL_Object l,
                 DDL_Object u, DDL_Object diag_blk);


#endif

