/* $Id: ddl_struct.h,v 1.2 1995/08/21 10:34:38 tim Exp $ */

/* general header file for sparse DDL routines */

/* Copyright 1994 IASC */

#ifndef _DDL_STRUCT_H
#define _DDL_STRUCT_H

#include <mpi.h>
#include "ddl_general.h"
#include "ddl_image.h"

/* object property list tags */

#define DDL_PROP_BASE                0
#define DDL_A                        DDL_PROP_BASE+1                   
#define DDL_IA                       DDL_PROP_BASE+2
#define DDL_JA                       DDL_PROP_BASE+3
#define DDL_PX                       DDL_PROP_BASE+4
#define DDL_PY                       DDL_PROP_BASE+5                   
#define DDL_COMM_PX                  DDL_PROP_BASE+6                    
#define DDL_COMM_PY                  DDL_PROP_BASE+7                    
#define DDL_COMM_IND                 DDL_PROP_BASE+8
#define DDL_PV                       DDL_PROP_BASE+9
#define DDL_IPV                      DDL_PROP_BASE+10
#define DDL_COMM_PV                  DDL_PROP_BASE+11
#define DDL_COMM_IPV                 DDL_PROP_BASE+12
#define DDL_IPX                      DDL_PROP_BASE+13
#define DDL_IPY                      DDL_PROP_BASE+14                  
#define DDL_VECTOR                   DDL_PROP_BASE+15                  
#define DDL_SPARSE                   DDL_PROP_BASE+16                  
#define DDL_DUMMY                    DDL_PROP_BASE+17

/* object property list types */

#define DDL_PROP_TYPE                0
#define DDL_PROP_VECTOR              DDL_PROP_TYPE+1                   
#define DDL_PROP_SPARSE              DDL_PROP_TYPE+2
#define DDL_PROP_GLOB                DDL_PROP_TYPE+3
/*#define DDL_PROP_PERM                DDL_PROP_TYPE+4*/
#define DDL_PROP_IND                 DDL_PROP_TYPE+5
#define DDL_PROP_DUMMY               DDL_PROP_TYPE+6

/* property list structures */

/* new global exchange data strctures */

struct DDLR_Sndglob_comm {
  int nproc;          /* number of processes to communicate with */
  int *proc;          /* vector of process numbers */
  int *ptr;           /* C pointer to start of index block for each process */
  int *cnt;           /* vector of counts for each process */
  int *ind;           /* vector of C indices on each process */
/*  int *pind; */          /* vector of permuted C indices on each process */
};

struct DDLR_Rcvglob_comm {
  int nproc;          /* number of processes to communicate with */
  int *proc;          /* vector of process numbers */
  int *ptr;           /* C pointer to start of index block for each process */
  int *cnt;           /* vector of counts for each process */
  int *ind;           /* vector of C indices on each process */
  int *pind;          /* vector of permuted C indices on each process */
};

struct DDLR_Glob {
  struct DDLR_Sndglob_comm snd; /* send data structure */
  struct DDLR_Rcvglob_comm rcv; /* receive data structure */
};

typedef struct DDLR_Glob *DDL_Glob;

/* global exchange data structure ordered by indices;
   for use in triangular solve etc. */
struct DDLR_Sndind_comm {
  int nind;           /* number of indices to send */
  int *ind;           /* vector of C indices to send */
/*  int *pind;*/          /* vector of permuted (global) C indices to send */
  int *ptr;           /* C pointer to start of process block for each index */
  int *cnt;           /* vector of counts for each index */
  int *proc;          /* vector of process numbers to send index to */
};

struct DDLR_Rcvind_comm {
  int nind;           /* number of indices to receive */
  int *ind;           /* vector of C indices to receive */
  int *pind;          /* vector of permuted C indices to receive */
  int *proc;          /* vector of process numbers to receive index from */

};

struct DDLR_Ind {
  struct DDLR_Sndind_comm snd; /* send data structure */
  struct DDLR_Rcvind_comm rcv; /* receive data structure */
};

typedef struct DDLR_Ind *DDL_Ind;

/* permutation vector structure */

struct DDLR_Vector; /* forward declaration */

/*struct DDLR_Perm {
  struct DDLR_Vector *pv;    
  struct DDLR_Vector *ipv;   
};

typedef struct DDLR_Perm *DDL_Perm;*/

struct DDLR_Sparse; /* forward declaration */

struct DDLR_Prop {
  int type; /* property type eg. DDL_Vector, DDL_Global etc. */
  int tag; /* name giving use of structure eg. mv, trisolve, etc. */
  struct DDLR_Prop *prop; /* pointer to next property */
  void *data; /* pointer to data structure */
};

typedef struct DDLR_Prop *DDL_Prop;

/* alternative object names for user programs */
typedef DDL_Prop DDL_Object;
typedef DDL_Object DDL_Vec;
typedef DDL_Object DDL_Spa;

/* dense vector structure */
struct DDLR_Vector {
  int format;            
  int block;
  struct DDLR_Image *image;
};
typedef struct DDLR_Vector *DDL_Vector;
  
/* standard sparse matrix data structure */
/* using C indexing from 0 */
struct DDLR_Sparse {
  MPI_Comm comm;
  MPI_Datatype type;
  int subtype;            /* eg. upper/lower triangular */ 
  int format;            
  int n;                  /* size[0] of create call */
  int mxnelt;             /* size[2] of create call */
  int size[2];            /* 0: number of rows on this process
			     1: local size of ia, ja, a vectors */
  int offset;             /* first row number on this process */
  int block;              /* All but the last process has a multiple of */
		          /* block rows/columns */
  int nelt;               /* number of used elements on this process */
  DDL_Object a;           /* pointer to sparse values */
  DDL_Object ia;          /* pointer to indices */
  DDL_Object ja;          /* pointer to indices */
};

typedef struct DDLR_Sparse *DDL_Sparse;

#endif

