/* $Id: ddl_support.h,v 1.4 1995/08/29 11:24:52 tim Exp $ */

/* support routine header file for DDL routines */

/* Copyright 1994 IASC */

#ifndef _DDL_SUPPORT_H
#define _DDL_SUPPORT_H

#include <mpi.h>
#include "ddl_general.h"
#include "ddl_image.h"
#include "ddl_sparse.h"

#define DDL_SUPPORT                  400
 
/* DDL library routine numbers */

extern int DDL_OPEN_HOST;
extern int DDL_CLOSE_HOST;
extern int DDL_SKIP_HOST_SPARSE;
extern int DDL_SKIP_HOST_VECTOR;

/* DDL library routine names */
#define DDL_T_OPEN_HOST              "DDL_Open_host"  
#define DDL_T_CLOSE_HOST             "DDL_Close_host"  
#define DDL_T_SKIP_HOST_SPARSE       "DDL_Skip_host_sparse"  
#define DDL_T_SKIP_HOST_VECTOR       "DDL_Skip_host_vector"  

/* function prototypes */
int DDL_Inhb(FILE *file, int *n, int *nelt, int ia[], int ja[], double a[],
	     int *isym, double soln[], double rhs[], double xinit[],
	     int *job);
int DDL_Out_sparse(FILE *file, int n, int nelt, 
		   int ia[], int ja[], double a[], MPI_Datatype type,
		   int isym, int format);
int DDL_In_sparse(FILE *file, int *n, int *nelt, 
		  int ia[], int ja[], double a[], MPI_Datatype *type,
		  int *isym, int *format);
int DDL_Out_vector(FILE *file, void *vec, MPI_Datatype type, int *n);
int DDL_In_vector(FILE *file, void *vec, MPI_Datatype *type, int *n);
int DDL_Open_host(FILE **file, char *fname, int mode);
int DDL_Open(FILE **file, char *fname, int mode, MPI_Comm comm);
int DDL_Close_host(FILE *file);
int DDL_Close(FILE *file, MPI_Comm comm);
int DDL_Fileformat_host(FILE *file, int *format, MPI_Datatype *type, 
			int a[]);
int DDL_Fileformat(FILE *file, int *format, MPI_Datatype *type, 
		   int a[], MPI_Comm comm);
int DDL_Coltotriad(int n, int nelt, 
		   int ia[], int ja[], void *a, MPI_Datatype type, int isym);
int DDL_Rowtotriad(int n, int nelt, 
		   int ia[], int ja[], void *a, MPI_Datatype type, int isym);
int DDL_Triadtocol(int n, int nelt, 
		  int ia[], int ja[], double a[], MPI_Datatype type, int isym);
int DDL_Triadtorow(int n, int nelt, 
		   int ia[], int ja[], void *a, MPI_Datatype type, int isym);
int DDL_DDLtype(MPI_Datatype MPItype, int *DDLtype);
int DDL_MPItype(int DDLtype, MPI_Datatype *MPItype);
int DDL_Skip_host_sparse(FILE *file);
int DDL_Skip_host_vector(FILE *file);
int DDL_Find_offset(int n, int block, int nproc, int proc);
int DDL_Find_size(int n, int block, int nproc, int proc);
int DDL_Find_proc(int n, int block, int nproc);
double DDL_Gettime();
#endif
