c     ILU preconditioner and solver

c     Copyright IASC 1994

c     ILU preconditioner
      subroutine DDL_Ilu_block(a_ddl, l_ddl, u_ddl, dinv_ddl,
     &     iwork, iwork_dim, rwork, rwork_dim, ierr)
      implicit none
      integer a_ddl, u_ddl, l_ddl, dinv_ddl, ierr
      
c     declare access procedure to be external
      external ilu_block_work

c     local variables
      integer size(3), offset
      integer ia, ja, a
      integer iu, ju, u
      integer il, jl, l
      integer dinv
      integer neltl, neltu
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)
      
      call DDL_Size_sparse(a_ddl, size, ierr)
      call DDL_Offset_sparse(a_ddl, offset, ierr)
      call DDL_Get_ia(a_ddl, ia, ierr)
      call DDL_Get_ja(a_ddl, ja, ierr)
      call DDL_Get_a(a_ddl, a, ierr)
      call DDL_Get_ia(u_ddl, iu, ierr)
      call DDL_Get_ja(u_ddl, ju, ierr)
      call DDL_Get_a(u_ddl, u, ierr)
      call DDL_Get_ia(l_ddl, il, ierr)
      call DDL_Get_ja(l_ddl, jl, ierr)
      call DDL_Get_a(l_ddl, l, ierr)
      call DDL_Get_vector(dinv_ddl, dinv, ierr)

c     call data access procedure
      call DDL_Access(ilu_block_work, 10, ia,ja,a, il,jl,l, iu,ju,u,
     &     dinv, 7, a_ddl, neltu, neltl, 
     &     iwork, iwork_dim, rwork, rwork_dim, ierr)

c     update nelt property of l_ddl, u_ddl
      call DDL_Setused_sparse(u_ddl, neltu, ierr)
      call DDL_Setused_sparse(l_ddl, neltl, ierr)

      end

c     data access procedure
      subroutine ilu_block_work(ia,ja,a, il,jl,l, iu,ju,u, dinv,
     &     a_ddl, neltu, neltl, iwork, iwork_dim, rwork, rwork_dim, 
     &     ierr)
      implicit none
      integer ia(*), ja(*)
      double precision a(*)
      integer iu(*), ju(*)
      double precision u(*)
      integer il(*), jl(*)
      double precision l(*)
      double precision dinv(*)
      integer a_ddl, neltu, neltl, ierr
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)

c     local variables
      integer iptr, aptr, jptr, nrptr, ncptr
      integer blocksize, sizes(3), offset, nelt
      integer nzblk, isym
      integer icol, jbgn, jend, j
      integer maxn, mxnelt

      call DDL_Size_sparse(a_ddl, sizes, ierr)
      blocksize = sizes(1)
      call DDL_Offset_sparse(a_ddl, offset, ierr)
      call DDL_Gsize_sparse(a_ddl, sizes, ierr)
      nelt = sizes(3)

c     Non symmetric matrix
      isym = 0

c     sort out some workspace parameters
      iptr = 1
      aptr = 1
      jptr = iptr+nelt
      nrptr = jptr+nelt
      ncptr = nrptr+blocksize      

      call sparse_row_to_block(blocksize,
     &     offset,offset+blocksize-1,ia,ja,a,iwork(iptr),
     &     iwork(jptr),rwork(aptr),nzblk)

c     change from SLAP triad format to SLAP column format
      call ds2y(blocksize,nzblk,iwork(iptr),iwork(jptr),
     &     rwork(aptr),isym)

c     zero element counts
      neltl=0
      neltu=0
      do icol = 1,blocksize
        jbgn = iwork((jptr-1)+icol)+1
        jend = iwork(jptr+icol)-1
        if (jbgn.le.jend) then
          do j = jbgn,jend
            if (iwork((iptr-1)+j).gt.icol) then
              neltl = neltl+1
              if (isym.ne.0) neltu=neltu+1
            else
              neltu=neltu+1
            endif
          end do
        end if
      end do

C     Compute the Incomplete lu decomposition.
      call dsilus( blocksize,nzblk,iwork(iptr),iwork(jptr),
     &    rwork(aptr),isym,neltl,il,jl,l,dinv,neltu,
     &    iu,ju,u,iwork(nrptr),iwork(ncptr))

      end

c     ILU solver
      subroutine DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, b_ddl, 
     &                         x_ddl, ierr)
      implicit none
      integer l_ddl, u_ddl, dinv_ddl, b_ddl, x_ddl
      integer ierr

c     declare access procedure to be external
      external ilu_solve_work

c     local variables
      integer il,jl,l, iu,ju,u, dinv, b, x
      integer size(3)

      call DDL_Size_sparse(l_ddl, size, ierr)
      call DDL_Get_ia(u_ddl, iu, ierr)
      call DDL_Get_ja(u_ddl, ju, ierr)
      call DDL_Get_a(u_ddl, u, ierr)
      call DDL_Get_ia(l_ddl, il, ierr)
      call DDL_Get_ja(l_ddl, jl, ierr)
      call DDL_Get_a(l_ddl, l, ierr)
      call DDL_Get_vector(dinv_ddl, dinv, ierr)
      call DDL_Get_vector(b_ddl, b, ierr)
      call DDL_Get_vector(x_ddl, x, ierr)

c     call data access procedure
      call DDL_Access(ilu_solve_work, 9, il,jl,l, iu,ju,u, dinv, b, x,
     &     1, size(1), ierr)

      end

c     data access procedure
      subroutine ilu_solve_work(il,jl,l, iu,ju,u, dinv, b, x,
     &     size, ierr)
      implicit none
      integer iu(*), ju(*)
      double precision u(*)
      integer il(*), jl(*)
      double precision l(*)
      double precision dinv(*), b(*), x(*)
      integer size, ierr

      call dslui2(size,b,x,il,jl,l,dinv,iu,ju,u)

      end
