c     ILU NC (?) solver

c     Copyright IASC 1994

c     ILU block factorization
      subroutine DDL_Ilunc_block(a_ddl, d_ddl, ivec_ddl, 
     &     nc, expinv, iwork, iwork_dim, rwork, rwork_dim, ierr)
      implicit none
      integer a_ddl, d_ddl, ivec_ddl, ierr
      integer nc
      logical expinv
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)

c     declare access procedure to be external
      external ilunc_block_work

c     local variables
      integer ia, ja, a
      integer d, ivec
     
      call DDL_Get_ia(a_ddl, ia, ierr)
      call DDL_Get_ja(a_ddl, ja, ierr)
      call DDL_Get_a(a_ddl, a, ierr)
      call DDL_Get_vector(d_ddl, d, ierr)
      call DDL_Get_vector(ivec_ddl, ivec, ierr)

c     call data access procedure
      call DDL_Access(ilunc_block_work, 5, ia,ja,a, d, ivec,
     &                7, a_ddl, nc, expinv, 
     &                iwork, iwork_dim, rwork, rwork_dim, ierr)

      end

c     data access procedure
      subroutine ilunc_block_work(ia,ja,a, d, ivec, a_ddl,
     &     nc, expinv, iwork, iwork_dim, rwork, rwork_dim, ierr)
      implicit none
      include 'mpif.h'
      include 'ddlf.h'
      integer a_ddl
      integer ia(*), ja(*)
      double precision a(*)
      integer ivec(*)
      double precision d(*)
      integer nc, ierr
      logical expinv
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)

c     local variables
      integer iptr, aptr, jptr, nrptr, ncptr
      integer blocksize, sizes(3), offset
      integer nzblk, isym, nelt
      integer icol, jbgn, jend, j
      integer den_l,den_u,iptrblk,list,numblks,dblk,ju,jl

      integer iprob,nzdblk

      integer i, nproc, comm

c     error checking
      integer memory_int, memory_real
      integer verr(DDL_MAXP)

c     Error Handling
c      integer ierr
      logical fatal_error
      character*(60) err_mess(0:3),err_outfile
      err_mess(0)='All went well - continue solve for slave '
      err_mess(1)='nc does not divide evenly into blocksize on slave '
      err_mess(2)='Insufficient storage allocation for RWORK or IWORK'
      err_mess(2)=err_mess(2)(1:50)//' on slave '
      err_mess(3)='Incomplete factorization broke down on slave '
      err_outfile='ilu_trace.lis'
      fatal_error=.FALSE.
      if (DDL_Ishost()) then
         open(unit=14, file=err_outfile)
      end if

      call DDL_Size_sparse(a_ddl, sizes, ierr)
      blocksize = sizes(1)
      call DDL_Offset_sparse(a_ddl, offset, ierr)
      call DDL_Used_sparse(a_ddl, nelt, ierr)

      ierr=0

c     Set value of iprob = 0
      iprob=0
c     Non symmetric matrix
      isym = 0

      numblks= blocksize/nc

c     sort out some workspace parameters
      iptr = 1
      aptr = 1
      jptr = iptr+nelt
      nrptr = jptr+nelt
      ncptr = nrptr+blocksize

c     check problem divides into blocks OK
      if (mod(blocksize,nc) .ne. 0) then
          ierr=1
          goto 1000
      end if

      call sparse_row_to_block(blocksize,
     & offset, offset+blocksize-1, ia, ja, a, iwork(iptr),
     & iwork(jptr), rwork(aptr), nzblk)

c     change from SLAP triad format to SLAP column format
      call ds2y(blocksize,nzblk,iwork(iptr),iwork(jptr),
     &     rwork(aptr),isym)

      dblk = 1
      den_l = dblk+(nc*blocksize)+nc
      den_u = den_l + nc*nc
      iptrblk = 11
      if (expinv) then
        list = iptrblk + nc
      else
        list = iptrblk + blocksize
      end if
      jl = list+2*numblks
      ju = jl+blocksize

      ivec(1)=numblks
      ivec(2)=dblk
      ivec(3)=jl
      ivec(4)=ju
      if (expinv) then
        call formilud(blocksize,nzblk,iwork(iptr),iwork(jptr),
     &    rwork(aptr),nc,d(dblk),ivec(jl),ivec(ju),ivec(list),
     &    ivec(iptrblk),d(den_l),d(den_u))
      else 
         call formilu(blocksize,nzblk,iwork(iptr),iwork(jptr),
     &    rwork(aptr),nc,d(dblk),ivec(jl),ivec(ju),ivec(list),
     &    ivec(iptrblk),d(den_l),d(den_u),iprob)
      end if

      if (iprob .gt. 0) then
         ierr=3
         goto 1000
      end if

1000  call DDL_Comm_sparse(a_ddl, comm, ierr)
      call MPI_Comm_size(comm, nproc, ierr)

c     gather error information from each process to host
      call MPI_Gather(ierr, 1, MPI_INTEGER, verr, 1, MPI_INTEGER, 
     &     0, comm, ierr)
      if (DDL_Ishost()) then
         do i=1, nproc
            write(14,*)'   slave ',i,' ierr  = ',verr(i)
            write(14,*)
            write(14,*)err_mess(verr(i)),i
            if (verr(i) .gt. 0) then
               fatal_error = .TRUE.
            end if
         end do
         close(14)
      end if

c     return final error to caller
c      ierr = 
      end

c     ILU NC solver
      subroutine DDL_Ilunc_solve(a_ddl, d_ddl, ivec_ddl,
     &     b_ddl, x_ddl, nc, expinv,
     &     iwork, iwork_dim, rwork, rwork_dim, ierr)
      implicit none
      integer a_ddl, d_ddl, ivec_ddl, x_ddl, b_ddl, ierr
      integer nc
      logical expinv
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)

c     declare access procedure to be external
      external ilunc_solve_work

c     local variables
      integer ia, ja, a, d, ivec, b, x

      call DDL_Get_ia(a_ddl, ia, ierr)
      call DDL_Get_ja(a_ddl, ja, ierr)
      call DDL_Get_a(a_ddl, a, ierr)
      call DDL_Get_vector(d_ddl, d, ierr)
      call DDL_Get_vector(ivec_ddl, ivec, ierr)
      call DDL_Get_vector(b_ddl, b, ierr)
      call DDL_Get_vector(x_ddl, x, ierr)

c     call data access procedure
      call DDL_Access(ilunc_solve_work, 7, ia, ja, a, d, ivec, x, b,
     &     7, a_ddl, nc, expinv, 
     &     iwork, iwork_dim, rwork, rwork_dim, ierr)

      end

c     data access procedure
      subroutine ilunc_solve_work(ia,ja,a, d, ivec, x, b,
     &     a_ddl, nc, expinv, iwork, iwork_dim, rwork, rwork_dim, ierr)
      implicit none
      integer ia(*), ja(*)
      double precision a(*)
      double precision b(*), x(*)
      integer ivec(*)
      double precision d(*)
      integer a_ddl, nc, ierr
      logical expinv
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      integer rwork(rwork_dim)

      double precision my_first_coeff
      integer nzblk,my_first_row,my_first_column,isym
      integer nwork,nwork1,blocksize,solwork,iptr,jptr,aptr
      integer j,icol,jbgn,jend,nrptr,ncptr
      integer den_l,den_u,iptrblk,list,numblks,dblk,ju,jl
      integer sizes(3), nelt, offset
      integer maxn, mxnelt, maxiw, maxrw
      integer memory_int, memory_real

      call DDL_Size_sparse(a_ddl, sizes, ierr)
      blocksize = sizes(1)
      call DDL_Offset_sparse(a_ddl, offset, ierr)
      call DDL_Used_sparse(a_ddl, nelt, ierr)

c     Non symmetric matrix
      isym = 0

      numblks= blocksize/nc

c     sort out some workspace parameters
      iptr = 1
      aptr = 1
      jptr = iptr+nelt
      nrptr = jptr+nelt
      ncptr = nrptr+blocksize

c      call sparse_row_to_block(blocksize,
c     &     offset, offset+blocksize-1,ia, ja, a, iwork(iptr),
c     &     iwork(jptr), rwork(aptr), nzblk)

c     change from SLAP triad format to SLAP column format
c      call ds2y(blocksize,nzblk,iwork(iptr),iwork(jptr),
c     &     rwork(aptr),isym)

      if (expinv) then
       call iludsol(blocksize,b,x,nzblk,iwork(iptr),iwork(jptr),
     &        rwork(aptr),isym,d,ivec)
      else
       call ilusol(blocksize,b,x,nzblk,iwork(iptr),iwork(jptr),
     &        rwork(aptr),isym,d,ivec)
      end if

      end

