c     DDL procedure to solve sparse linear system of equations 
      subroutine DDL_Linsolve(a_ddl, x_ddl, b_ddl, itmax, iter, 
     &     tol, tolno, precon, nc, expinv, error, 
     &     iwork, iwork_dim, rwork, rwork_dim, ierr)

      implicit none
      include 'mpif.h'
      include 'ddlf.h'

c     DDL object handles
      integer a_ddl, x_ddl, b_ddl
      
      integer itmax, iter, tolno, precon, nc, ierr
      double precision tol(tolno), error
      logical expinv
      integer iwork_dim
      integer iwork(iwork_dim)
      integer rwork_dim
      double precision rwork(rwork_dim)

c     local variables
c     tfqmrnc variables
      integer i,n,m
      double precision alpha,theta,betan,
     + tau,omega,rhon,rhonm1,const,xnrm,c,sigma,malpha,oldcgs,
     + rnrm,rnrm0,eta,xtnrm,bnrm,gamma,delta
      integer nelt, tolcount
      logical ilu,ilunc

c     DDL/MPI 
      integer infile, outfile
      integer block, format
      integer comm
      integer m_ddl, l_ddl, u_ddl, dinv_ddl
      integer pip_ddl, rcgs_ddl, r0_ddl
      integer p_ddl, v_ddl, d_ddl, tmp_ddl, q_ddl, w_ddl, z_ddl
      integer uv_ddl, rvec_ddl, ivec_ddl
      integer type, fformat, sizes(3), len
      integer numblks, blocksize, lnelt, memory_int, memory_real

c     check preconditioning type
      if (precon .eq. 0) then
         ilu = .FALSE.
         ilunc = .FALSE.
      else if (precon .eq. 1) then
         ilu = .TRUE.
         ilunc = .FALSE.
      else if (precon .eq. 2) then
         ilu = .FALSE.
         ilunc = .TRUE.
      end if

c     find problem size
      call DDL_Gsize(a_ddl, sizes, ierr) 
      n = sizes(1)
      nelt = sizes(3)
      call DDL_Type(a_ddl, type, ierr)
      call DDL_Format(a_ddl, format, ierr)
      call DDL_Comm(a_ddl, comm, ierr)
      call DDL_Block(a_ddl, block, ierr)

c     check enough memory is allocated
      if (ilunc) then
         call DDL_Size(a_ddl, sizes, ierr)
         blocksize = sizes(1)
         call DDL_Used_sparse(a_ddl, lnelt)
         numblks = blocksize/nc
         memory_int = 10+nc+2*(numblks+blocksize)+ lnelt+blocksize+1
         memory_real = 8*blocksize+nc*(blocksize+3*nc)+ lnelt
         if (memory_int.gt.iwork_dim.or.memory_real.gt.rwork_dim) then
            ierr=2
            goto 1000
         end if
      end if

      if (ilu) then
         sizes(1) = n
         sizes(2) = n
         sizes(3) = nelt*2
         l_ddl = DDL_NULL
         call DDL_Create_sparse(sizes, type, format, 
     &                          block, comm, l_ddl, ierr)
         u_ddl = DDL_NULL
         call DDL_Create_sparse(sizes, type, format, 
     &                          block, comm, u_ddl, ierr)
      end if

      if (ilu) then
         dinv_ddl = DDL_NULL
         call DDL_Create_vector(n, type, block, comm, dinv_ddl, ierr)
c        factorize A to block preconditioner M
         call DDL_Ilu_block(a_ddl, l_ddl, u_ddl, dinv_ddl,
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
      end if

      if (ilunc) then
c        create integer vector for ilunc calls
         ivec_ddl = DDL_NULL
         call DDL_Create_vector(10*n, MPI_INTEGER, block, 
     &                          comm, ivec_ddl, ierr)
         rvec_ddl = DDL_NULL
         call DDL_Create_vector(10*n, MPI_DOUBLE_PRECISION, block, 
     &                          comm, rvec_ddl, ierr)
c        factorize A to block preconditioner ILUnc
         call DDL_Ilunc_block(a_ddl, rvec_ddl, ivec_ddl, nc, expinv, 
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
      end if

c     create vector for residual
      rcgs_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, rcgs_ddl, ierr)
c     create vector for initial residual
      r0_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, r0_ddl, ierr)
c     create vectors for tfqmr soln method
      uv_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, uv_ddl, ierr)
      p_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, p_ddl, ierr)
      v_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, v_ddl, ierr)
      d_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, d_ddl, ierr)
      tmp_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, tmp_ddl, ierr)
      q_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, q_ddl, ierr)
      w_ddl = DDL_NULL
      call DDL_Create_vector(n, type, block, 
     &                       comm, w_ddl, ierr)
      if(ilu .or. ilunc) then
         z_ddl = DDL_NULL
         call DDL_Create_vector(n, type, block, 
     &                          comm, z_ddl, ierr)
      end if

      gamma = 1.0d0
      delta = 0.0d0
c     calculate mv communications
      call DDL_Smv_comm(a_ddl, ierr)

      call DDL_Nrm2(b_ddl, bnrm, ierr)

      call DDL_Copy(b_ddl, r0_ddl, ierr)
      call DDL_Smv(a_ddl, x_ddl, r0_ddl, -1.0d0, gamma, ierr)     

      call DDL_Nrm2(x_ddl, xnrm, ierr)

      if (ilu) then
         call DDL_Copy(r0_ddl, tmp_ddl, ierr)
         call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl, 
     &                      r0_ddl, ierr)
         call DDL_Copy(b_ddl, tmp_ddl, ierr)
         call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl, 
     &                      b_ddl, ierr)
      else if (ilunc) then
         call DDL_Copy(r0_ddl, tmp_ddl, ierr)
         call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &        r0_ddl, nc, expinv,
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
         call DDL_Copy(b_ddl, tmp_ddl, ierr)
         call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &        b_ddl, nc, expinv, 
     &        iwork, iwork_dim, rwork, rwork_dim, ierr)
      end if

      call DDL_Copy(r0_ddl, p_ddl, ierr)
      call DDL_Copy(r0_ddl, uv_ddl, ierr)
      call DDL_Copy(r0_ddl, rcgs_ddl, ierr)
      call DDL_Scal(0.0d0, d_ddl, ierr)

c     initialise v to Ap
      call DDL_Smv(a_ddl, p_ddl, v_ddl, gamma, delta, ierr)

      call DDL_Nrm2(x_ddl, xnrm, ierr)
      call DDL_Nrm2(rcgs_ddl, rnrm, ierr)
      call DDL_Nrm2(b_ddl, bnrm, ierr)

      rnrm0 = rnrm
      omega = rnrm
      tau = rnrm
      theta = 0.0d0
      eta = 0.0d0
      iter = 0
      tolcount = 1
      error = rnrm/bnrm
      rhonm1 = 1.0d0

      call DDL_Dot(r0_ddl, rcgs_ddl, rhonm1, ierr)

C*****************************************************************
C*****************************************************************
C
C                  START ITERATIVE LOOP
C
C*****************************************************************
C*****************************************************************
  

      do while ((error.GE.(tol(tolno))).AND.(iter.LT.itmax)) 
        iter = iter + 1


C*****************************************************************
C
C                      SECTION (c)
C
C*****************************************************************
        
         call DDL_Dot(r0_ddl, rcgs_ddl, rhon, ierr)

         betan = rhon/rhonm1

         if (iter.eq.1) then
            call DDL_Copy(rcgs_ddl, uv_ddl, ierr)
            call DDL_Copy(rcgs_ddl, p_ddl, ierr)
         else
            call DDL_Copy(rcgs_ddl, uv_ddl, ierr)
            call DDL_Axpy(betan, q_ddl, uv_ddl, ierr)
            call DDL_Copy(q_ddl, tmp_ddl, ierr)
            call DDL_Axpy(betan, p_ddl, tmp_ddl, ierr)
            call DDL_Copy(uv_ddl, p_ddl, ierr)
            call DDL_Axpy(betan, tmp_ddl, p_ddl, ierr)
         end if

         call DDL_Smv(a_ddl, p_ddl, v_ddl, gamma, delta, ierr)

         if (ilu) then
            call DDL_Copy(v_ddl, tmp_ddl, ierr)
            call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl,
     &                         v_ddl, ierr)
         else if (ilunc) then
            call DDL_Copy(v_ddl, tmp_ddl, ierr)
         call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &           v_ddl, nc, expinv,
     &           iwork, iwork_dim, rwork, rwork_dim, ierr)
         end if
      
C*****************************************************************
C
C                      SECTION (a)
C
C*****************************************************************

         call DDL_Dot(r0_ddl, v_ddl, sigma, ierr)
         alpha = rhon/sigma
         malpha = -alpha
         oldcgs = rnrm
         call DDL_Copy(uv_ddl, q_ddl, ierr)
         call DDL_Axpy(malpha, v_ddl, q_ddl, ierr)
         call DDL_Copy(uv_ddl, w_ddl, ierr)
         call DDL_Axpy(1.0d0, q_ddl, w_ddl, ierr)
         if (ilu) then
            call DDL_Smv(a_ddl, w_ddl, z_ddl, gamma, delta, ierr)
            call DDL_Copy(z_ddl, tmp_ddl, ierr)
            call DDL_Ilu_solve(l_ddl, u_ddl, dinv_ddl, tmp_ddl,
     &                         z_ddl, ierr)
            call DDL_Axpy(malpha, z_ddl, rcgs_ddl, ierr)
         else if (ilunc) then
            call DDL_Smv(a_ddl, w_ddl, z_ddl, gamma, delta, ierr)
            call DDL_Copy(z_ddl, tmp_ddl, ierr)
            call DDL_Ilunc_solve(a_ddl, rvec_ddl, ivec_ddl, tmp_ddl,
     &           z_ddl, nc, expinv,
     &           iwork, iwork_dim, rwork, rwork_dim, ierr)
            call DDL_Axpy(malpha, z_ddl, rcgs_ddl, ierr)
         else
            call DDL_Smv(a_ddl, w_ddl, tmp_ddl, gamma, delta, ierr)
            call DDL_Axpy(malpha, tmp_ddl, rcgs_ddl, ierr) 
         end if

C*****************************************************************
C
C                      SECTION (b)
C
C*****************************************************************

         call DDL_Nrm2(rcgs_ddl, rnrm, ierr)  
         omega = dsqrt(oldcgs*rnrm)
         m = 2*iter-1   
         const = theta*theta*eta/alpha
         theta = omega/tau
         c = 1.0d0/(dsqrt(1.0d0+theta*theta))
         tau = tau*theta*c
         eta = c*c*alpha
         call DDL_Scal(const, d_ddl, ierr)
         call DDL_Axpy(1.0d0, uv_ddl, d_ddl, ierr)
         call DDL_Axpy(eta, d_ddl, x_ddl, ierr) 
         error = dsqrt(dble(m+1))*dabs(tau)/bnrm
         omega = rnrm
         m = 2*iter
         const = theta*theta*eta/alpha
         theta = omega/tau
         c = 1.0d0/(dsqrt(1.0d0+theta*theta))
         tau = tau*theta*c
         eta = c*c*alpha
         call DDL_Scal(const, d_ddl, ierr) 
         call DDL_Axpy(1.0d0, q_ddl, d_ddl, ierr)  
         call DDL_Axpy(eta, d_ddl, x_ddl, ierr)
         error = dsqrt(dble(m+1))*dabs(tau)/bnrm
         call DDL_Nrm2(x_ddl, xnrm, ierr)  

         if (error.LT.(tol(tolcount))) then
            tolcount = tolcount + 1
         end if
        
         rhonm1 = rhon
      end do

c     remove distributed objects
      if (ilu) then
         call DDL_Free(l_ddl, ierr)
         call DDL_Free(u_ddl, ierr)
         call DDL_Free(dinv_ddl, ierr)
         call DDL_Free(z_ddl, ierr)
      end if
      if (ilunc) then
         call DDL_Free(ivec_ddl, ierr)
         call DDL_Free(rvec_ddl, ierr)
         call DDL_Free(z_ddl, ierr)
      end if
      call DDL_Free(rcgs_ddl, ierr)
      call DDL_Free(r0_ddl, ierr)
      call DDL_Free(uv_ddl, ierr)
      call DDL_Free(p_ddl, ierr)
      call DDL_Free(v_ddl, ierr)
      call DDL_Free(d_ddl, ierr)
      call DDL_Free(tmp_ddl, ierr)
      call DDL_Free(q_ddl, ierr)
      call DDL_Free(w_ddl, ierr)
 1000 end
