      SUBROUTINE DSblqm(N, B, X, NELT, IA, JA, A, ISYM, numblks, ITOL,
     $   TOL, ITMAX, ITER, ERR, IERR, IUNIT, RWORK, LENW, IWORK, LENIW )
c    Documentation needs to be considerably rewritten to allow for the
c    changes to allow a block factorisation.
C***BEGIN PROLOGUE  DSLUqm
C***DATE WRITTEN   890404   (YYMMDD)
C***REVISION DATE  890404   (YYMMDD)
C***CATEGORY NO.  D2A4
C***KEYWORDS  LIBRARY=SLATEC(SLAP),
C             TYPE=DOUBLE PRECISION(SSLUCS-D),
C             Non-Symmetric Linear system, Sparse,
C             Iterative incomplete LU Precondition
C***AUTHOR  Greenbaum, Anne, Courant Institute
C           Seager, Mark K., (LLNL)
C             Lawrence Livermore National Laboratory
C             PO BOX 808, L-300
C             Livermore, CA 94550 (415) 423-3141
C             seager@lll-crg.llnl.gov
C***PURPOSE  Incomplete LU BiConjugate Gradient Sparse Ax=b solver.
C            Routine to solve a linear system  Ax = b  using the
C            BiConjugate Gradient  method  with  Incomplete   LU
C            decomposition preconditioning.
C***DESCRIPTION
C *Usage:
C     INTEGER N, NELT, IA(NELT), JA(NELT), ISYM, ITOL, ITMAX
C     INTEGER ITER, IERR, IUNIT, LENW, IWORK(NEL+NU+4*N+2), LENIW
C     DOUBLE PRECISION B(N), X(N), A(NELT), TOL, ERR, RWORK(NEL+NU+8*N)
C
C     CALL DSLUqm(N, B, X, NELT, IA, JA, A, ISYM, ITOL, TOL,
C    $     ITMAX, ITER, ERR, IERR, IUNIT, RWORK, LENW, IWORK, LENIW)
C
C *Arguments:
C N      :IN       Integer.
C         Order of the Matrix.
C B      :IN       Double Precision B(N).
C         Right-hand side vector.
C X      :INOUT    Double Precision X(N).
C         On input X is your initial guess for solution vector.
C         On output X is the final approximate solution.
C NELT   :IN       Integer.
C         Number of Non-Zeros stored in A.
C IA     :INOUT    Integer IA(NELT).
C JA     :INOUT    Integer JA(NELT).
C A      :INOUT    Double Precision A(NELT).
C         These arrays should hold the matrix A in either the SLAP
C         Triad format or the SLAP Column format.  See "Description",
C         below.  If the SLAP Triad format is chosen it is changed
C         internally to the SLAP Column format.
C ISYM   :IN       Integer.
C         Flag to indicate symmetric storage format.
C         If ISYM=0, all nonzero entries of the matrix are stored.
C         If ISYM=1, the matrix is symmetric, and only the upper
C         or lower triangle of the matrix is stored.
C ITOL   :IN       Integer.
C         Flag to indicate type of convergence criterion.
C         If ITOL=1, iteration stops when the 2-norm of the residual
C         divided by the 2-norm of the right-hand side is less than TOL.
C         This routine must calculate the residual from R = A*X - B.
C         This is un-natural and hence expensive for this type of iter-
C         ative method.  ITOL=2 is *STRONGLY* recommended.
C         If ITOL=2, iteration stops when the 2-norm of M-inv times the
C         residual divided by the 2-norm of M-inv times the right hand
C         side is less than tol, where M-inv time a vector is the pre-
C         conditioning step.  This is the *NATURAL* stopping for this
C         iterative method and is *STRONGLY* recommended.
C TOL    :IN       Double Precision.
C         Convergence criterion, as described above.
C ITMAX  :IN       Integer.
C         Maximum number of iterations.
C ITER   :OUT      Integer.
C         Number of iterations required to reach convergence, or
C         ITMAX+1 if convergence criterion could not be achieved in
C         ITMAX iterations.
C ERR    :OUT      Double Precision.
C         Error estimate of error in final approximate solution, as
C         defined by ITOL.
C IERR   :OUT      Integer.
C         Return error flag.
C           IERR = 0 => All went well.
C           IERR = 1 => Insufficient storage allocated
C                       for WORK or IWORK.
C           IERR = 2 => Method failed to converge in
C                       ITMAX steps.
C           IERR = 3 => Error in user input.  Check input
C                       value of N, ITOL.
C           IERR = 4 => User error tolerance set too tight.
C                       Reset to 500.0*D1MACH(3).  Iteration proceeded.
C           IERR = 5 => Breakdown of the method detected.
C                       $(r0,r) approximately 0.0$.
C           IERR = 6 => Stagnation of the method detected.
C                        $(r0,v) approximately 0.0$.
C           IERR = 7 => Incomplete factorization broke down
C                       and was fudged.  Resulting preconditioning may
C                       be less than the best.
C IUNIT  :IN       Integer.
C         Unit number on which to write the error at each iteration,
C         if this is desired for monitoring convergence.  If unit
C         number is 0, no writing will occur.
C RWORK  :WORK     Double Precision RWORK(LENW).
C         Double Precision array used for workspace.  NEL is the
C         number of non-
C         zeros in the lower triangle of the matrix (including the
C         diagonal).  NU is the number of nonzeros in the upper
C         triangle of the matrix (including the diagonal).
C LENW   :IN       Integer.
C         Length of the double precision workspace, RWORK.
C         LENW >= NEL+NU+9*N.
C IWORK  :WORK     Integer IWORK(LENIW).
C         Integer array used for workspace.  NEL is the number of non-
C         zeros in the lower triangle of the matrix (including the
C         diagonal).  NU is the number of nonzeros in the upper
C         triangle of the matrix (including the diagonal).
C         Upon return the following locations of IWORK hold information
C         which may be of use to the user:
C         IWORK(9)  Amount of Integer workspace actually used.
C         IWORK(10) Amount of Double Precision workspace actually used.
C LENIW  :IN       Integer.
C         Length of the integer workspace, IWORK.
C         LENIW >= NEL+NU+4*N+12.
C
C *Description:
C       This routine is simply a  driver for the DCGSN  routine.  It
C       calls the DSILUS routine to set  up the  preconditioning and
C       then  calls DCGSN with  the appropriate   MATVEC, MTTVEC and
C       MSOLVE, MTSOLV routines.
C
C       The Sparse Linear Algebra Package (SLAP) utilizes two matrix
C       data structures: 1) the  SLAP Triad  format or  2)  the SLAP
C       Column format.  The user can hand this routine either of the
C       of these data structures and SLAP  will figure out  which on
C       is being used and act accordingly.
C
C       =================== S L A P Triad format ===================
C
C       This routine requires that the  matrix A be   stored in  the
C       SLAP  Triad format.  In  this format only the non-zeros  are
C       stored.  They may appear in  *ANY* order.  The user supplies
C       three arrays of  length NELT, where  NELT is  the number  of
C       non-zeros in the matrix: (IA(NELT), JA(NELT), A(NELT)).  For
C       each non-zero the user puts the row and column index of that
C       matrix element  in the IA and  JA arrays.  The  value of the
C       non-zero   matrix  element is  placed  in  the corresponding
C       location of the A array.   This is  an  extremely  easy data
C       structure to generate.  On  the  other hand it   is  not too
C       efficient on vector computers for  the iterative solution of
C       linear systems.  Hence,   SLAP changes   this  input    data
C       structure to the SLAP Column format  for  the iteration (but
C       does not change it back).
C
C       Here is an example of the  SLAP Triad   storage format for a
C       5x5 Matrix.  Recall that the entries may appear in any order.
C
C           5x5 Matrix       SLAP Triad format for 5x5 matrix on left.
C                              1  2  3  4  5  6  7  8  9 10 11
C       |11 12  0  0 15|   A: 51 12 11 33 15 53 55 22 35 44 21
C       |21 22  0  0  0|  IA:  5  1  1  3  1  5  5  2  3  4  2
C       | 0  0 33  0 35|  JA:  1  2  1  3  5  3  5  2  5  4  1
C       | 0  0  0 44  0|
C       |51  0 53  0 55|
C
C       =================== S L A P Column format ==================
C       This routine requires  that the  matrix  A be  stored in the
C       SLAP Column format.  In this format the non-zeros are stored
C       counting down columns (except for  the diagonal entry, which
C       must appear first in each  "column")  and are stored  in the
C       double precision array A.   In other words,  for each column
C       in the matrix put the diagonal entry in  A.  Then put in the
C       other non-zero  elements going down  the column (except  the
C       diagonal) in order.   The  IA array holds the  row index for
C       each non-zero.  The JA array holds the offsets  into the IA,
C       A arrays  for  the  beginning  of each   column.   That  is,
C       IA(JA(ICOL)),  A(JA(ICOL)) points   to the beginning  of the
C       ICOL-th   column    in    IA and   A.      IA(JA(ICOL+1)-1),
C       A(JA(ICOL+1)-1) points to  the  end of the   ICOL-th column.
C       Note that we always have  JA(N+1) = NELT+1,  where N is  the
C       number of columns in  the matrix and NELT  is the number  of
C       non-zeros in the matrix.
C
C       Here is an example of the  SLAP Column  storage format for a
C       5x5 Matrix (in the A and IA arrays '|'  denotes the end of a
C       column):
C
C       5x5 Matrix      SLAP Column format for 5x5 matrix on left.
C       1  2  3    4  5    6  7    8    9 10 11
C       |11 12  0  0 15|   A: 11 21 51 | 22 12 | 33 53 | 44 | 55 15 35
C       |21 22  0  0  0|  IA:  1  2  5 |  2  1 |  3  5 |  4 |  5  1  3
C       | 0  0 33  0 35|  JA:  1  4  6    8  9   12
C       | 0  0  0 44  0|
C       |51  0 53  0 55|
C
C *Precision:           Double Precision
C *Side Effects:
C       The SLAP Triad format (IA, JA,  A) is modified internally to
C       be the   SLAP Column  format.  See above.
C
C *See Also:
C       DCGS, DSDCGS
C***REFERENCES  1. P. Sonneveld, ``CGS, a fast Lanczos-type solver
C                 for nonsymmetric linear systems'', Delft University
C                 of Technology Report 84-16, Department of Math-
C                 ematics and Informatics, Julianalaan 132, 2628 BL
C                 Delft, Phone 015-784568.
C
C               2. E.F. Kaasschieter, ``The solution of non-symmetric
C                 linear systems by bi-conjugate gradients or conjugate
C                 gradients squared,''  Delft University of Tech-
C                 nology Report 86-21, Department of Mathematics and
C                 Informatics, Julianalaan 132, 2628 BL Delft,
C                 Phone 015-784568.
C***ROUTINES CALLED  DS2Y, DCHKW, DSILUS, DCGS, DSMV, DSLUI
C***END PROLOGUE  DSLUCS
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      INTEGER N, NELT, IA(NELT), JA(NELT), ISYM, ITOL, ITMAX, ITER
      INTEGER IERR, IUNIT, LENW, IWORK(LENIW), LENIW
      DOUBLE PRECISION B(N), X(N), A(NELT), TOL, ERR, RWORK(LENW)
      DOUBLE PRECISION secs, second
      integer i1mach
      EXTERNAL  DSMV, blksol, second, i1mach
c
c     Declarations for the UMFPACK material
c
      INTEGER PMEM(7,2), PRLEV, umferr, ihead, itail, xhead, xtail,
     1        ndupl, ndrop, ninvld, lu(2,2), losrch, hisrch, nb, extra,
     2        npiv, igar, xgar, io, nzblk, solwork, ablk, iptrblk,
     3        blksize, matstep, pivstep, modmat, modpiv, div, rem
      logical trans, dupl, drop, scale, block, symsrc, dense
      double precision dtol, asrdum, ascdum, relpt, abspt, gro, pmin
      common/umfreal/abspt
      common/umfint/nzblk,solwork,pmem,lu,prlev,io
      common/denseinfo/maxblk,nblks,ablk,iptrblk,dense
C
C         Change the SLAP input matrix IA, JA, A to SLAP-Column format.
C***FIRST EXECUTABLE STATEMENT  DSLUCS
      IERR = 0
      IF( N.LT.1 .OR. NELT.LT.1 ) THEN
         IERR = 3
         RETURN
      ENDIF
      CALL DS2Y( N, NELT, IA, JA, A, ISYM )
C
C
      io = I1MACH(2)
      LOCR   = 1
      LOCD   = LOCR + N
      LOCR0  = LOCD + N
      LOCP   = LOCR0 + N
      LOCQ   = LOCP + N
      LOCU   = LOCQ + N
      LOCV1  = LOCU + N
      LOCV2  = LOCV1 + N
      iptrblk = 1
c
c     If there is enough memory, and the maximum block size is smaller
c     than 20, use a dense solver. This initial version uses a general
c     LU routine, a banded solver from LAPACK would be a better choice.
c
      rem = mod(n,numblks)
      div = n / numblks
      if (rem .eq. 0) then 
         maxblk = div
      else 
         maxblk = div + 1
      end if
      dense = maxblk .le. 20 .and. numblks*maxblk**2 .lt. lenw-8*n+1
      secs = second()
      if (dense) then 
	 write(io,*) 'Doing dense decomposition.'
	 ablk = LOCV2 + N 
	 infoerrs = iptrblk + numblks*maxblk 
	 nblks = numblks
c
c        Set the area for the matrices to zero
c        and then put the nonzero elements in their
c        correct place.
c
         call dfill(numblks*maxblk**2,rwork(ablk),0.d0)
	 call formden(n,nelt,ia,ja,a,numblks,rwork(ablk))
	 call xlaenv(1,1)
	 matstep = maxblk**2
	 do i=0,rem-1
	    call dgetrf(maxblk,maxblk,rwork(i*matstep+ablk),maxblk,
     1  	iwork(i*maxblk+iptrblk),iwork(infoerrs+i))
	 end do
	 modmat = rem*matstep + ablk
	 modpiv = rem*maxblk + iptrblk
	 matstep = div**2
	 do i=rem,numblks-1
	    call dgetrf(div,div,rwork((i-rem)*matstep+modmat),
     1  	div,iwork((i-rem)*div+modpiv),iwork(infoerrs+i))
	 end do
	 do i = 0,numblks-1
	   if (iwork(infoerrs+i) .ne. 0) then 
	      write(io,*) 'Dense block ',i,' is singular, info= ',
     1		iwork(infoerrs+i)
	      stop
           endif
         end do
      else
         solwork =  LOCV2 + N
         ablk = solwork + 2*n
         xhead = ablk + nelt
         xtail = lenw+1
         ihead = iptrblk + 2*nelt+1
         itail = leniw + 1
         prlev = 1
c
c        This version of formblk is (unfortunately) specific to
c        UMFPACK in that the row and column indices are adjacent in
c        memory, rather than in two column vectors.
c
         call formblk(n,nelt,ia,ja,a,numblks,nzblk,iwork(iptrblk),
     1             rwork(ablk))
c
c        Set-up the workspace for the UMF routines
c
         call umdini(rwork, iwork, pmem, prlev, iunit, umferr, ihead,
     1            itail, xhead, xtail)
         if (umferr .ne. 0) then
           write(io, *) 'Error in umdini', umferr
           stop
         end if
c
c        Set-up the basic parameters for the factorisation. We assume that
c        we do not what the symmetric problem solved and that scaling is not
c        performed. Other than that, these parameters can change  -- testing
c        will reveal those that will propagate into the calling sequence.
c
         trans = .false.
         dupl = .false.
         drop = .true.
         dtol = 1.d-14
         block = .false.
         scale = .false.
         symsrc = .false.
         relpt = 0.1d0
         abspt = 0.d0
         gro = 2.d0
         losrch = 0
         hisrch = 4
c
c        The block size is normally around 64, but with a preconditioner, only
c        smaller blocks are involved, so the block size for supernodes is 
c        set to the small value of 20.
c
         nb = 20
         extra = 5

c
c        Obtain the block lu preconditioner
c
         call umdfac(rwork,iwork,pmem,rwork(ablk), iwork(iptrblk), n,
     1             nzblk,trans, dupl, drop, dtol, ndupl, ndrop, ninvld,
     2             asrdum, ascdum, .false., lu, prlev, iunit, umferr,
     3             block,relpt,abspt,gro,losrch,hisrch,symsrc,nb,extra,
     4             npiv, pmin, igar, xgar)
         if (umferr .ne. 0) then
            write(io,*) 'Error in umdfac: ',umferr
            stop
         end if
      endif
      secs = second() - secs
      write(io,*) 'Set-up required: ',secs,' seconds.'
C
C
C     Perform the incomplete LU preconditioned
C     BiConjugate Gradient Squared algorithm.
C
      CALL Dqmr(N, B, X, NELT, IA, JA, A, ISYM, DSMV,
     $     blksol, ITOL, TOL, ITMAX, ITER, ERR, IERR, IUNIT,
     $     rwork(locd), RWORK(LOCR), RWORK(LOCR0), RWORK(LOCP),
     $     RWORK(LOCQ), RWORK(LOCU), RWORK(LOCV1),
     $     RWORK(LOCV2), RWORK, IWORK )
      RETURN
C------------- LAST LINE OF DSLUCS FOLLOWS ----------------------------
      END

      subroutine formden(n,nz,ia,ja,a,numblks,ablk)
c
c     This routine pulls the numblks block diagonal matrices out of 
c     the sparse matrix (ia,ja,a) and forms numblks dense matrix problems
c     in the array ablk and these can then be factored individually using a 
c     standard LINPACK or LAPACK call. 
c
c     CRITICAL ASSUMPTIONS: This routine assumes that the memory for the
c     matrix ablk has already been zeroed, preferably via a data statement
c     in the main program (to speed execution). It only puts the non-zero
c     elements into the correct block matrices.
c
c     In addition, this routine assumes that the input matrix is either in
c     triad format or in compressed column format (equivalently the 
c     Harwell Boeing format or the SLAP column format).
c
      integer n, nz, ia(*), ja(*), numblks 
      double precision a(*), ablk(*)
c
c     Local variables
c
      integer count, i, j, iold, blocksize, k, rem, div, threshold,
     1        mylow, myhigh, row, col, mathresh, offset, divsq
      logical triad
      triad = ja(n+1) .ne. nz + 1
c
c     Be able to account for the number of blocks not dividing the problem
c     size evenly, in which case, the first rem blocks are of size div+1
c     and the remaining numblks - rem blocks are of size div.
c
      rem = mod(n,numblks)
      div = n / numblks
      divsq = div**2
      divp1sq = (div+1)**2
      threshold = (div+1)*rem
      mathresh = (div+1)**2*rem
c
      if (triad) then
         iold = 0
         offset = 0
         do count = 1,nz
c
c           Triad format has no restrictions on ordering, so none
c           can be assumed, although the Math Tech problems TEND
c           to group entries by row.
c
            if (ia(count) .ne. iold) then
               iold = ia(count)
               if (iold .le. threshold) then
                  blocksize = div+1
                  k = (iold-1)/blocksize
                  mylow = k*blocksize
                  myhigh = mylow  + blocksize
                  offset = k*divp1sq - mylow
               else
                  blocksize = div
                  k = (iold-threshold-1)/div
                  mylow = threshold + k*div
                  myhigh = mylow  + div
                  offset = mathresh + k*divsq - mylow
               end if
             end if
             if (ja(count) .gt. mylow .and.
     1           ja(count) .le. myhigh) then
                row = ia(count)  + offset
 		col = ja(count)  - mylow
                ablk(row+(col-1)*blocksize) = a(count)
             end if
         end do
      else
         kold = -1
	 offset = 0
         do j = 1,n
            if ( j .le. threshold) then
               blocksize = div + 1
               k = (j-1)/blocksize
               mylow = k*blocksize
               myhigh = mylow  + blocksize
  	       offset =  k*divp1sq - mylow
            else
               blocksize = div 
               k = (j-threshold-1)/div
               if (k .ne. kold) then
                  mylow = threshold + k*div
                  myhigh = mylow  + div
                  kold = k
	          offset = mathresh + k*divsq - mylow
               end if
            end if
	    col = (j-1-mylow)*blocksize
            do i = ja(j), ja(j+1)-1
               if (ia(i) .gt. mylow .and.
     1             ia(i) .le. myhigh) then
                 row = ia(i) + offset
                 ablk(row+col) = a(i)
               end if
            end do
         end do
      end if
      end



      subroutine formblk(n,nz,ia,ja,a,numblks,nzblk,iptrblk,ablk)
      integer n, nz, ia(*), ja(*), numblks, nzblk, iptrblk(2,*)
      double precision a(*), ablk(*)
c
c     Routine to produce a block sub-matrix in triad format from an
c     input matrix in either triad or gathered column (aka SLAP column or
c     Harwell-Boeing format) given the input matrix and the number of blocks
c     to be used.
c
c     Warning: while the routine will happily define a block matrix with
c     unequal blocksizes, there is the possibility that such a block matrix
c     will be a very poor preconditioner. If the underlying problem arises
c     from a discretisation over an x by y  by z grid, it is strongly
c     recommended that n = k * x * numblks for an integer value k.
c
c
c     This version of formblk is (unfortunately) specific to
c     UMFPACK in that the row and column indices are adjacent in
c     memory, rather than in two column vectors.
c
c     Initial version:
c     19/07/93
c     C. Addison
c
      integer count, i, j, iold, blocksize, k, rem, div, threshold,
     1        mylow, myhigh
      logical triad
      triad = ja(n+1) .ne. nz + 1
c
c     Be able to account for the number of blocks not dividing the problem
c     size evenly, in which case, the first rem blocks are of size div+1
c     and the remaining numblks - rem blocks are of size div.
c
c     write(*,*) 'formblk from n, nz: ',n, nz
      rem = mod(n,numblks)
      div = n / numblks
      threshold = (div+1)*rem
c     write(*,*) rem, div, threshold, n, numblks
c
      if (triad) then
         nzblk = 0
         iold = 0
         do count = 1,nz
c
c           Triad format has no restrictions on ordering, so none
c           can be assumed, although the Math Tech problems TEND
c           to group entries by row.
c
            if (ia(count) .ne. iold) then
               iold = ia(count)
               if (iold .le. threshold) then
                  blocksize = div+1
                  k = (iold-1)/blocksize
                  mylow = k*blocksize
                  myhigh = mylow  + blocksize
               else
                  k = (iold-threshold-1)/div
                  mylow = threshold + k*div
                  myhigh = mylow  + div
               end if
             end if
             if (ja(count) .gt. mylow .and.
     1           ja(count) .le. myhigh) then
                nzblk = nzblk + 1
                ablk(nzblk) = a(count)
                iptrblk(1,nzblk) = ia(count)
                iptrblk(2,nzblk) = ja(count)
             end if
         end do
      else
         nzblk = 0
	 kold = -1
         do j = 1,n
            if ( j .le. threshold) then
               blocksize = div + 1
               k = (j-1)/blocksize
               mylow = k*blocksize
               myhigh = mylow  + blocksize
            else
               k = (j-threshold-1)/div
	       if (k .ne. kold) then 
                  mylow = threshold + k*div
                  myhigh = mylow  + div
c		  write(*,*) '>>> ',k,mylow,myhigh,j
		  kold = k
               end if
            end if
            do i = ja(j), ja(j+1)-1
               if (ia(i) .gt. mylow .and.
     1             ia(i) .le. myhigh) then
                 nzblk = nzblk + 1
                 ablk(nzblk) = a(i)
                 iptrblk(1,nzblk) = ia(i)
                 iptrblk(2,nzblk) = j
               end if
            end do
         end do
      end if
      end

      subroutine blksol(n,b,x,nz,ia, ja, a,isym,rwork,iwork)
      integer n, nz, ia(*), ja(*), isym, iwork(*)
      double precision b(*), x(*), a(*), rwork(*)
      INTEGER pmem(7,2), prlev, umferr, lu(2,2),
     1        npiv, io, nzblk, solwork, vl, info, i, 
     2        matstep, modmat,modpiv,modx, numblks, ablk,
     3	      iptrblk, rem, div
      logical trans, scale,dense
      double precision asrdum, ascdum, abspt
      common/umfreal/abspt
      common/umfint/nzblk,solwork,pmem,lu,prlev,io
      common/denseinfo/maxblk,numblks,ablk,iptrblk,dense
      if (dense) then 
         call dcopy(n,b,1,x,1)
         div = n / numblks
	 rem = mod(n,numblks)
         matstep = maxblk**2
         do i=0,rem-1
            call dgetrs('N',maxblk,1,rwork(i*matstep+ablk),maxblk,
     1          iwork(i*maxblk+iptrblk),x(i*maxblk+1),maxblk,
     2		info)
         end do
         modmat = rem*matstep + ablk
         modpiv = rem*maxblk + iptrblk
         modx = rem*maxblk + 1
         matstep = div**2
         do i=rem,numblks-1
            call dgetrs('N',div,1,rwork((i-rem)*matstep+modmat),
     1          div,iwork((i-rem)*div+modpiv),x(modx+(i-rem)*div),
     2		div,info)
         end do
      else
        vl = 8
        scale = .false.
        trans = .false.
        call umdsol(rwork, iwork, pmem, asrdum, ascdum, scale, lu,
     1            prlev, io, umferr, b, trans, abspt, vl, x, npiv,
     2            rwork(solwork))
        if (umferr .ne. 0) then
          write(io,*) 'Error in umdsol: ',umferr
          return
        endif
      endif
      end
